Object subclass: #IRBuilder
	instanceVariableNames: 'ir tempMap jumpBackTargetStacks jumpAheadStacks currentSequence sourceMapNodes sourceMapByteIndex'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-IR'!
!IRBuilder commentStamp: 'ajh 6/25/2004 11:23' prior: 0!
I provide a simple interface for constructing an IRMethod.  For example, to create an ir method that compares first instVar to first arg and returns 'yes' or 'no' (same example as in BytecodeGenerator), do:

	IRBuilder new
		numRargs: 2;
		addTemps: #(self a z);		"rcvr, arg, & extra temp (not used here)"
		pushTemp: #self;
		getField: 1;
		pushTemp: #a;
		send: #>;
		jumpAheadTo: #else if: false;
		pushLiteral: 'yes';
		returnTop;
		jumpAheadTarget: #else;
		pushLiteral: 'no';
		returnTop;
		ir

Sending #compiledMethod to an ir method will generate its compiledMethod.  Sending #methodNode to it will decompile to its parse tree.
!


!IRBuilder methodsFor: 'initialize' stamp: 'ajh 6/30/2004 14:28'!
addTemp: tempKey

	self addTemps: {tempKey}! !

!IRBuilder methodsFor: 'initialize' stamp: 'ajh 6/30/2004 14:28'!
addTemps: newKeys

	| keys i new |
	keys _ ir tempKeys.
	i _ keys size - 1. "zero-based (index 0 equals receiver)"
	new _ OrderedCollection new.
	newKeys do: [:key |
		tempMap at: key ifAbsentPut: [
			new add: key.
			i _ i + 1]
	].
	ir tempKeys: keys, new.
! !

!IRBuilder methodsFor: 'initialize' stamp: 'ajh 6/25/2004 11:24'!
numRargs: n

	ir numRargs: n.
! !

!IRBuilder methodsFor: 'initialize' stamp: 'ajh 3/10/2003 14:10'!
primitiveNode: primNode

	ir primitiveNode: primNode! !


!IRBuilder methodsFor: 'instructions' stamp: 'ajh 7/2/2004 13:43'!
getField: instVarIndex
	"Receiver must be on top"

	self pushLiteral: instVarIndex.
	self send: #privGetField:.
! !

!IRBuilder methodsFor: 'instructions' stamp: 'ajh 6/22/2003 14:25'!
jumpAheadTarget: labelSymbol
	"Pop latest jumpAheadTo: with this labelSymbol and have it point to this new instruction sequence"

	| jumpInstr |
	self startNewSequence.
	jumpInstr _ (jumpAheadStacks at: labelSymbol ifAbsent: [self error: 'Missing jumpAheadTo: ', labelSymbol printString]) removeLast.
	jumpInstr destination: currentSequence.
! !

!IRBuilder methodsFor: 'instructions' stamp: 'ajh 6/22/2003 14:44'!
jumpAheadTo: labelSymbol
	"Jump to the sequence that will be created when jumpAheadTarget: labelSymbol is sent to self.  This is and its corresponding target is only good for one use.  Other jumpAheadTo: with the same label will be put on a stack and superceed existing ones until its jumpAheadTarget: is called."

	"jumpAheadTarget: label will pop this and replace destination with its basic block"
	(jumpAheadStacks at: labelSymbol ifAbsentPut: [OrderedCollection new])
		addLast: (self add: IRJump new).
	self startNewSequence.
! !

!IRBuilder methodsFor: 'instructions' stamp: 'ajh 6/22/2003 14:44'!
jumpAheadTo: labelSymbol if: boolean
	"Conditional jump to the sequence that will be created when jumpAheadTarget: labelSymbol is sent to self.  This and its corresponding target is only good for one use.  Other jumpAheadTo:... with the same label will be put on a stack and superceed existing ones until its jumpAheadTarget: is called."

	| instr |
	"jumpAheadTarget: label will pop this and replace destination with its basic block"
	(jumpAheadStacks at: labelSymbol ifAbsentPut: [OrderedCollection new])
		addLast: (instr _ self add: (IRJumpIf new boolean: boolean)).
	self startNewSequence.
	instr otherwise: currentSequence.
! !

!IRBuilder methodsFor: 'instructions' stamp: 'ajh 6/22/2003 14:44'!
jumpBackTarget: labelSymbol
	"Remember this basic block for a future jumpBackTo: labelSymbol.  Stack up remembered targets with same name and remove them from stack for each jumpBackTo: called with same name."

	self startNewSequence.
	(jumpBackTargetStacks at: labelSymbol ifAbsentPut: [OrderedCollection new])
		addLast: currentSequence.
! !

!IRBuilder methodsFor: 'instructions' stamp: 'ajh 6/22/2003 14:25'!
jumpBackTo: labelSymbol
	"Pop last remembered position with this label and write an unconditional jump to it"

	| sequence |
	sequence _ (jumpBackTargetStacks at: labelSymbol ifAbsent: [self error: 'Missing jumpBackTarget: ', labelSymbol printString]) removeLast.
	self add: (IRJump new destination: sequence).
	self startNewSequence.
! !

!IRBuilder methodsFor: 'instructions' stamp: 'ajh 3/10/2003 01:09'!
popTop

	self add: IRInstruction popTop! !

!IRBuilder methodsFor: 'instructions' stamp: 'ajh 6/27/2004 21:38'!
pushBlock: irMethod

	self add: (IRInstruction pushBlock: irMethod)! !

!IRBuilder methodsFor: 'instructions' stamp: 'ajh 6/27/2004 21:38'!
pushBlockMethod: irMethod

	self add: (IRInstruction pushBlockMethod: irMethod)! !

!IRBuilder methodsFor: 'instructions' stamp: 'ajh 3/10/2003 01:09'!
pushDup

	self add: IRInstruction pushDup! !

!IRBuilder methodsFor: 'instructions' stamp: 'ajh 3/10/2003 01:09'!
pushLiteral: object

	self add: (IRInstruction pushLiteral: object)! !

!IRBuilder methodsFor: 'instructions' stamp: 'ajh 3/10/2003 14:52'!
pushReceiver

	self add: (IRInstruction pushTemp: 0)! !

!IRBuilder methodsFor: 'instructions' stamp: 'ajh 3/17/2003 11:06'!
pushTemp: key

	| index |
	index _ tempMap at: key.
	self add: (IRInstruction pushTemp: index)! !

!IRBuilder methodsFor: 'instructions' stamp: 'ajh 6/25/2004 17:41'!
pushThisContext

	self add: (IRInstruction pushTemp: -2)! !

!IRBuilder methodsFor: 'instructions' stamp: 'ajh 6/26/2004 13:47'!
pushThisEnv

	self add: (IRInstruction pushTemp: -1)! !

!IRBuilder methodsFor: 'instructions' stamp: 'ajh 3/15/2003 01:55'!
remoteReturn

	self add: IRInstruction remoteReturn.
	self startNewSequence.
! !

!IRBuilder methodsFor: 'instructions' stamp: 'ajh 3/15/2003 01:55'!
returnTop

	self add: IRInstruction returnTop.
	self startNewSequence.
! !

!IRBuilder methodsFor: 'instructions' stamp: 'ajh 3/10/2003 01:09'!
send: selector

	self add: (IRInstruction send: selector)! !

!IRBuilder methodsFor: 'instructions' stamp: 'ajh 3/10/2003 01:09'!
send: selector toSuperOf: behavior

	self add: (IRInstruction send: selector toSuperOf: behavior)! !

!IRBuilder methodsFor: 'instructions' stamp: 'ajh 7/3/2004 15:15'!
setField: instVarIndex
	"receiver must be on top with new field value underneath"

	self pushLiteral: instVarIndex.
	self send: #privStoreIn:field:.
! !

!IRBuilder methodsFor: 'instructions' stamp: 'ajh 3/17/2003 11:06'!
storeTemp: key

	| index |
	index _ tempMap at: key.
	self add: (IRInstruction storeTemp: index)! !

!IRBuilder methodsFor: 'instructions' stamp: 'ajh 6/26/2004 13:48'!
storeThisEnv

	self add: (IRInstruction storeTemp: -1)! !


!IRBuilder methodsFor: 'private' stamp: 'ajh 3/13/2003 13:20'!
add: instr

	"Associate instr with current parse node or byte range"
	instr sourceNode: self sourceNode.
	instr bytecodeIndex: self sourceByteIndex.
	^ currentSequence add: instr! !

!IRBuilder methodsFor: 'private' stamp: 'ajh 3/17/2003 11:06'!
initialize

	ir _ IRMethod new.
	tempMap _ Dictionary new.
	jumpAheadStacks _ IdentityDictionary new.
	jumpBackTargetStacks _ IdentityDictionary new.
	sourceMapNodes _ OrderedCollection new.  "stack"

	"Leave an empty sequence up front (guaranteed not to be in loop)"
	ir startSequence: (IRSequence new orderNumber: 0).
	currentSequence _ IRSequence new orderNumber: 1.
	ir startSequence add: (IRJump new destination: currentSequence).
! !

!IRBuilder methodsFor: 'private' stamp: 'ajh 3/10/2003 17:45'!
startNewSequence
	"End current instruction sequence and start a new sequence to add instructions to.  If ending block just falls through to new block then add an explicit jump to it so they stay linked"

	| newSequence |
	currentSequence isEmpty ifTrue: [^ self].  "block is still empty, continue using it"
	newSequence _ IRSequence new
		orderNumber: currentSequence orderNumber + 1.
	currentSequence last isJumpOrReturn ifFalse: [
		self add: (IRJump new destination: newSequence)].
	currentSequence _ newSequence.
! !


!IRBuilder methodsFor: 'mapping' stamp: 'ajh 3/13/2003 13:17'!
mapToByteIndex: index
	"decompiling"

	sourceMapByteIndex _ index! !

!IRBuilder methodsFor: 'mapping' stamp: 'ajh 6/22/2003 14:45'!
mapToNode: object
	"new instructions will be associated with object"

	sourceMapNodes addLast: object! !

!IRBuilder methodsFor: 'mapping' stamp: 'ajh 6/22/2003 14:25'!
popMap

	sourceMapNodes removeLast! !

!IRBuilder methodsFor: 'mapping' stamp: 'ajh 3/13/2003 13:19'!
sourceByteIndex
	"decompiling"

	^ sourceMapByteIndex! !

!IRBuilder methodsFor: 'mapping' stamp: 'ajh 6/22/2003 21:03'!
sourceNode

	^ sourceMapNodes isEmpty
		ifTrue: [nil]
		ifFalse: [sourceMapNodes last]! !


!IRBuilder methodsFor: 'results' stamp: 'ajh 3/10/2003 15:51'!
ir

	^ ir! !


!IRBuilder methodsFor: 'decompiling' stamp: 'ajh 6/22/2003 14:44'!
addJumpBackTarget: label to: sequence

	(jumpBackTargetStacks at: label ifAbsentPut: [OrderedCollection new])
		addLast: sequence! !

!IRBuilder methodsFor: 'decompiling' stamp: 'ajh 3/21/2003 01:48'!
testJumpAheadTarget: label

	jumpAheadStacks at: label ifPresent: [:stack |
		[stack isEmpty] whileFalse: [self jumpAheadTarget: label]
	]! !

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

IRBuilder class
	instanceVariableNames: ''!

!IRBuilder class methodsFor: 'as yet unclassified' stamp: 'ajh 3/10/2003 17:56'!
new

	^ super new initialize! !


Link subclass: #IRInstruction
	instanceVariableNames: 'sourceNode bytecodeIndex'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-IR'!
!IRInstruction commentStamp: 'ajh 6/27/2004 23:02' prior: 0!
I am an instruction in the IR (intermediate representation) language.  The IR serves as the intermediary between the Smalltalk language and the bytecode language.  It is easier to optimize and translate to/from this language than it is to optimize/translate directly from Smalltalk to bytecodes.  The IR is generic and simple consisting of just twelve instructions.  They are:

	goto: labelNum
	if: boolean goto: labelNum1 otherwise: labelNum2
	label: labelNum
	popTop
	pushDup
	pushLiteral: object
	pushBlock: irMethod
	pushBlockMethod: irMethod
	pushTemp: tempIndex
	remoteReturn
	returnTop
	send: selector
	send: selector toSuperOf: behavior
	storeTemp: tempIndex

Each instruction is reified as an instance of one of my eight subclasses and grouped by basic block (IRSequence) into an IRMethod.  IRInterpreter visits each instruction in a IRMethod responding to the above instruction messages sent to it.
!


!IRInstruction methodsFor: 'accessing' stamp: 'ajh 3/6/2003 14:32'!
executeOn: interpreter
	"Send approriate message to interpreter"

	self subclassResponsibility! !

!IRInstruction methodsFor: 'accessing' stamp: 'ajh 3/6/2003 14:32'!
successorSequences
	"sent to last instruction in sequence which is expected to be a jump and return instruction"

	^ #()! !


!IRInstruction methodsFor: 'testing' stamp: 'ajh 3/10/2003 16:12'!
isConstant: valueTest

	^ false! !

!IRInstruction methodsFor: 'testing' stamp: 'ajh 3/11/2003 00:19'!
isGoto
	"is unconditional jump"

	^ false! !

!IRInstruction methodsFor: 'testing' stamp: 'ajh 3/11/2003 00:21'!
isIf

	^ false! !

!IRInstruction methodsFor: 'testing' stamp: 'ajh 3/11/2003 00:22'!
isJump
	"goto or if"

	^ false! !

!IRInstruction methodsFor: 'testing' stamp: 'ajh 3/10/2003 16:13'!
isJumpOrReturn

	^ self isJump or: [self isReturn]! !

!IRInstruction methodsFor: 'testing' stamp: 'ajh 3/10/2003 16:09'!
isReturn

	^ false! !


!IRInstruction methodsFor: 'mapping' stamp: 'ajh 3/13/2003 13:07'!
bytecodeIndex

	^ bytecodeIndex! !

!IRInstruction methodsFor: 'mapping' stamp: 'ajh 3/13/2003 13:14'!
bytecodeIndex: index

	bytecodeIndex _ index! !

!IRInstruction methodsFor: 'mapping' stamp: 'ajh 3/6/2003 14:32'!
sourceNode

	^ sourceNode
	! !

!IRInstruction methodsFor: 'mapping' stamp: 'ajh 3/6/2003 14:32'!
sourceNode: parseNode

	sourceNode _ parseNode
	! !

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

IRInstruction class
	instanceVariableNames: ''!

!IRInstruction class methodsFor: 'instance creation' stamp: 'ajh 3/10/2003 01:07'!
goto: seq

	^ IRJump new
		destination: seq! !

!IRInstruction class methodsFor: 'instance creation' stamp: 'ajh 3/10/2003 01:07'!
if: bool goto: seq1 otherwise: seq2

	^ IRJumpIf new
		boolean: bool;
		destination: seq1;
		otherwise: seq2! !

!IRInstruction class methodsFor: 'instance creation' stamp: 'ajh 3/10/2003 01:07'!
popTop

	^ IRPop new! !

!IRInstruction class methodsFor: 'instance creation' stamp: 'ajh 6/27/2004 22:58'!
pushBlock: irMethod

	^ IRConstant new
		constant: irMethod;
		type: #block! !

!IRInstruction class methodsFor: 'instance creation' stamp: 'ajh 6/27/2004 21:40'!
pushBlockMethod: irMethod

	^ IRConstant new
		constant: irMethod;
		type: #blockMethod! !

!IRInstruction class methodsFor: 'instance creation' stamp: 'ajh 3/10/2003 01:08'!
pushDup

	^ IRDup new! !

!IRInstruction class methodsFor: 'instance creation' stamp: 'ajh 3/10/2003 01:08'!
pushLiteral: object

	^ IRConstant new
		constant: object! !

!IRInstruction class methodsFor: 'instance creation' stamp: 'ajh 3/12/2003 12:16'!
pushTemp: index

	^ IRTemp new
		number: index;
		isStore: false! !

!IRInstruction class methodsFor: 'instance creation' stamp: 'ajh 3/10/2003 01:08'!
remoteReturn

	^ IRReturn new
		isRemote: true! !

!IRInstruction class methodsFor: 'instance creation' stamp: 'ajh 3/10/2003 01:08'!
returnTop

	^ IRReturn new
		isRemote: false! !

!IRInstruction class methodsFor: 'instance creation' stamp: 'ajh 3/10/2003 01:08'!
send: selector

	^ IRSend new
		selector: selector! !

!IRInstruction class methodsFor: 'instance creation' stamp: 'ajh 7/3/2004 19:10'!
send: selector toSuperOf: behavior

	behavior ifNil: [self error: 'super of nil does not exist'].
	^ IRSend new
		selector: selector;
		superOf: behavior! !

!IRInstruction class methodsFor: 'instance creation' stamp: 'ajh 3/12/2003 12:16'!
storeTemp: index

	^ IRTemp new
		number: index;
		isStore: true! !


IRInstruction subclass: #IRConstant
	instanceVariableNames: 'constant type'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-IR'!
!IRConstant commentStamp: 'ajh 3/24/2003 23:56' prior: 0!
Instruction "pushLiteral: object"!


!IRConstant methodsFor: 'as yet unclassified' stamp: 'ajh 3/10/2003 00:43'!
constant

	^ constant! !

!IRConstant methodsFor: 'as yet unclassified' stamp: 'ajh 3/10/2003 00:43'!
constant: object

	constant _ object! !

!IRConstant methodsFor: 'as yet unclassified' stamp: 'ajh 6/27/2004 21:41'!
executeOn: interpreter

	^ type caseOf: {
		[nil] -> [interpreter pushLiteral: constant].
		[#block] -> [interpreter pushBlock: constant].
		[#blockMethod] -> [interpreter pushBlockMethod: constant]
	 }! !

!IRConstant methodsFor: 'as yet unclassified' stamp: 'ajh 3/10/2003 16:12'!
isConstant: valueTest

	^ valueTest value: constant! !

!IRConstant methodsFor: 'as yet unclassified' stamp: 'ajh 6/27/2004 21:39'!
type
	"type is nil, #block, or #blockMethod"

	^ type! !

!IRConstant methodsFor: 'as yet unclassified' stamp: 'ajh 6/27/2004 21:39'!
type: symbol
	"symbol is nil, #block, or #blockMethod"

	type _ symbol! !


IRInstruction subclass: #IRDup
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-IR'!
!IRDup commentStamp: 'ajh 3/24/2003 23:56' prior: 0!
Instruction "pushDup"!


!IRDup methodsFor: 'as yet unclassified' stamp: 'ajh 3/10/2003 00:46'!
executeOn: interpreter

	^ interpreter pushDup! !


Object subclass: #IRInterpreter
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-IR'!
!IRInterpreter commentStamp: 'ajh 3/24/2003 23:55' prior: 0!
I visit each IRInstruction in an IRMethod in order.  Each instruction sends its instruction message to me upon being visited.  See my 'instructions' method category for complete list of instructions.  Subclasses should override them.!


!IRInterpreter methodsFor: 'instructions' stamp: 'ajh 3/9/2003 13:26'!
goto: seqNum! !

!IRInterpreter methodsFor: 'instructions' stamp: 'ajh 3/9/2003 13:26'!
if: bool goto: seqNum1 otherwise: seqNum2! !

!IRInterpreter methodsFor: 'instructions' stamp: 'ajh 3/9/2003 13:25'!
label: seqNum! !

!IRInterpreter methodsFor: 'instructions' stamp: 'ajh 3/9/2003 14:11'!
popTop! !

!IRInterpreter methodsFor: 'instructions' stamp: 'ajh 6/27/2004 21:34'!
pushBlock: irMethod! !

!IRInterpreter methodsFor: 'instructions' stamp: 'ajh 6/27/2004 21:26'!
pushBlockMethod: irMethod! !

!IRInterpreter methodsFor: 'instructions' stamp: 'ajh 3/9/2003 14:12'!
pushDup! !

!IRInterpreter methodsFor: 'instructions' stamp: 'ajh 3/2/2003 16:08'!
pushLiteral: object! !

!IRInterpreter methodsFor: 'instructions' stamp: 'ajh 3/2/2003 16:09'!
pushTemp: index! !

!IRInterpreter methodsFor: 'instructions' stamp: 'ajh 3/2/2003 16:10'!
remoteReturn! !

!IRInterpreter methodsFor: 'instructions' stamp: 'ajh 3/2/2003 16:08'!
returnTop! !

!IRInterpreter methodsFor: 'instructions' stamp: 'ajh 3/2/2003 16:11'!
send: selector! !

!IRInterpreter methodsFor: 'instructions' stamp: 'ajh 3/2/2003 16:11'!
send: selector toSuperOf: behavior! !

!IRInterpreter methodsFor: 'instructions' stamp: 'ajh 3/2/2003 16:11'!
storeTemp: index! !


!IRInterpreter methodsFor: 'interpret' stamp: 'ajh 3/9/2003 12:33'!
interpret: ir

	self interpretAll: ir allSequences! !

!IRInterpreter methodsFor: 'interpret' stamp: 'ajh 3/10/2003 23:30'!
interpretAll: irSequences

	irSequences do: [:seq | self interpretSequence: seq]! !

!IRInterpreter methodsFor: 'interpret' stamp: 'ajh 3/6/2003 15:31'!
interpretInstruction: irInstruction

	irInstruction executeOn: self! !

!IRInterpreter methodsFor: 'interpret' stamp: 'ajh 3/9/2003 13:20'!
interpretSequence: instructionSequence

	self label: instructionSequence orderNumber.
	instructionSequence do: [:instr | self interpretInstruction: instr].
! !


IRInterpreter subclass: #IRDecompiler
	instanceVariableNames: 'stack sp scope currentInstr valueLabelMap'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-IR'!
!IRDecompiler commentStamp: 'ajh 3/25/2003 00:37' prior: 0!
I interpret IRMethod instructions and generate a Smalltalk abstract syntax tree rooted at a RBMethodNode.

This is implemented like a shift-reduce parser.  Each instruction either causes a node to be pushed on the stack (shift), or causes one or more nodes to be popped and combined into a single node which is push back on the stack (reduce).  Most reduction is done at the "label: labelNum" instruction where it tries to reduce jump structures into control messages like #ifTrue:, whileFalse:, etc.

Several pseudo nodes (RBPseudoNode and subclasses) are used to represent basic instructions that have not been reduced to real AST nodes yet.
!


!IRDecompiler methodsFor: 'init' stamp: 'ajh 6/25/2004 11:15'!
decompileIR: ir

	| sequenceNode temps args goto seq value |
	1 to: ir numArgs do: [:i | scope tempVarAt: i].
	self interpret: ir.
	self label: #return.  "final label to return"

	self Label: #return.
	goto _ self Goto.
	value _ self ValueOrNone.
	seq _ self Sequence.
	sp = 1 ifFalse: [self patternError].

	value ifNotNil: [seq addNode: value].
	sequenceNode _ (self newBlock: seq return: goto) body.

	temps _ scope tempVars asArray.
	ir tempKeys: temps.
	args _ (temps first: ir numRargs) allButFirst.
	args _ args collect: [:var | self newVar: var].
	temps _ temps allButFirst: ir numRargs.
	sequenceNode temporaries: (temps collect: [:var | self newVar: var]).

	^ RBMethodNode new
		selectorParts: (self newSelectorParts: (self class dummySelector: args size));
		arguments: args;
		body: sequenceNode;
		privIR: ir;
		primitiveNode: ir primitiveNode;
		scope: scope! !

!IRDecompiler methodsFor: 'init' stamp: 'ajh 6/27/2004 15:14'!
scope: aLexicalScope

	scope _ aLexicalScope! !


!IRDecompiler methodsFor: 'instructions' stamp: 'ajh 3/23/2003 12:47'!
goto: seqNum

	self stackPush: (RBPseudoGotoNode new destination: seqNum).
! !

!IRDecompiler methodsFor: 'instructions' stamp: 'ajh 3/20/2003 17:21'!
if: bool goto: seqNum1 otherwise: seqNum2

	self stackPush: (RBPseudoIfNode new
		boolean: bool;
		destination: seqNum1;
		otherwise: seqNum2)! !

!IRDecompiler methodsFor: 'instructions' stamp: 'ajh 6/22/2003 14:46'!
label: seqNum

	stack isEmpty ifTrue: [  "start"
		^ stack addLast: (RBPseudoLabelNode new destination: seqNum)].

	"Reduce jump structures to one of the following if possible"
	[	(self endIfNil: seqNum) or: [
		 (self endAndOr: seqNum) or: [
		  (self endAndOr2: seqNum) or: [
		   (self endIfThen: seqNum) or: [
		    (self endIfThenElse: seqNum) or: [
		     (self endCase: seqNum) or: [
		      (self endToDo: seqNum) or: [
		       (self endWhile: seqNum)]]]]]]]
	] whileTrue.

	stack addLast: (RBPseudoLabelNode new destination: seqNum).
! !

!IRDecompiler methodsFor: 'instructions' stamp: 'ajh 6/22/2003 21:05'!
popTop

	| value |
	stack last ifNil: [^ stack removeLast].  "pop no-op from #simplifyTempAssign:"
	stack last isValue ifTrue: [
		(stack atLast: 2) isSequence ifTrue: [
			value _ stack removeLast.
			^ stack last addNode: value.
		] ifFalse: [(stack atLast: 2) isPseudo ifTrue: [
			value _ stack removeLast.
			^ stack addLast: (RBSequenceNode statements: {value}).
		]].
	].
	stack addLast: RBPseudoPopNode new
! !

!IRDecompiler methodsFor: 'instructions' stamp: 'ajh 6/27/2004 21:35'!
pushBlock: irMethod

	self block: irMethod env: nil! !

!IRDecompiler methodsFor: 'instructions' stamp: 'ajh 6/27/2004 21:37'!
pushBlockMethod: irMethod

	"block will recognized when send: #createBlock:"
	self pushLiteral: irMethod! !

!IRDecompiler methodsFor: 'instructions' stamp: 'ajh 6/22/2003 14:51'!
pushDup

	stack addLast: RBPseudoDupNode new! !

!IRDecompiler methodsFor: 'instructions' stamp: 'ajh 6/27/2004 21:36'!
pushLiteral: object

	self stackPush: (RBLiteralNode
		literalToken: (SmaCCToken value: object name)
		value: object).
! !

!IRDecompiler methodsFor: 'instructions' stamp: 'ajh 3/17/2003 10:15'!
pushTemp: tempIndex

	| var |
	var _ scope tempVarAt: tempIndex.
	self stackPush: (self newVar: var)! !

!IRDecompiler methodsFor: 'instructions' stamp: 'ajh 6/22/2003 14:27'!
remoteReturn

	stack removeLast.  "pop home context free var"
	self goto: #return.
! !

!IRDecompiler methodsFor: 'instructions' stamp: 'ajh 3/20/2003 17:26'!
returnTop

	self goto: #return.
! !

!IRDecompiler methodsFor: 'instructions' stamp: 'ajh 6/27/2004 16:12'!
send: selector

	| args rcvr |
	args _ OrderedCollection new.
	[	selector numArgs timesRepeat: [args addFirst: self Value].
		rcvr _ self Value.
	] on: Abort do: [
		self stackPush: (RBPseudoSendNode new selector: selector).
		^ self cascade
	].

	(rcvr isLiteral and: [selector = #createBlock:]) ifTrue: [
		^ self block: rcvr value env: args first].

	self stackPush: (self simplify: (RBMessageNode new
		receiver: rcvr
		selectorParts: (self newSelectorParts: selector)
		arguments: args)).
! !

!IRDecompiler methodsFor: 'instructions' stamp: 'ajh 3/24/2003 00:45'!
send: selector toSuperOf: behavior

	| args rcvr |
	args _ OrderedCollection new.
	selector numArgs timesRepeat: [args addFirst: self Value].
	rcvr _ self Value.
	(rcvr isVariable and: [rcvr name = 'self']) ifFalse: [self patternError].

	rcvr identifierToken: (SmaCCToken value: 'super').
	self stackPush: (RBMessageNode new
		receiver: rcvr
		selectorParts: (self newSelectorParts: selector)
		arguments: args).
! !

!IRDecompiler methodsFor: 'instructions' stamp: 'ajh 3/24/2003 15:55'!
storeTemp: tempIndex

	| var |
	var _ scope tempVarAt: tempIndex.
	self stackPush: (self simplifyTempAssign:
		(RBAssignmentNode variable: (self newVar: var) value: self Value)).
! !


!IRDecompiler methodsFor: 'priv instructions' stamp: 'ajh 3/21/2003 00:47'!
block: method capturedVars: varNodes

	| vars |
	vars _ varNodes collect: [:v |
		v isVariable ifTrue: [v binding]
		  ifFalse: [
			(v isMessage and: [v selector = #contextTag and: [v receiver isVariable]])
				ifFalse: [self patternError].
			v receiver binding]
	].
	self stackPush: (IRDecompiler new
		scope: (scope newBlockScope capturedVars: vars);
		decompileIR: method ir)
		asBlock! !

!IRDecompiler methodsFor: 'priv instructions' stamp: 'ajh 6/29/2004 14:33'!
block: method env: envRefNode

	self halt.
	envRefNode isVariable ifTrue: [envRefNode binding]
	  ifFalse: [
		(envRefNode isMessage and: [envRefNode selector = #myEnv and: [envRefNode receiver isVariable]])
			ifFalse: [self patternError].
		envRefNode receiver binding].

	self stackPush: (IRDecompiler new
		scope: (scope newBlockScope "capturedVars: vars");
		decompileIR: method ir)
		asBlock! !

!IRDecompiler methodsFor: 'priv instructions' stamp: 'ajh 3/24/2003 01:54'!
cascade

	| messages selector args rcvr |
	messages _ OrderedCollection new.
	[	"last message"
		selector _ self Send selector.
		args _ OrderedCollection new.
		selector numArgs timesRepeat: [args addFirst: self Value].
		messages addFirst: selector -> args.

		"rest of messages"
		[(rcvr _ self ValueOrNone) isNil] whileTrue: [
			self Pop.
			selector _ self Send selector.
			args _ OrderedCollection new.
			selector numArgs timesRepeat: [args addFirst: self Value].
			self Dup.
			messages addFirst: selector -> args.
		].
	] on: Abort do: [^ false].

	messages _ messages collect: [:assoc |
		RBMessageNode
			receiver: rcvr
			selector: assoc key
			arguments: assoc value].
	self stackPush: (RBCascadeNode messages: messages).
! !

!IRDecompiler methodsFor: 'priv instructions' stamp: 'ajh 6/22/2003 14:46'!
endAndOr2: seqNum

	| goto seq p if2 test else o if1 seqValue elseTest |
	[
		goto _ self Goto.
		seqValue _ self ValueOrNone.
		seq _ self Sequence.
		p _ self Label destination.
		if2 _ self IfGoto: seqNum otherwise: p.
		elseTest _ self Value.
		else _ self SequenceBackTo: goto destination.
		o _ self Label destination.
		o = goto destination ifTrue: [self abort].
		if1 _ self IfGoto: seqNum otherwise: o.
		test _ self Value.
	] on: Abort do: [^ false].

	if1 boolean = if2 boolean ifFalse: [
		test _ RBMessageNode
			receiver: test selector: #not arguments: #().
	].
	self stackPush: (RBMessageNode
		receiver: test
		selector: (if2 boolean ifTrue: [#or:] ifFalse: [#and:])
		arguments: {self newBlock: (else addNode: elseTest) return: nil}).
	stack addLast: if2.
	self label: p.
	stack addLast: seq.
	seqValue ifNotNil: [stack addLast: seqValue].
	stack addLast: goto.
	^ true
! !

!IRDecompiler methodsFor: 'priv instructions' stamp: 'ajh 6/22/2003 14:46'!
endAndOr: seqNum

	| o test branches if body block sel1 sel2 if2 |
	branches _ OrderedCollection new.
	[
		(if2 _ self If) otherwise = seqNum ifFalse: [self abort].
		[	test _ self Value.
			body _ self Sequence.
			branches add: {body. test}.
			o _ self Label destination.
			(if _ self If) otherwise = o ifFalse: [self abort].
			if destination = seqNum
		] whileFalse: [
			if boolean = if2 boolean ifFalse: [self abort].
			if destination = if2 destination ifFalse: [self abort].
		].
		if boolean = if2 boolean ifTrue: [self abort].
		test _ self Value.
	] on: Abort do: [^ false].

	if boolean
		ifTrue: [sel1 _ #or:. sel2 _ #and:]
		ifFalse: [sel1 _ #and:. sel2 _ #or:].
	block _ self newBlock: (branches first first addNode: branches first second).
	branches allButFirstDo: [:pair |
		block _ self newBlock: (pair first addNode: (RBMessageNode
				receiver: pair second
				selector: sel2
				arguments: {block})).
	].
	self stackPush: (RBMessageNode
		receiver: test
		selector: sel1
		arguments: {block}).
	stack addLast: if2.
	^ true! !

!IRDecompiler methodsFor: 'priv instructions' stamp: 'ajh 3/24/2003 00:07'!
endCase: seqNum

	| otherwiseGoto goto node otherwiseValue otherwiseSeq n branchValue branchSeq f caseValue caseSeq rcvr branches message |
	branches _ OrderedCollection new.
	[	"otherwise"
		otherwiseGoto _ self GotoOrReturn: seqNum.
		node _ self stackDown.
		(node isPop or: [node isPseudoSend]) ifTrue: [
			node isPop ifTrue: [node _ self Send].
			node selector == #caseError ifFalse: [self abort].
		] ifFalse: [
			sp _ sp + 1.  "stackUp"
			otherwiseValue _ self ValueOrNone.
			otherwiseSeq _ self Sequence.
		].
		n _ self Label destination.

		"last case branch"
		goto _ self GotoOrReturn: seqNum.
		branchValue _ self ValueOrNone.
		branchSeq _ self Sequence.
		f _ self Label destination.
		"last case"
		self IfGoto: n otherwise: f.
		self Send selector == #= ifFalse: [self abort].
		caseValue _ self Value.
		caseSeq _ self Sequence.
		otherwiseSeq ifNil: [self Dup].
		branches addFirst: ({caseSeq. caseValue} -> {branchSeq. branchValue. goto}).

		[(rcvr _ self ValueOrNone) isNil] whileTrue: [
			"case branch"
			n _ self Label destination.
			goto _ self GotoOrReturn: seqNum.
			branchValue _ self ValueOrNone.
			branchSeq _ self Sequence.
			self Pop.
			f _ self Label destination.
			"case"
			self IfGoto: n otherwise: f.
			self Send selector == #= ifFalse: [self abort].
			caseValue _ self Value.
			caseSeq _ self Sequence.
			self Dup.
			branches addFirst: ({caseSeq. caseValue} -> {branchSeq. branchValue. goto}).
		].
	] on: Abort do: [^ false].

	branches _ branches collect: [:assoc |
		assoc key second
			ifNotNil: [assoc key first addNode: assoc key second].
		assoc value second
			ifNotNil: [assoc value first addNode: assoc value second].
		RBMessageNode
			receiver: (self newBlock: assoc key first return: nil)
			selector: #->
			arguments:
				{self newBlock: assoc value first return: assoc value third}
	].
	message _ otherwiseSeq
		ifNil: [
			RBMessageNode
				receiver: rcvr
				selector: #caseOf:
				arguments: {RBArrayNode statements: branches}]
		ifNotNil: [
			otherwiseValue
				ifNotNil: [otherwiseSeq addNode: otherwiseValue].
			RBMessageNode
				receiver: rcvr
				selector: #caseOf:otherwise:
				arguments: 
					{RBArrayNode statements: branches.
					self newBlock: otherwiseSeq return: otherwiseGoto}.
		].
	self stackPush: message.
	branchValue ifNil: [self popTop].
	self goto: seqNum.
	^ true! !

!IRDecompiler methodsFor: 'priv instructions' stamp: 'ajh 3/24/2003 02:39'!
endIfNil: seqNum

	| goto branch o if rcvr value |
	[
		goto _ self GotoOrReturn: seqNum.
		value _ self Value.
		branch _ self Sequence.
		self Pop.
		o _ self Label destination.
		if _ self IfGoto: seqNum otherwise: o.
		self Send selector == #== ifFalse: [self abort].
		(self Value isLiteral: [:v | v isNil]) ifFalse: [self abort].
		self Dup.
		rcvr _ self Value.
	] on: Abort do: [^ false].

	branch addNode: value.
	self stackPush: (RBMessageNode
		receiver: rcvr
		selector: (if boolean ifTrue: [#ifNotNil:] ifFalse: [#ifNil:])
		arguments: {self newBlock: branch return: goto}).
	self goto: seqNum.
	^ true! !

!IRDecompiler methodsFor: 'priv instructions' stamp: 'ajh 3/24/2003 01:46'!
endIfThen: seqNum

	| goto branch o if test value |
	[
		goto _ self Goto.
		(goto destination == seqNum or: [self isExplicitReturn: goto]) ifFalse: [self abort].
		goto isRet ifTrue: [value _ self Value].
		branch _ self Sequence.
		o _ self Label destination.
		if _ self IfGoto: seqNum otherwise: o.
		test _ self Value.
	] on: Abort do: [^ false].

	value ifNotNil: [branch addNode: value].
	self stackPush: (self simplify: (RBMessageNode
		receiver: test
		selector: (if boolean ifTrue: [#ifFalse:] ifFalse: [#ifTrue:])
		arguments: {self newBlock: branch return: goto})).
	self popTop.
	self goto: seqNum.
	^ true! !

!IRDecompiler methodsFor: 'priv instructions' stamp: 'ajh 3/24/2003 19:11'!
endIfThenElse: seqNum

	| goto2 else d goto1 then o if test value2 value1 |
	[
		goto2 _ self Goto.
		value2 _ self ValueOrNone.
		else _ self Sequence.
		d _ self Label destination.
		goto1 _ self Goto.
		((self isExplicitReturn: goto2) or: [goto2 destination == goto1 destination]) ifFalse: [self abort].
		value1 _ self ValueOrNone.
		then _ self Sequence.
		o _ self Label destination.
		if _ self IfGoto: d otherwise: o.
		test _ self Value.
	] on: Abort do: [^ false].

	value2 ifNotNil: [else addNode: value2].
	value1 ifNotNil: [then addNode: value1].
	self stackPush: (self simplify: (else statements isEmpty
		ifTrue: [RBMessageNode
			receiver: test
			selector: (if boolean ifTrue: [#ifFalse:] ifFalse: [#ifTrue:])
			arguments: {self newBlock: then return: goto1}]
		ifFalse: [RBMessageNode
			receiver: test
			selector: (if boolean
				ifTrue: [#ifFalse:ifTrue:]
				ifFalse: [#ifTrue:ifFalse:])
			arguments: {
				self newBlock: then return: goto1.
				self newBlock: else return: goto2}])).
	value1 ifNil: [self popTop].
	self goto: goto1 destination.
	(else statements isEmpty and:
	 [stack anySatisfy: [:n | n isIf and: [n destination = d]]]
	) ifTrue: [
		self label: d.
		self goto: goto2 destination.
	].
	^ true! !

!IRDecompiler methodsFor: 'priv instructions' stamp: 'ajh 3/24/2003 02:04'!
endToDo: seqNum

	| start limit incr iter step loopBlock o if test limitExpr init |
	[
		start _ self Goto destination.
		limit _ self Value.
		incr _ self Assignment.
		iter _ incr variable.
		(incr value isMessage and:
		 [incr value selector == #+ and:
		  [incr value receiver isVariable and: 
		   [incr value receiver binding == iter binding]]]
		) ifFalse: [self abort].
		step _ incr value arguments first.
		loopBlock _ self Sequence.
		o _ self Label destination.
		if _ self IfGoto: seqNum otherwise: o.
		test _ self Value.
		(test isMessage and:
		 [(test selector == #<= or: [test selector == #>=]) and:
		  [(valueLabelMap at: test arguments first ifAbsent: [self abort]) destination = start]]
		) ifFalse: [self abort].
		limitExpr _ test arguments first.
		limitExpr isAssignment ifTrue: [
			limitExpr variable binding == limit binding ifFalse: [self abort].
			limitExpr _ limitExpr value.
		].
		init _ test receiver.
		(init isAssignment and: [init variable binding == iter binding])
			ifFalse: [self abort].
	] on: Abort do: [^ false].

	loopBlock _ self newBlock: loopBlock return: nil.
	loopBlock arguments: {iter}.
	self stackPush: ((step isLiteral: [:c | c = 1])
		ifTrue: [RBMessageNode
				receiver: init value
				selector: #to:do:
				arguments: {limitExpr. loopBlock}]
		ifFalse: [RBMessageNode
				receiver: init value
				selector: #to:by:do:
				arguments: {limitExpr. step. loopBlock}]).
	self popTop.
	self goto: seqNum.
	^ true
! !

!IRDecompiler methodsFor: 'priv instructions' stamp: 'ajh 3/24/2003 02:18'!
endWhile: seqNum

	| start loopBlock if test sequence o |
	[
		start _ self Goto destination.
		loopBlock _ self Sequence.
		o _ self Label destination.
		if _ self IfGoto: seqNum otherwise: o.
		test _ self Value.
		sequence _ self SequenceBackTo: start.
		self Label: start.
		sp _ sp + 1.  "stackUp"
	] on: Abort do: [^ false].

	self stackPush: (self simplify: (RBMessageNode
		receiver: (self newBlock: (sequence addNode: test) return: nil)
		selector: (if boolean ifTrue: [#whileFalse:] ifFalse: [#whileTrue:])
		arguments: {self newBlock: loopBlock return: nil})).
	self popTop.
	self goto: seqNum.
	^ true
! !


!IRDecompiler methodsFor: 'interpret' stamp: 'ajh 3/13/2003 13:48'!
interpretInstruction: irInstruction

	currentInstr _ irInstruction.
	super interpretInstruction: irInstruction.
! !

!IRDecompiler methodsFor: 'interpret' stamp: 'ajh 3/20/2003 23:37'!
interpretSequence: instructionSequence

	super interpretSequence: instructionSequence.
	currentInstr _ nil.
! !


!IRDecompiler methodsFor: 'private' stamp: 'ajh 3/23/2003 21:17'!
initialize

	stack _ OrderedCollection new.
	scope _ nil parseScope newMethodScope.  "in case never set"
	valueLabelMap _ IdentityDictionary new.
! !

!IRDecompiler methodsFor: 'private' stamp: 'ajh 3/22/2003 11:03'!
isExplicitReturn: goto

	^ goto isRet and:
	  [goto mapInstr notNil and:
	   [goto mapInstr isRemote or: [scope isBlockScope not]]]! !

!IRDecompiler methodsFor: 'private' stamp: 'ajh 3/20/2003 23:21'!
mapNode: node

	currentInstr ifNil: [^ self].
	node isPseudo
		ifTrue: [node mapInstr: currentInstr]
		ifFalse: [currentInstr sourceNode: node]! !

!IRDecompiler methodsFor: 'private' stamp: 'ajh 3/24/2003 12:07'!
newBlock: sequence

	^ self newBlock: sequence return: nil! !

!IRDecompiler methodsFor: 'private' stamp: 'ajh 3/24/2003 01:24'!
newBlock: sequence return: goto

	| statements ret |
	statements _ sequence statements.
	(goto notNil and: [self isExplicitReturn: goto]) ifTrue: [
		ret _ RBReturnNode value: statements last.
		goto mapInstr sourceNode: ret.
		statements atLast: 1 put: ret.
	].
	sequence statements: statements.
	^ RBBlockNode body: sequence! !

!IRDecompiler methodsFor: 'private' stamp: 'ajh 3/13/2003 14:10'!
newSelectorParts: selector

	^ selector keywords collect: [:word |
		SmaCCToken value: word]! !

!IRDecompiler methodsFor: 'private' stamp: 'ajh 3/17/2003 10:14'!
newVar: semVar

	^ RBVariableNode new
		identifierToken: (SmaCCToken value: semVar name);
		binding: semVar! !

!IRDecompiler methodsFor: 'private' stamp: 'ajh 7/3/2004 15:16'!
simplify: mess
	"mess is a messageNode.  If it is a message created by the compiler convert it back to its normal form"

	| rcvr sc |
	(mess selector == #value and: [mess receiver isLiteral]) ifTrue: [
		^ self newVar: (GlobalVar new assoc: mess receiver value; scope: scope)
	].

	(mess selector = #privSetInHolder: and: [mess arguments first isLiteral]) ifTrue: [
		^ RBAssignmentNode
			variable: (self newVar: (GlobalVar new assoc: mess arguments first value; scope: scope))
			value: mess receiver
	].

	(mess selector = #privGetField: and:
	 [mess arguments first isLiteral and:
	  [mess receiver isVariable]]) ifTrue: [
		rcvr _ mess receiver binding.
		rcvr == scope receiverVar ifTrue: [
			^ self newVar: (scope receiverVarAt: mess arguments first value)].
		(rcvr isFieldVar and: [rcvr originalTempVar notNil]) ifTrue: [
			sc _ rcvr originalTempVar scope.
			rcvr = scope receiverVar ifTrue: [
				^ self newVar: (sc receiverVarAt: mess arguments first value)]].
	].

	(mess selector = #privStoreIn:field: and:
	 [mess arguments last isLiteral and:
	  [mess arguments first isVariable]]) ifTrue: [
		rcvr _ mess arguments first binding.
		rcvr == scope receiverVar ifTrue: [
			^ RBAssignmentNode
				variable: (self newVar: (scope receiverVarAt: mess arguments last value))
				value: mess receiver
		].
		(rcvr isFieldVar and: [rcvr originalTempVar notNil]) ifTrue: [
			sc _ rcvr originalTempVar scope.
			rcvr = scope receiverVar ifTrue: [
				^ RBAssignmentNode
					variable: (self newVar: (sc receiverVarAt: mess arguments last value))
					value: mess receiver]
		].
	].

	(mess selector = #tempHolderValue and: [mess receiver isVariable]) ifTrue: [
		^ mess receiver
	].

	(mess selector = #storeInTempHolder: and: [mess arguments first isVariable]) ifTrue: [
		^ RBAssignmentNode
			variable: mess arguments first
			value: mess receiver
	].

	^ mess! !

!IRDecompiler methodsFor: 'private' stamp: 'ajh 3/24/2003 16:01'!
simplifyTempAssign: assignment
	"If it is a assignment created by the compiler convert it back to its normal form"

	| mess |
	((mess _ assignment value) isMessage and: 
	 [mess selector = #wrapInTempHolder and:
	  [mess receiver isLiteral: [:v | v isNil]]]
	) ifTrue: [
		^ nil  "no-op"
	].

	^ assignment! !


!IRDecompiler methodsFor: 'stack' stamp: 'ajh 3/22/2003 17:03'!
Assignment

	| node |
	(node _ self stackDown) isAssignment ifTrue: [^ node].
	self abort! !

!IRDecompiler methodsFor: 'stack' stamp: 'ajh 3/23/2003 22:34'!
Dup

	| node |
	(node _ self stackDown) isDup ifTrue: [^ node].
	self abort! !

!IRDecompiler methodsFor: 'stack' stamp: 'ajh 3/22/2003 16:33'!
Goto

	| node |
	(node _ self stackDown) isGoto ifTrue: [^ node].
	self abort! !

!IRDecompiler methodsFor: 'stack' stamp: 'ajh 3/23/2003 12:57'!
Goto: seqNum

	| goto |
	(goto _ self Goto) destination = seqNum ifTrue: [^ goto].
	self abort! !

!IRDecompiler methodsFor: 'stack' stamp: 'ajh 3/23/2003 23:17'!
GotoOrReturn: seqNum

	| goto |
	goto _ self Goto.
	(goto destination = seqNum or: [goto isRet]) ifTrue: [^ goto].
	self abort! !

!IRDecompiler methodsFor: 'stack' stamp: 'ajh 3/22/2003 16:33'!
If

	| node |
	(node _ self stackDown) isIf ifTrue: [^ node].
	self abort! !

!IRDecompiler methodsFor: 'stack' stamp: 'ajh 3/22/2003 17:07'!
IfGoto: seqNum otherwise: seqNum2

	| if |
	((if _ self If) destination = seqNum and: [if otherwise = seqNum2])
		ifTrue: [^ if].
	self abort! !

!IRDecompiler methodsFor: 'stack' stamp: 'ajh 3/22/2003 16:33'!
Label

	| node |
	(node _ self stackDown) isLabel ifTrue: [^ node].
	self abort! !

!IRDecompiler methodsFor: 'stack' stamp: 'ajh 3/22/2003 16:34'!
Label: seqNum

	| label |
	(label _ self Label) destination = seqNum ifTrue: [^ label].
	self abort! !

!IRDecompiler methodsFor: 'stack' stamp: 'ajh 3/22/2003 17:36'!
Pop

	| node |
	(node _ self stackDown) isPop ifTrue: [^ node].
	self abort! !

!IRDecompiler methodsFor: 'stack' stamp: 'ajh 3/23/2003 22:32'!
Send

	| node |
	(node _ self stackDown) isPseudoSend ifTrue: [^ node].
	self abort! !

!IRDecompiler methodsFor: 'stack' stamp: 'ajh 3/24/2003 01:08'!
Sequence

	| node seq i goto |
	seq _ RBSequenceNode statements: #().
	i _ sp ifNil: [sp _ stack size].
	[	node _ stack at: i.
		node isSequence ifTrue: [
			seq addNodesFirst: node statements.
			node _ stack at: (i _ i - 1).
		].
		"include sequences before fall-through gotos"
		(node isLabel and: [i > 1 "not first"]) ifFalse: [
			sp _ i.
			^ seq].
		goto _ stack at: (i _ i - 1).
		goto isGoto and: [goto destination = node destination]
	] whileTrue: [i _ i - 1].
	sp _ i + 1.  "points to label"
	^ seq! !

!IRDecompiler methodsFor: 'stack' stamp: 'ajh 3/24/2003 01:08'!
SequenceBackTo: labelNum
	"Return previous sequence stopping at non-sequence or label: num, whichever comes first"

	| node seq i goto |
	seq _ RBSequenceNode statements: #().
	i _ sp ifNil: [sp _ stack size].
	[	node _ stack at: i.
		node isSequence ifTrue: [
			seq addNodesFirst: node statements.
			node _ stack at: (i _ i - 1).
		].
		"includes sequences back to labelNum"
		(node isLabel and: [i > 1 "not first"]) ifFalse: [
			sp _ i.
			^ seq].
		node destination = labelNum ifTrue: [
			sp _ i.
			^ seq].
		goto _ stack at: (i _ i - 1).
		goto isGoto and: [goto destination = node destination]
	] whileTrue: [i _ i - 1].
	sp _ i + 1.  "points to label"
	^ seq! !

!IRDecompiler methodsFor: 'stack' stamp: 'ajh 3/23/2003 21:44'!
Value

	| node |
	node _ self ValueOrNone.
	node ifNil: [self abort].
	^ node! !

!IRDecompiler methodsFor: 'stack' stamp: 'ajh 3/23/2003 21:43'!
ValueOrNone

	| node i label |
	i _ sp ifNil: [sp _ stack size].
	[	node _ stack at: i.
		node isValue ifTrue: [
			label ifNotNil: [valueLabelMap at: node put: label].
			sp _ i - 1.
			^ node].
		"test for value before fall-through goto"
		(node isLabel and: [i > 1 "not first"]) ifFalse: [^ nil].
		label _ node.
		node _ stack at: (i _ i - 1).
		node isGoto and: [node destination = label destination]
	] whileTrue: [i _ i - 1].
	^ nil! !

!IRDecompiler methodsFor: 'stack' stamp: 'ajh 3/24/2003 00:59'!
abort

	| spWas |
	spWas _ sp.
	sp _ nil.
	Abort signal! !

!IRDecompiler methodsFor: 'stack' stamp: 'ajh 6/22/2003 14:41'!
fixStack

	sp ifNotNil: [stack removeLast: (stack size - sp)].
	sp _ nil.
! !

!IRDecompiler methodsFor: 'stack' stamp: 'ajh 3/22/2003 19:24'!
stackDown

	| node |
	sp ifNil: [sp _ stack size].
	sp = 0 ifTrue: [self abort].
	node _ stack at: sp.
	sp _ sp - 1.
	^ node! !

!IRDecompiler methodsFor: 'stack' stamp: 'ajh 6/22/2003 14:51'!
stackPush: node

	self fixStack.
	stack addLast: node.
	node ifNil: [^ self].  "no op"
	self mapNode: node.
! !

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

IRDecompiler class
	instanceVariableNames: ''!

!IRDecompiler class methodsFor: 'as yet unclassified' stamp: 'ajh 6/5/2003 12:34'!
dummySelector: numArgs
	"Answer a dummy selector with number of args"

	| sel |
	sel _ 'unknown'.
	1 to: numArgs do: [:i |
		sel _ sel, 'with:'].
	^ sel asSymbol! !


IRInstruction subclass: #IRJump
	instanceVariableNames: 'destination'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-IR'!
!IRJump commentStamp: 'ajh 3/24/2003 23:56' prior: 0!
Instruction "goto: labelNum"!


!IRJump methodsFor: 'as yet unclassified' stamp: 'ajh 3/10/2003 23:08'!
destination

	^ destination! !

!IRJump methodsFor: 'as yet unclassified' stamp: 'ajh 3/10/2003 23:08'!
destination: sequence

	destination _ sequence! !

!IRJump methodsFor: 'as yet unclassified' stamp: 'ajh 3/10/2003 00:47'!
executeOn: interpreter

	^ interpreter goto: destination orderNumber! !

!IRJump methodsFor: 'as yet unclassified' stamp: 'ajh 3/11/2003 00:19'!
isGoto
	"is unconditional jump"

	^ true! !

!IRJump methodsFor: 'as yet unclassified' stamp: 'ajh 3/11/2003 00:22'!
isJump
	"goto or if"

	^ true! !

!IRJump methodsFor: 'as yet unclassified' stamp: 'ajh 3/11/2003 00:02'!
successorSequences

	^ {destination}! !


IRJump subclass: #IRJumpIf
	instanceVariableNames: 'boolean otherwise'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-IR'!
!IRJumpIf commentStamp: 'ajh 3/24/2003 23:56' prior: 0!
Instruction "if: boolean goto: labelNum1 otherwise: labelNum2"!


!IRJumpIf methodsFor: 'as yet unclassified' stamp: 'ajh 3/10/2003 00:43'!
boolean

	^ boolean! !

!IRJumpIf methodsFor: 'as yet unclassified' stamp: 'ajh 3/10/2003 00:43'!
boolean: bool

	boolean _ bool! !

!IRJumpIf methodsFor: 'as yet unclassified' stamp: 'ajh 3/10/2003 00:47'!
executeOn: interpreter

	^ interpreter if: boolean goto: destination orderNumber otherwise: otherwise orderNumber! !

!IRJumpIf methodsFor: 'as yet unclassified' stamp: 'ajh 3/11/2003 00:19'!
isGoto
	"is unconditional jump"

	^ false! !

!IRJumpIf methodsFor: 'as yet unclassified' stamp: 'ajh 3/11/2003 00:21'!
isIf

	^ true! !

!IRJumpIf methodsFor: 'as yet unclassified' stamp: 'ajh 3/10/2003 00:43'!
otherwise

	^ otherwise! !

!IRJumpIf methodsFor: 'as yet unclassified' stamp: 'ajh 3/10/2003 00:44'!
otherwise: sequence

	otherwise _ sequence! !

!IRJumpIf methodsFor: 'as yet unclassified' stamp: 'ajh 3/11/2003 00:02'!
successorSequences

	^ {destination. otherwise}! !


Object subclass: #IRMethod
	instanceVariableNames: 'startSequence primitiveNode tempKeys numRargs compiledMethod'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-IR'!
!IRMethod commentStamp: 'ajh 5/23/2003 11:08' prior: 0!
I am a method in the IR (intermediate representation) language consisting of IRInstructions grouped by IRSequence (basic block).  The IRSequences form a control graph (therefore I only have to hold onto the starting sequence).  #compiledMethod will convert me to a CompiledMethod.  #methodNode will convert me back to a parse tree.
!


!IRMethod methodsFor: 'initialize' stamp: 'ajh 6/25/2004 11:44'!
initialize

	primitiveNode _ PrimitiveNode null.
	tempKeys _ #().
! !

!IRMethod methodsFor: 'initialize' stamp: 'ajh 6/25/2004 10:53'!
numRargs: n

	numRargs _ n! !

!IRMethod methodsFor: 'initialize' stamp: 'ajh 3/10/2003 23:08'!
primitiveNode: aPrimitiveNode

	primitiveNode _ aPrimitiveNode! !

!IRMethod methodsFor: 'initialize' stamp: 'ajh 3/10/2003 17:52'!
startSequence: irSequence

	startSequence _ irSequence! !

!IRMethod methodsFor: 'initialize' stamp: 'ajh 6/25/2004 11:19'!
tempKeys: objects

	tempKeys _ objects! !


!IRMethod methodsFor: 'accessing' stamp: 'ajh 3/9/2003 15:35'!
allSequences

	^ startSequence withAllSuccessors! !

!IRMethod methodsFor: 'accessing' stamp: 'ajh 6/25/2004 10:53'!
numArgs

	^ self numRargs - 1! !

!IRMethod methodsFor: 'accessing' stamp: 'ajh 6/25/2004 10:53'!
numRargs

	^ numRargs! !

!IRMethod methodsFor: 'accessing' stamp: 'ajh 3/10/2003 18:10'!
primitiveNode

	^ primitiveNode! !

!IRMethod methodsFor: 'accessing' stamp: 'ajh 3/10/2003 17:53'!
startSequence

	^ startSequence! !

!IRMethod methodsFor: 'accessing' stamp: 'ajh 6/25/2004 10:49'!
tempKeys

	^ tempKeys! !

!IRMethod methodsFor: 'accessing' stamp: 'ajh 6/25/2004 10:52'!
tempNames
	"All temp names in context order"

	| varNames |
	varNames _ OrderedCollection new.
	self tempKeys do: [:var | | name |
		name _ var asString.
		"vars are unique but inlined to:do: loop vars may have the same name, so munge the names to make them different"
		[varNames includes: name] whileTrue: [name _ name, 'X'].
		varNames add: name.
	].
	^ varNames asArray! !


!IRMethod methodsFor: 'printing' stamp: 'ajh 3/9/2003 15:53'!
longPrintOn: stream

	IRPrinter new
		indent: 0;
		stream: stream;
		interpret: self! !


!IRMethod methodsFor: 'optimizing' stamp: 'ajh 3/10/2003 15:45'!
absorbConstantConditionalJumps

	startSequence absorbConstantConditionalJumps: IdentitySet new! !

!IRMethod methodsFor: 'optimizing' stamp: 'ajh 3/10/2003 15:45'!
absorbJumpsToSingleInstrs

	startSequence absorbJumpToSingleInstr: IdentitySet new! !

!IRMethod methodsFor: 'optimizing' stamp: 'ajh 3/18/2003 19:05'!
optimize

	self removeEmptyStart.
	self absorbJumpsToSingleInstrs.  "do before next to get in right form"
	self absorbConstantConditionalJumps.
	self absorbJumpsToSingleInstrs.  "do again since new opportunities may have arised after last step" 
! !

!IRMethod methodsFor: 'optimizing' stamp: 'ajh 3/18/2003 19:25'!
removeEmptyStart

	startSequence size = 1 ifTrue: [
		"startSeq is just unconditional jump, forget it"
		startSequence _ startSequence last destination].
! !


!IRMethod methodsFor: 'translating' stamp: 'ajh 3/10/2003 15:55'!
compiledMethod

	^ compiledMethod ifNil: [self compiledMethodWith: #(0)]! !

!IRMethod methodsFor: 'translating' stamp: 'ajh 6/28/2004 14:44'!
compiledMethodWith: trailer

	^ compiledMethod _ IRTranslator new
		trailer: trailer;
		interpret: self;
		compiledMethod! !

!IRMethod methodsFor: 'translating' stamp: 'ajh 3/10/2003 15:54'!
privCompiledMethod: aCompiledMethod

	compiledMethod _ aCompiledMethod! !


!IRMethod methodsFor: 'mapping' stamp: 'ajh 3/19/2003 13:38'!
sourceMap
	"Return a mapping from bytecode pcs to source code ranges"

	| start map |
	"Besides getting start position, make sure bytecodeIndices are filled in"
	start _ self compiledMethod initialPC - 1.  
	map _ OrderedCollection new.
	self allSequences do: [:seq |
		seq do: [:instr |  | node |
			((node _ instr sourceNode) notNil and: 
			 [node debugHighlightStart notNil and:
			  [node debugHighlightStop notNil and:
			   [instr bytecodeIndex notNil]]]) ifTrue: [
				map add:
					instr bytecodeIndex + start
						-> (node debugHighlightStart to: node debugHighlightStop)]
		]
	].
	^ map! !


!IRMethod methodsFor: 'decompiling' stamp: 'ajh 6/28/2004 13:44'!
ast

	^ IRDecompiler new decompileIR: self! !

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

IRMethod class
	instanceVariableNames: ''!

!IRMethod class methodsFor: 'as yet unclassified' stamp: 'ajh 3/10/2003 23:17'!
new

	^ super new initialize! !


IRInstruction subclass: #IRPop
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-IR'!
!IRPop commentStamp: 'ajh 3/24/2003 23:57' prior: 0!
Instruction "popTop"!


!IRPop methodsFor: 'as yet unclassified' stamp: 'ajh 3/10/2003 00:47'!
executeOn: interpreter

	^ interpreter popTop! !


IRInterpreter subclass: #IRPrinter
	instanceVariableNames: 'stream indent'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-IR'!
!IRPrinter commentStamp: 'ajh 3/25/2003 00:22' prior: 0!
I interpret IRMethod instructions and write them out to a print stream.!


!IRPrinter methodsFor: 'initialize' stamp: 'ajh 3/9/2003 15:49'!
indent: tabs

	indent _ tabs! !

!IRPrinter methodsFor: 'initialize' stamp: 'ajh 3/9/2003 15:50'!
stream: stringWriteStream

	stream _ stringWriteStream! !


!IRPrinter methodsFor: 'instructions' stamp: 'ajh 3/9/2003 15:41'!
goto: seqNum

	stream nextPutAll: 'goto: '.
	seqNum printOn: stream.
! !

!IRPrinter methodsFor: 'instructions' stamp: 'ajh 3/9/2003 15:42'!
if: bool goto: seqNum1 otherwise: seqNum2

	stream nextPutAll: 'if: '.
	bool printOn: stream.
	stream nextPutAll: ' goto: '.
	seqNum1 printOn: stream.
	stream nextPutAll: ' else: '.
	seqNum2 printOn: stream.
! !

!IRPrinter methodsFor: 'instructions' stamp: 'ajh 3/11/2003 00:36'!
label: seqNum

	"add tab and cr since this does not get called within interpretInstruction:"
	stream cr.  "extra cr just to space out sequences"
	indent timesRepeat: [stream tab].
	stream nextPutAll: 'label: '.
	seqNum printOn: stream.
	stream cr.
! !

!IRPrinter methodsFor: 'instructions' stamp: 'ajh 3/9/2003 15:42'!
popTop

	stream nextPutAll: 'popTop'! !

!IRPrinter methodsFor: 'instructions' stamp: 'ajh 6/27/2004 21:44'!
pushBlock: irMethod

	stream nextPutAll: 'pushBlock:'.
	IRPrinter new
		indent: indent + 1;
		stream: stream;
		interpret: irMethod removeEmptyStart.
! !

!IRPrinter methodsFor: 'instructions' stamp: 'ajh 6/27/2004 21:44'!
pushBlockMethod: irMethod

	stream nextPutAll: 'pushBlockMethod:'.
	IRPrinter new
		indent: indent + 1;
		stream: stream;
		interpret: irMethod removeEmptyStart.
! !

!IRPrinter methodsFor: 'instructions' stamp: 'ajh 3/9/2003 15:42'!
pushDup

	stream nextPutAll: 'pushDup'! !

!IRPrinter methodsFor: 'instructions' stamp: 'ajh 3/21/2003 01:49'!
pushLiteral: object

	stream nextPutAll: 'pushLiteral: '.
	object isVariableBinding ifTrue: [^ stream nextPutAll: object key].
	object printOn: stream.

	((object isKindOf: BlockClosure) or: [object isKindOf: CompiledMethod])
		ifTrue: [
			IRPrinter new
				indent: indent + 1;
				stream: stream;
				interpret: object method ir removeEmptyStart].
! !

!IRPrinter methodsFor: 'instructions' stamp: 'ajh 6/26/2004 18:37'!
pushTemp: index

	stream nextPutAll: 'pushTemp: '.
	index printOn: stream.
	index = 0 ifTrue: [stream nextPutAll: ' "receiver"'].
	index = -1 ifTrue: [stream nextPutAll: ' "thisEnv"'].
	index = -2 ifTrue: [stream nextPutAll: ' "thisContext"'].
! !

!IRPrinter methodsFor: 'instructions' stamp: 'ajh 3/9/2003 15:43'!
remoteReturn

	stream nextPutAll: 'remoteReturn'.
! !

!IRPrinter methodsFor: 'instructions' stamp: 'ajh 3/9/2003 15:43'!
returnTop

	stream nextPutAll: 'returnTop'.
! !

!IRPrinter methodsFor: 'instructions' stamp: 'ajh 3/9/2003 15:44'!
send: selector

	stream nextPutAll: 'send: '.
	selector printOn: stream.
! !

!IRPrinter methodsFor: 'instructions' stamp: 'ajh 3/9/2003 15:44'!
send: selector toSuperOf: behavior

	stream nextPutAll: 'send: '.
	selector printOn: stream.
	stream nextPutAll: ' toSuperOf: '.
	behavior printOn: stream.
! !

!IRPrinter methodsFor: 'instructions' stamp: 'ajh 6/26/2004 23:15'!
storeTemp: index

	stream nextPutAll: 'storeTemp: '.
	index printOn: stream.
	index = -1 ifTrue: [stream nextPutAll: ' "thisEnv"'].
! !


!IRPrinter methodsFor: 'interpret' stamp: 'ajh 3/9/2003 15:48'!
interpretInstruction: irInstruction

	indent timesRepeat: [stream tab].
	super interpretInstruction: irInstruction.
	stream cr.
! !


IRInstruction subclass: #IRReturn
	instanceVariableNames: 'isRemote'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-IR'!
!IRReturn commentStamp: 'ajh 3/24/2003 23:57' prior: 0!
Instruction "returnTop" or "remoteReturn"!


!IRReturn methodsFor: 'as yet unclassified' stamp: 'ajh 3/10/2003 00:48'!
executeOn: interpreter

	^ isRemote
		ifTrue: [interpreter remoteReturn]
		ifFalse: [interpreter returnTop]! !

!IRReturn methodsFor: 'as yet unclassified' stamp: 'ajh 3/10/2003 00:44'!
isRemote

	^ isRemote! !

!IRReturn methodsFor: 'as yet unclassified' stamp: 'ajh 3/10/2003 00:44'!
isRemote: boolean

	isRemote _ boolean! !

!IRReturn methodsFor: 'as yet unclassified' stamp: 'ajh 3/10/2003 16:10'!
isReturn

	^ true! !


IRInstruction subclass: #IRSend
	instanceVariableNames: 'selector superOf'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-IR'!
!IRSend commentStamp: 'ajh 3/24/2003 23:57' prior: 0!
Instruction "send: selector" or "send: selector toSuperOf: behavior"!


!IRSend methodsFor: 'as yet unclassified' stamp: 'ajh 3/10/2003 00:48'!
executeOn: interpreter

	^ superOf
		ifNil: [interpreter send: selector]
		ifNotNil: [interpreter send: selector toSuperOf: superOf]! !

!IRSend methodsFor: 'as yet unclassified' stamp: 'ajh 3/10/2003 00:44'!
selector

	^ selector! !

!IRSend methodsFor: 'as yet unclassified' stamp: 'ajh 3/10/2003 00:44'!
selector: symbol

	selector _ symbol! !

!IRSend methodsFor: 'as yet unclassified' stamp: 'ajh 3/10/2003 00:45'!
superOf

	^ superOf! !

!IRSend methodsFor: 'as yet unclassified' stamp: 'ajh 3/10/2003 00:45'!
superOf: behavior

	superOf _ behavior! !


LinkedList subclass: #IRSequence
	instanceVariableNames: 'orderNumber'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-IR'!
!IRSequence commentStamp: 'ajh 3/14/2003 00:33' prior: 0!
I hold a sequence of IRInstructions where only the last instruction jumps or returns, the rest are guaranteed to execute in order (ie. a basic block).!


!IRSequence methodsFor: 'comparing' stamp: 'ajh 3/11/2003 00:29'!
= other
	"Override collection equal with identity equal"

	^ self == other! !

!IRSequence methodsFor: 'comparing' stamp: 'ajh 6/18/2002 15:09'!
hash

	^ self identityHash! !


!IRSequence methodsFor: 'successor sequences' stamp: 'ajh 3/19/2003 22:22'!
nextSequence

	| sequences i |
	sequences _ self withAllSuccessors.
	i _ sequences findFirst: [:seq | seq orderNumber = self orderNumber].
	(i = 0 or: [i = sequences size]) ifTrue: [^ nil].
	^ sequences at: i + 1! !

!IRSequence methodsFor: 'successor sequences' stamp: 'ajh 3/19/2003 01:24'!
successorSequences

	self isEmpty ifTrue: [^ #()].
	^ self last successorSequences! !

!IRSequence methodsFor: 'successor sequences' stamp: 'ajh 9/26/2002 16:14'!
withAllSuccessors
	"Return me and all my successors sorted by sequence orderNumber"

	| list |
	list _ OrderedCollection new: 20.
	self withAllSuccessorsDo: [:seq | list add: seq].
	^ list asSortedCollection: [:x :y | x orderNumber <= y orderNumber]! !

!IRSequence methodsFor: 'successor sequences' stamp: 'ajh 7/18/2002 01:37'!
withAllSuccessorsDo: block
	"Iterate over me and all my successors only once"

	self withAllSuccessorsDo: block alreadySeen: IdentitySet new! !

!IRSequence methodsFor: 'successor sequences' stamp: 'ajh 3/6/2003 01:31'!
withAllSuccessorsDo: block alreadySeen: set
	"Iterate over me and all my successors only once"

	(set includes: self) ifTrue: [^ self].
	set add: self.
	block value: self.
	self successorSequences do: [:seq |
		seq ifNotNil: [seq withAllSuccessorsDo: block alreadySeen: set]].
! !


!IRSequence methodsFor: 'optimizing' stamp: 'ajh 3/13/2003 00:26'!
absorbConstantConditionalJumps: alreadySeen
	"Collapse sequences that look like:
			[if] goto s1
			...
		 s1:	pushConst: true/false
			goto s2
		 s2:	if true/false goto s3 else s4
	into:
			[if] goto s3/s4

	These sequences are produced by and:/or: messages"

	| seq bool if |
	(alreadySeen includes: self) ifTrue: [^ self].
	alreadySeen add: self.

	[(seq _ self successorSequences) size > 0  "not return"
	  and: [(seq _ seq first "destination") size = 2
	   and: [(seq first isConstant: [:obj | (bool _ obj) isKindOf: Boolean])
	    and: [seq last isGoto
	     and: [(if _ seq last destination first) isIf]]]]
	] whileTrue: [ "absorb"
		self last destination: (bool == if boolean
			ifTrue: [if destination]
			ifFalse: [if otherwise]).
	].

	self successorSequences do: [:instrs | instrs absorbConstantConditionalJumps: alreadySeen].
! !

!IRSequence methodsFor: 'optimizing' stamp: 'ajh 3/20/2003 00:10'!
absorbJumpToSingleInstr: alreadySeen
	"Collapse jumps to single return instructions into caller"

	| seqs seq |
	(alreadySeen includes: self) ifTrue: [^ self].
	alreadySeen add: self.

	[ (seqs _ self successorSequences) size = 1  "unconditional jump..." 
	   and: [(seq _ seqs first) size = 1  "...to single instruction..."
	    and: [seq successorSequences size < 2]]  "...but don't collapse conditional jumps so their otherwiseSequences can stay right after them"
	] whileTrue: [ "replace goto with single instruction"
		self removeLast.
		seq do: [:instr | self add: instr copy].
	].

	seqs do: [:instrs | instrs absorbJumpToSingleInstr: alreadySeen].
! !


!IRSequence methodsFor: 'printing' stamp: 'ajh 3/11/2003 00:57'!
longPrintOn: stream

	[IRPrinter new
		indent: 0;
		stream: stream;
		interpretSequence: self
	] onDNU: #orderNumber do: [:ex | ex resume: ex receiver]! !

!IRSequence methodsFor: 'printing' stamp: 'ajh 3/11/2003 00:25'!
printOn: stream

	stream nextPutAll: 'an '.
	self class printOn: stream.
	stream space.
	stream nextPut: $(.
	self orderNumber printOn: stream.
	stream nextPut: $).
! !


!IRSequence methodsFor: 'manipulating' stamp: 'ajh 3/19/2003 22:25'!
splitAfter: instruction

	| newSeq instr next |
	next _ self nextSequence.
	next _ next
		ifNil: [self orderNumber + 1]
		ifNotNil: [(next orderNumber + self orderNumber) / 2].
	newSeq _ IRSequence new orderNumber: next.
	instr _ instruction.
	[(instr _ instr nextLink) isNil] whileFalse: [newSeq add: instr].
	instruction nextLink: nil.
	lastLink _ instruction.
	self add: (IRJump new destination: newSeq).
	^ newSeq! !


!IRSequence methodsFor: 'initializing' stamp: 'ajh 7/18/2002 01:23'!
orderNumber: n
	"Sequences are sorted by this number"

	orderNumber _ n! !


!IRSequence methodsFor: 'accessing' stamp: 'ajh 7/18/2002 01:23'!
orderNumber
	"Sequences are sorted by this number"

	^ orderNumber! !


IRInstruction subclass: #IRTemp
	instanceVariableNames: 'number isStore'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-IR'!
!IRTemp commentStamp: 'ajh 3/24/2003 23:58' prior: 0!
Instruction "pushTemp: tempIndex" or "storeTemp: tempIndex"!


!IRTemp methodsFor: 'as yet unclassified' stamp: 'ajh 3/12/2003 12:18'!
executeOn: interpreter

	^ isStore
		ifTrue: [interpreter storeTemp: number]
		ifFalse: [interpreter pushTemp: number]! !

!IRTemp methodsFor: 'as yet unclassified' stamp: 'ajh 3/10/2003 00:45'!
isStore

	^ isStore! !

!IRTemp methodsFor: 'as yet unclassified' stamp: 'ajh 3/10/2003 00:45'!
isStore: boolean

	isStore _ boolean! !

!IRTemp methodsFor: 'as yet unclassified' stamp: 'ajh 3/12/2003 12:17'!
number

	^ number! !

!IRTemp methodsFor: 'as yet unclassified' stamp: 'ajh 3/12/2003 12:16'!
number: n

	number _ n! !


IRInterpreter subclass: #IRTranslator
	instanceVariableNames: 'pending gen currentInstr trailerBytes'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-IR'!
!IRTranslator commentStamp: 'ajh 3/25/2003 00:26' prior: 0!
I interpret IRMethod instructions, sending the appropriate bytecode messages to my BytecodeGenerator (gen).  I hold some messages back in pending awaiting certain sequences of them that can be consolidated into single bytecode instructions, otherwise the pending messages are executed in order as if they were executed when they first appeared.!


!IRTranslator methodsFor: 'initialize' stamp: 'ajh 6/28/2004 11:27'!
initialize

	gen _ BytecodeGenerator new.
	trailerBytes _ #(0).
! !

!IRTranslator methodsFor: 'initialize' stamp: 'ajh 6/28/2004 11:23'!
trailer: bytes

	trailerBytes _ bytes! !


!IRTranslator methodsFor: 'instructions' stamp: 'ajh 3/9/2003 22:09'!
goto: seqNum

	self doPending.
	gen goto: seqNum.
! !

!IRTranslator methodsFor: 'instructions' stamp: 'ajh 3/9/2003 22:09'!
if: bool goto: seqNum1 otherwise: seqNum2

	self doPending.
	gen if: bool goto: seqNum1 otherwise: seqNum2.
! !

!IRTranslator methodsFor: 'instructions' stamp: 'ajh 3/9/2003 22:10'!
label: seqNum

	pending _ OrderedCollection new.
	gen label: seqNum.
! !

!IRTranslator methodsFor: 'instructions' stamp: 'ajh 3/9/2003 22:35'!
popTop

	"if last was storeTemp or storeInstVar then convert to storePopTemp or storePopInstVar"
	#storeTemp: == self pendingSelector ifTrue: [
		^ self pendingSelector: #storePopTemp:].
	#storeInstVar: == self pendingSelector ifTrue: [
		^ self pendingSelector: #storePopInstVar:].

	"otherwise do normal pop"
	self doPending.
	gen popTop.
! !

!IRTranslator methodsFor: 'instructions' stamp: 'ajh 6/28/2004 11:20'!
pushBlock: irMethod

	| meth block |
	meth _ irMethod compiledMethodWith: trailerBytes.
	meth isBlockMethod: true.
	block _ meth createBlock: nil.
	self addPending: (Message selector: #pushLiteral: argument: block)! !

!IRTranslator methodsFor: 'instructions' stamp: 'ajh 6/28/2004 11:20'!
pushBlockMethod: irMethod

	| meth |
	meth _ irMethod compiledMethodWith: trailerBytes.
	meth isBlockMethod: true.
	self addPending: (Message selector: #pushLiteral: argument: meth)! !

!IRTranslator methodsFor: 'instructions' stamp: 'ajh 3/9/2003 22:10'!
pushDup

	self doPending.
	gen pushDup.
! !

!IRTranslator methodsFor: 'instructions' stamp: 'ajh 3/12/2003 11:58'!
pushLiteral: object

	self addPending: (Message selector: #pushLiteral: argument: object)! !

!IRTranslator methodsFor: 'instructions' stamp: 'ajh 7/3/2004 15:17'!
pushTemp: index

	index = 0 ifTrue: [^ self addPending: (Message selector: #pushReceiver)].

	(self pendingMatches: {
		[:m | m selector == #storePopTemp: and: [m argument = index]]}
		) ifTrue: [^ self pendingSelector: #storeTemp:].

	self doPending.

	index = -2 ifTrue: [^ gen pushThisContext].
	index = -1 ifTrue: [
		^ gen pushThisContext;
			pushLiteral: MethodContext myEnvFieldIndex;
			send: #privGetField:].

	gen pushTemp: index.
! !

!IRTranslator methodsFor: 'instructions' stamp: 'ajh 3/9/2003 22:12'!
remoteReturn

	self doPending.
	gen remoteReturn.
! !

!IRTranslator methodsFor: 'instructions' stamp: 'ajh 3/10/2003 16:27'!
returnTop

	#pushReceiver == self pendingSelector ifTrue: [
		self pendingSelector: #returnReceiver.
		^ self doPending
	].
	#pushLiteral: == self pendingSelector ifTrue: [
		self pendingSelector: #returnConstant:.
		^ self doPending
	].
	#pushInstVar: == self pendingSelector ifTrue: [
		self pendingSelector: #returnInstVar:.
		^ self doPending
	].
	self doPending.
	gen returnTop.
! !

!IRTranslator methodsFor: 'instructions' stamp: 'ajh 7/3/2004 15:16'!
send: selector

	"If get/set inst var, access it directly"
	| index |
	((#(privGetField: #privStoreIn:field:) identityIncludes: selector) and:
	 [self pendingMatches: {
		[:m | m selector == #pushReceiver].
		[:m | m selector == #pushLiteral: and: [m argument isInteger]]}]
	) ifTrue: [
		index _ self popPending argument.
		self popPending.  "pop pushReceiver"
		self addPending: (Message
			selector: (selector == #privGetField:
				ifTrue: [#pushInstVar:] ifFalse: [#storeInstVar:])
			argument: index).
		(self pendingMatches: {
			[:m | m selector == #storePopInstVar: and: [m argument = index]].
			[:m | m selector == #pushInstVar: and: [m argument = index]]}
		) ifTrue: [
			self popPending.
			self pendingSelector: #storeInstVar:.
		].
		^ self
	].

	"otherwise do normal send"
	self doPending.
	gen send: selector.
! !

!IRTranslator methodsFor: 'instructions' stamp: 'ajh 3/9/2003 22:14'!
send: selector toSuperOf: behavior

	self doPending.
	gen send: selector toSuperOf: behavior.
! !

!IRTranslator methodsFor: 'instructions' stamp: 'ajh 7/3/2004 15:17'!
storeTemp: index

	index = -1 "thisEnv" ifTrue: [
		self doPending.
		^ gen pushThisContext;
			pushLiteral: MethodContext myEnvFieldIndex;
			send: #privStoreIn:field:].

	self addPending: (Message selector: #storeTemp: argument: index)! !


!IRTranslator methodsFor: 'interpret' stamp: 'ajh 3/15/2003 02:02'!
interpret: ir

	ir optimize.
	gen primitiveNode: ir primitiveNode.
	gen numArgs: ir numArgs.
	super interpret: ir.
! !

!IRTranslator methodsFor: 'interpret' stamp: 'ajh 3/13/2003 18:07'!
interpretAll: irSequences

	irSequences withIndexDo: [:seq :i | seq orderNumber: i].
	super interpretAll: irSequences.
! !

!IRTranslator methodsFor: 'interpret' stamp: 'ajh 3/13/2003 04:50'!
interpretInstruction: irInstruction

	currentInstr _ irInstruction.
	super interpretInstruction: irInstruction.
! !


!IRTranslator methodsFor: 'priv pending' stamp: 'ajh 6/22/2003 14:54'!
addPending: message

	pending addLast: currentInstr -> message! !

!IRTranslator methodsFor: 'priv pending' stamp: 'ajh 3/13/2003 04:49'!
doPending
	"execute pending instructions"

	| assoc |
	[pending isEmpty] whileFalse: [
		assoc _ pending removeFirst.
		gen mapBytesTo: assoc key "instr".
		assoc value "message" sendTo: gen.
	].
	gen mapBytesTo: currentInstr.
! !

!IRTranslator methodsFor: 'priv pending' stamp: 'ajh 6/5/2003 12:41'!
pendingMatches: blocks
	"Return true if each message at end of pending list satisfies its corresponding block.  The number of elements tested equals the number of blocks.  If not enough elements return false"

	| messages i |
	messages _ pending collect: [:assoc | assoc value].
	blocks size > messages size ifTrue: [^ false].
	i _ messages size - blocks size.
	blocks do: [:b |
		(b value: (messages at: (i _ i + 1))) ifFalse: [^ false].
	].
	^ true! !

!IRTranslator methodsFor: 'priv pending' stamp: 'ajh 6/22/2003 21:06'!
pendingSelector

	pending isEmpty ifTrue: [^ nil].
	^ pending last value "message" selector! !

!IRTranslator methodsFor: 'priv pending' stamp: 'ajh 6/22/2003 21:06'!
pendingSelector: selector

	pending last value "message" setSelector: selector! !

!IRTranslator methodsFor: 'priv pending' stamp: 'ajh 6/22/2003 14:27'!
popPending

	^ pending removeLast value "message"! !


!IRTranslator methodsFor: 'results' stamp: 'ajh 6/28/2004 11:23'!
compiledMethod

	^ gen compiledMethodWith: trailerBytes! !

!IRTranslator methodsFor: 'results' stamp: 'ajh 3/13/2003 18:28'!
compiledMethodWith: trailer

	^ gen compiledMethodWith: trailer! !
