RBProgramNodeVisitor subclass: #ASTChecker
	instanceVariableNames: 'scope'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Semantics'!
!ASTChecker commentStamp: 'ajh 6/23/2004 19:43' prior: 0!
I visit each node in the abstract syntax tree while growing and shrinking a SemScope chain. Each method and block node is linked with its corresponding scope object, and each variable def and ref is linked with its corresponding ScopeVar. Exceptions are raised for undefined variable references and so on (see subclasses of SemanticWarning).
!


!ASTChecker methodsFor: 'visitor-double dispatching' stamp: 'ajh 2/25/2003 19:43'!
acceptArrayNode: anArrayNode 
	anArrayNode statements do: [:each | self visitNode: each]! !

!ASTChecker methodsFor: 'visitor-double dispatching' stamp: 'ajh 3/11/2003 18:21'!
acceptAssignmentNode: anAssignmentNode

	| var |
	self visitNode: anAssignmentNode value.
	var _ (scope lookupVar: anAssignmentNode variable name)
		ifNil: [self undeclaredVariable: anAssignmentNode variable].
	var markWrite.
	anAssignmentNode variable binding: var.
! !

!ASTChecker methodsFor: 'visitor-double dispatching' stamp: 'ajh 6/25/2004 17:28'!
acceptBlockNode: aBlockNode

	aBlockNode isInlined ifTrue: [^ self acceptInlinedBlockNode: aBlockNode].

	scope _ scope newFunctionScope.
	aBlockNode scope: scope.
	(scope addTemp: 'parent env') markArg. "first temp is receiver"
	aBlockNode arguments do: [:node | (self declareVariableNode: node) markArg].
	self visitNode: aBlockNode body.
	scope _ scope popScope.
! !

!ASTChecker methodsFor: 'visitor-double dispatching' stamp: 'ajh 2/25/2003 22:15'!
acceptCascadeNode: aCascadeNode 
	aCascadeNode messages do: [:each | self visitNode: each]! !

!ASTChecker methodsFor: 'visitor-double dispatching' stamp: 'ajh 7/8/2004 20:31'!
acceptDoItNode: aDoItNode

	| doItReceiverName |
	doItReceiverName _ scope isInstanceScope
		ifTrue: ['self']
		ifFalse: [(scope rawVar: 'theContext')
			ifNil: ['theContext']
			ifNotNil: ['the context']].

	scope _ scope newMethodScope.
	aDoItNode scope: scope.

	(scope addTemp: doItReceiverName) markArg. "first temp is receiver"
	self visitNode: aDoItNode body.
	scope _ scope popScope.
! !

!ASTChecker methodsFor: 'visitor-double dispatching' stamp: 'ajh 6/28/2004 16:49'!
acceptInlinedBlockNode: aBlockNode

	aBlockNode arguments do: [:node | (self declareVariableNode: node) markWrite "given"].
	self visitNode: aBlockNode body.
! !

!ASTChecker methodsFor: 'visitor-double dispatching' stamp: 'ajh 2/25/2003 19:34'!
acceptLiteralNode: aLiteralNode! !

!ASTChecker methodsFor: 'visitor-double dispatching' stamp: 'ajh 3/11/2003 22:51'!
acceptMessageNode: aMessageNode

	(aMessageNode isCascaded not or: [aMessageNode isFirstCascaded]) 
		ifTrue: [self visitNode: aMessageNode receiver].
	aMessageNode arguments do: [:each | self visitNode: each].
	"aMessageNode binding:"
		((scope lookupSelector: aMessageNode selectorString)
			ifNil: [self undeclaredSelector: aMessageNode])! !

!ASTChecker methodsFor: 'visitor-double dispatching' stamp: 'ajh 6/29/2004 16:36'!
acceptMethodNode: aMethodNode

	scope _ scope newMethodScope.
	aMethodNode scope: scope.
	(scope addTemp: 'self') markArg.  "first temp is receiver"
	aMethodNode arguments do: [:node | (self declareVariableNode: node) markArg].
	aMethodNode primitiveNode: (PrimitiveParser new
		startPosition: aMethodNode primitiveStartPosition;
		parse: aMethodNode primitiveText).
	self visitNode: aMethodNode body.
	scope _ scope outerScope.
! !

!ASTChecker methodsFor: 'visitor-double dispatching' stamp: 'ajh 6/26/2004 18:22'!
acceptReturnNode: aReturnNode

	| var |
	self visitNode: aReturnNode value.
	var _ scope lookupVar: 'top env'.  "nil var means local return"
	var ifNotNil: [var markRead].
	aReturnNode homeBinding: var.
! !

!ASTChecker methodsFor: 'visitor-double dispatching' stamp: 'ajh 6/29/2004 16:56'!
acceptSequenceNode: aSequenceNode

	aSequenceNode temporaries do: [:node | self declareVariableNode: node].
	aSequenceNode statements do: [:each | self visitNode: each].
	aSequenceNode temporaries do: [:node |
		node binding isUnused ifTrue: [self unusedVariable: node]].
! !

!ASTChecker methodsFor: 'visitor-double dispatching' stamp: 'ajh 6/25/2004 16:26'!
acceptVariableNode: aVariableNode

	| var name |
	name _ aVariableNode name.
	name = 'super' ifTrue: [name _ 'self'].
	var _ (scope lookupVar: name)
		ifNil: [self undeclaredVariable: aVariableNode].
	var isUndefined
		ifTrue: [self uninitializedVariable: aVariableNode].
	var markRead.
	aVariableNode binding: var.
! !

!ASTChecker methodsFor: 'visitor-double dispatching' stamp: 'ajh 6/30/2004 14:17'!
declareVariableNode: aVariableNode

	| name var |
	name _ aVariableNode name.
	var _ scope rawVar: name.
	var ifNotNil: [
		var scope = scope ifTrue: [
			"Reuse same var"
			var _ scope lookupVar: name.
		] ifFalse: [
			"Create new var that shadows outer one"
			self variable: aVariableNode shadows: var.
			var _ scope addTemp: name.
		]
	] ifNil: [
		"new var"
		var _ scope addTemp: name.
	].
	aVariableNode binding: var.
	^ var! !


!ASTChecker methodsFor: 'initialize-release' stamp: 'ajh 6/23/2004 20:40'!
initialize

	scope _ GlobalScope new.  "in case never initialized"! !

!ASTChecker methodsFor: 'initialize-release' stamp: 'ajh 2/26/2003 19:58'!
scope: aSemScope

	scope _ aSemScope! !


!ASTChecker methodsFor: 'error handling' stamp: 'ajh 3/11/2003 22:46'!
undeclaredSelector: messageNode

	^ UndeclaredSelectorWarning new
		messageNode: messageNode;
		signal! !

!ASTChecker methodsFor: 'error handling' stamp: 'ajh 3/11/2003 12:06'!
undeclaredVariable: variableNode

	^ UndeclaredVariableWarning new
		variableNode: variableNode;
		signal! !

!ASTChecker methodsFor: 'error handling' stamp: 'ajh 3/12/2003 13:35'!
uninitializedVariable: variableNode

	^ UninitializedVariableWarning new
		variableNode: variableNode;
		signal! !

!ASTChecker methodsFor: 'error handling' stamp: 'ajh 3/12/2003 14:10'!
unusedVariable: variableNode

	^ UnusedVariableWarning new
		variableNode: variableNode;
		signal! !

!ASTChecker methodsFor: 'error handling' stamp: 'ajh 3/19/2003 13:33'!
variable: variableNode shadows: semVar

	^ ShadowVariableWarning new
		variableNode: variableNode;
		shadowedVar: semVar;
		signal! !


RBProgramNodeVisitor subclass: #ASTTranslator
	instanceVariableNames: 'methodBuilder effectTranslator valueTranslator'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Semantics'!
!ASTTranslator commentStamp: 'ajh 3/24/2003 22:19' prior: 0!
I visit an abstract syntax tree and generate IR (intermediate representation) instructions for each node by sending the appropriate message to my methodBuilder (an IRBuilder).  I hold onto my two subclasses one for generating instructions for value, the other for generating instructions for effect.!


!ASTTranslator methodsFor: 'visitor-double dispatching' stamp: 'ajh 7/3/2004 15:26'!
acceptArrayNode: anArrayNode 

	| elementNodes |
	elementNodes _ anArrayNode children.
	elementNodes size <= 4 ifTrue: [
		"Short form: Array braceWith: a with: b ..."
		methodBuilder pushLiteral: Array holder; send: #value.
		elementNodes do: [:node | valueTranslator visitNode: node].
		methodBuilder send: (#(braceWithNone braceWith: braceWith:with: braceWith:with:with: braceWith:with:with:with:) at: elementNodes size + 1).
	] ifFalse: [
		"Long form: (Array braceStream: N) nextPut: a; nextPut: b; ...; braceArray"
		methodBuilder pushLiteral: Array holder; send: #value.
		methodBuilder pushLiteral: elementNodes size.
		methodBuilder send: #braceStream:.
		elementNodes do: [:node |
			methodBuilder pushDup.
			valueTranslator visitNode: node.
			methodBuilder send: #nextPut:.
			methodBuilder popTop.
		].
		methodBuilder send: #braceArray.
	].
! !

!ASTTranslator methodsFor: 'visitor-double dispatching' stamp: 'ajh 7/8/2004 10:43'!
acceptAssignmentNode: anAssignmentNode 

	valueTranslator visitNode: anAssignmentNode value.
	anAssignmentNode variable binding emitStore: methodBuilder from: anAssignmentNode owningScope.
! !

!ASTTranslator methodsFor: 'visitor-double dispatching' stamp: 'ajh 2/25/2003 00:51'!
acceptBlockNode: aBlockNode 
! !

!ASTTranslator methodsFor: 'visitor-double dispatching' stamp: 'ajh 3/10/2003 14:53'!
acceptCascadeNode: aCascadeNode 

	valueTranslator visitNode: aCascadeNode receiver.
	aCascadeNode messages allButLastDo: [:node |
		methodBuilder pushDup.
		effectTranslator visitNode: node.
	].
	valueTranslator visitNode: aCascadeNode messages last.
! !

!ASTTranslator methodsFor: 'visitor-double dispatching' stamp: 'ajh 6/26/2004 13:48'!
acceptDoItNode: aDoItNode 

	methodBuilder numRargs: 1.
	methodBuilder addTemps: aDoItNode scope tempVars.
	aDoItNode scope emitPrologue: methodBuilder.
	valueTranslator visitNode: aDoItNode body.
	aDoItNode body lastIsReturn ifFalse: [
		methodBuilder returnTop].
! !

!ASTTranslator methodsFor: 'visitor-double dispatching' stamp: 'ajh 2/24/2003 20:06'!
acceptLiteralNode: aLiteralNode! !

!ASTTranslator methodsFor: 'visitor-double dispatching' stamp: 'ajh 2/25/2003 13:23'!
acceptMessageNode: aMessageNode 

	aMessageNode isInlineIf ifTrue: [^ self emitIfNode: aMessageNode].
	aMessageNode isInlineIfNil ifTrue: [^ self emitIfNilNode: aMessageNode].
	aMessageNode isInlineAndOr ifTrue: [^ self emitAndOrNode: aMessageNode].
	aMessageNode isInlineWhile ifTrue: [^ self emitWhileNode: aMessageNode].
	aMessageNode isInlineToDo ifTrue: [^ self emitToDoNode: aMessageNode].
	aMessageNode isInlineCase ifTrue: [^ self emitCaseNode: aMessageNode].
	^ self emitMessageNode: aMessageNode! !

!ASTTranslator methodsFor: 'visitor-double dispatching' stamp: 'ajh 6/26/2004 13:48'!
acceptMethodNode: aMethodNode 

	methodBuilder numRargs: aMethodNode arguments size + 1.
	methodBuilder primitiveNode: aMethodNode primitiveNode.
	methodBuilder addTemps: aMethodNode scope tempVars.
	aMethodNode scope emitPrologue: methodBuilder.
	effectTranslator visitNode: aMethodNode body.
	aMethodNode body lastIsReturn ifFalse: [
		methodBuilder pushReceiver; returnTop].
! !

!ASTTranslator methodsFor: 'visitor-double dispatching' stamp: 'ajh 7/8/2004 19:01'!
acceptReturnNode: aReturnNode

	valueTranslator visitNode: aReturnNode value.
	aReturnNode homeBinding ifNil: [
		methodBuilder returnTop.
	] ifNotNil: [
		aReturnNode homeBinding emitValue: methodBuilder from: aReturnNode owningScope.
		methodBuilder remoteReturn.
	].
! !

!ASTTranslator methodsFor: 'visitor-double dispatching' stamp: 'ajh 6/24/2004 14:57'!
acceptSequenceNode: aSequenceNode 

	aSequenceNode statements allButLastDo: [:n | effectTranslator visitNode: n].
	aSequenceNode statements notEmpty ifTrue: [self visitNode: aSequenceNode statements last].
! !

!ASTTranslator methodsFor: 'visitor-double dispatching' stamp: 'ajh 2/24/2003 20:06'!
acceptVariableNode: aVariableNode! !

!ASTTranslator methodsFor: 'visitor-double dispatching' stamp: 'ajh 6/26/2004 18:32'!
translateBlockNode: aBlockNode 

	methodBuilder numRargs: aBlockNode arguments size + 1.
	methodBuilder addTemps: aBlockNode scope tempVars.
	aBlockNode scope emitPrologue: methodBuilder.
	valueTranslator visitNode: aBlockNode body.
	aBlockNode body lastIsReturn ifFalse: [
		methodBuilder returnTop].
! !


!ASTTranslator methodsFor: 'inline messages' stamp: 'ajh 3/13/2003 00:21'!
emitAndOrNode: aMessageNode

	valueTranslator visitNode: aMessageNode receiver.
	methodBuilder jumpAheadTo: #else if: (aMessageNode selector == #or:).
	self visitNode: aMessageNode arguments first body.
	self == valueTranslator ifTrue: [
		methodBuilder jumpAheadTo: #end].
	methodBuilder jumpAheadTarget: #else.
	self == valueTranslator ifTrue: [
		methodBuilder pushLiteral: (aMessageNode selector == #or:).
		methodBuilder jumpAheadTarget: #end.
	].
! !

!ASTTranslator methodsFor: 'inline messages' stamp: 'ajh 7/8/2004 19:02'!
emitCaseNode: aMessageNode

	| cases assocMessageNode s |
	cases _ aMessageNode arguments first statements.
	valueTranslator visitNode: aMessageNode receiver.
	1 to: cases size - 1 do: [:i |
		methodBuilder pushDup.
		assocMessageNode _ cases at: i.
		valueTranslator visitNode: assocMessageNode receiver body.
		methodBuilder send: #=.
		methodBuilder jumpAheadTo: #next if: false.
		methodBuilder popTop.
		self visitNode: assocMessageNode arguments first body.
		methodBuilder jumpAheadTo: #end.
		methodBuilder jumpAheadTarget: #next.
	].
	aMessageNode arguments size = 2 ifTrue: [
		"last case with otherwise"
		assocMessageNode _ cases last.
		valueTranslator visitNode: assocMessageNode receiver body.
		methodBuilder send: #=.
		methodBuilder jumpAheadTo: #next if: false.
		self visitNode: assocMessageNode arguments first body.
		methodBuilder jumpAheadTo: #end.
		methodBuilder jumpAheadTarget: #next.
		self visitNode: aMessageNode arguments last body.
	] ifFalse: [
		"last case without otherwise"
		methodBuilder pushDup.
		assocMessageNode _ cases last.
		valueTranslator visitNode: assocMessageNode receiver body.
		methodBuilder send: #=.
		methodBuilder jumpAheadTo: #next if: false.
		methodBuilder popTop.
		self visitNode: assocMessageNode arguments first body.
		methodBuilder jumpAheadTo: #end.
		methodBuilder jumpAheadTarget: #next.
		methodBuilder send: #caseError.
		aMessageNode lastIsReturn ifTrue: [
			(s _ aMessageNode owningScope) isBlockScope ifTrue: [
				(s lookupVar: 'top env') emitValue: methodBuilder from: aMessageNode owningScope.
				methodBuilder remoteReturn.
			] ifFalse: [
				methodBuilder returnTop.
			]
		] ifFalse: [
			self == effectTranslator ifTrue: [methodBuilder popTop].
		].
	].

	"there has to be a one-to-one correspondence between every jump and target"	
	1 to: cases size do: [:i | methodBuilder jumpAheadTarget: #end].
! !

!ASTTranslator methodsFor: 'inline messages' stamp: 'ajh 7/8/2004 19:00'!
emitIfNilNode: aMessageNode

	| args selector |
	valueTranslator visitNode: aMessageNode receiver.
	args _ aMessageNode arguments.
	(selector _ aMessageNode selector) caseOf: {
		[#ifNil:] -> [self == valueTranslator ifTrue: [methodBuilder pushDup]].
		[#ifNil:ifNotNil:] -> [args last arguments isEmpty ifFalse: [args first arguments first binding emitLocalStore: methodBuilder]].
		[#ifNotNil:] -> [args first arguments isEmpty ifFalse: [args first arguments first binding emitLocalStore: methodBuilder]].
		[#ifNotNil:ifNil:] -> [args first arguments isEmpty ifFalse: [args first arguments first binding emitLocalStore: methodBuilder]]
	}.
	methodBuilder pushLiteral: nil.
	methodBuilder send: #==.
	methodBuilder jumpAheadTo: #else if: (selector beginsWith: #ifNotNil:).
	(self == valueTranslator and: [selector == #ifNil:])
		ifTrue: [methodBuilder popTop].
	self visitNode: args first body.
	(args size > 1 or: [self == valueTranslator and: [selector == #ifNotNil:]])
		ifTrue: [methodBuilder jumpAheadTo: #end].
	methodBuilder jumpAheadTarget: #else.
	(args size > 1 or: [self == valueTranslator and: [selector == #ifNotNil:]]) ifTrue: [
		args size > 1
			ifTrue: [self visitNode: args last body]
			ifFalse: [methodBuilder pushLiteral: nil].
		methodBuilder jumpAheadTarget: #end.
	].
! !

!ASTTranslator methodsFor: 'inline messages' stamp: 'ajh 2/27/2003 14:32'!
emitIfNode: aMessageNode

	| args |
	valueTranslator visitNode: aMessageNode receiver.
	methodBuilder jumpAheadTo: #else if: (aMessageNode selector beginsWith: #ifFalse:).
	args _ aMessageNode arguments.
	self visitNode: args first body.
	(args size > 1 or: [self == valueTranslator]) ifTrue: [
		methodBuilder jumpAheadTo: #end].
	methodBuilder jumpAheadTarget: #else.
	(args size > 1 or: [self == valueTranslator]) ifTrue: [
		args size > 1
			ifTrue: [self visitNode: args last body]
			ifFalse: [methodBuilder pushLiteral: nil].
		methodBuilder jumpAheadTarget: #end.
	].
! !

!ASTTranslator methodsFor: 'inline messages' stamp: 'ajh 3/13/2003 04:20'!
emitMessageNode: aMessageNode

	| r |
	aMessageNode isCascaded ifFalse: [
		valueTranslator visitNode: aMessageNode receiver].
	aMessageNode arguments do: [:each | 
		valueTranslator visitNode: each].
	r _ aMessageNode receiver.
	(r isVariable and: [r name = 'super' and: [r binding name = 'self']])
		ifTrue: [methodBuilder send: aMessageNode selector toSuperOf: aMessageNode owningScope classEncoding]
		ifFalse: [methodBuilder send: aMessageNode selector].
! !

!ASTTranslator methodsFor: 'inline messages' stamp: 'ajh 7/8/2004 19:03'!
emitToDoNode: aMessageNode

	| limit step block iterator limitEmit |
	limit _ aMessageNode arguments first.
	step _ 1.
	aMessageNode arguments size = 3 ifTrue: [
		"to:by:do:"
		step _ aMessageNode arguments second.
		step isLiteral ifFalse: [self error: 'should not have been inlined'].
		step _ step value.
	].
	block _ aMessageNode arguments last.

	"push start. allocate and initialize iterator"
	valueTranslator visitNode: aMessageNode receiver.
	iterator _ block arguments first binding.
	iterator emitLocalStore: methodBuilder.

	"push limit, store in a temp if it's an expression or a changing variable"
	valueTranslator visitNode: limit.
	limitEmit _ [valueTranslator visitNode: limit].
	(limit isLiteral or: 
	 [limit isVariable and:
	  [ParseTreeSearcher new
		matches: limit name , ' := `@object' do: [:n :a | false];
		executeTree: block initialAnswer: true]]
	) ifFalse: [
		limit _ {iterator. #limit}.  "must be unique throughout method"
		methodBuilder addTemp: limit.
		methodBuilder storeTemp: limit.
		limitEmit _ [methodBuilder pushTemp: limit].
	].

	"loop"
	methodBuilder jumpBackTarget: #start.
	methodBuilder send: (step > 0 ifTrue: [#<=] ifFalse: [#>=]).
	methodBuilder jumpAheadTo: #done if: false.
	effectTranslator visitNode: block body.
	iterator emitLocalValue: methodBuilder.
	methodBuilder pushLiteral: step.
	methodBuilder send: #+.
	iterator emitLocalStore: methodBuilder.
	limitEmit value.
	methodBuilder jumpBackTo: #start.
	methodBuilder jumpAheadTarget: #done.
! !

!ASTTranslator methodsFor: 'inline messages' stamp: 'ajh 2/27/2003 14:33'!
emitWhileNode: aMessageNode

	methodBuilder jumpBackTarget: #begin.
	valueTranslator visitNode: aMessageNode receiver body.
	aMessageNode selector caseOf: {
		[#whileTrue:]	-> [methodBuilder jumpAheadTo: #end if: false].
		[#whileTrue]	-> [methodBuilder jumpAheadTo: #end if: false].
		[#whileFalse:]	-> [methodBuilder jumpAheadTo: #end if: true].
		[#whileFalse]	-> [methodBuilder jumpAheadTo: #end if: true]
	}.
	aMessageNode arguments notEmpty ifTrue: [
		effectTranslator visitNode: aMessageNode arguments first body].
	methodBuilder jumpBackTo: #begin.
	methodBuilder jumpAheadTarget: #end.
! !


!ASTTranslator methodsFor: 'visiting' stamp: 'ajh 3/2/2003 03:31'!
visitNode: aNode

	methodBuilder mapToNode: aNode.
	super visitNode: aNode.
	methodBuilder popMap.
! !


!ASTTranslator methodsFor: 'initialize' stamp: 'ajh 6/23/2004 19:53'!
initialize

	methodBuilder _ IRBuilder new.
	effectTranslator _ self as: ASTTranslatorForEffect.
	valueTranslator _ self as: ASTTranslatorForValue.
	effectTranslator instVarNamed: #effectTranslator put: effectTranslator.
	effectTranslator instVarNamed: #valueTranslator put: valueTranslator.
	valueTranslator instVarNamed: #valueTranslator put: valueTranslator.
! !


!ASTTranslator methodsFor: 'accessing' stamp: 'ajh 3/10/2003 17:59'!
ir

	^ methodBuilder ir! !


ASTTranslator subclass: #ASTTranslatorForEffect
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Semantics'!
!ASTTranslatorForEffect commentStamp: 'ajh 3/24/2003 22:20' prior: 0!
I override some methods in my super to generate instructions for effect only.!


!ASTTranslatorForEffect methodsFor: 'visitor-double dispatching' stamp: 'ajh 3/10/2003 14:56'!
acceptArrayNode: anArrayNode 

	super acceptArrayNode: anArrayNode.
	methodBuilder popTop.
! !

!ASTTranslatorForEffect methodsFor: 'visitor-double dispatching' stamp: 'ajh 3/10/2003 14:56'!
acceptAssignmentNode: anAssignmentNode 

	super acceptAssignmentNode: anAssignmentNode.
	methodBuilder popTop.
! !

!ASTTranslatorForEffect methodsFor: 'visitor-double dispatching' stamp: 'ajh 3/10/2003 14:56'!
acceptCascadeNode: aCascadeNode 

	super acceptCascadeNode: aCascadeNode.
	methodBuilder popTop.
! !


!ASTTranslatorForEffect methodsFor: 'emit messages' stamp: 'ajh 3/10/2003 14:57'!
emitMessageNode: aMessageNode 

	super emitMessageNode: aMessageNode.
	methodBuilder popTop.
! !


ASTTranslator subclass: #ASTTranslatorForValue
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Semantics'!
!ASTTranslatorForValue commentStamp: 'ajh 3/24/2003 22:20' prior: 0!
I override some methods in my super to generate instructions for effect and value only.!


!ASTTranslatorForValue methodsFor: 'visitor-double dispatching' stamp: 'ajh 7/8/2004 19:19'!
acceptBlockNode: aBlockNode 

	| blockMethod |
	blockMethod _ aBlockNode generateIR.

	aBlockNode freeVars isEmpty ifTrue: [
		"Create block at compile time"
		methodBuilder pushBlock: blockMethod.
	] ifFalse: [
		"Create block at run time"
		| outerScope envVar |
		outerScope _ aBlockNode scope outerScope.
		envVar _ outerScope hasEscapingEnv
			ifTrue: [outerScope thisEnvVar]
			ifFalse: [outerScope receiverVar].
		methodBuilder pushBlockMethod: blockMethod.
		envVar emitLocalValue: methodBuilder.
		methodBuilder send: #createBlock:.
	].
! !

!ASTTranslatorForValue methodsFor: 'visitor-double dispatching' stamp: 'ajh 2/27/2003 18:18'!
acceptLiteralNode: aLiteralNode

	methodBuilder pushLiteral: aLiteralNode value.
! !

!ASTTranslatorForValue methodsFor: 'visitor-double dispatching' stamp: 'ajh 2/25/2003 18:21'!
acceptSequenceNode: aSequenceNode 

	super acceptSequenceNode: aSequenceNode.
	aSequenceNode statements isEmpty ifTrue: [
		methodBuilder pushLiteral: nil].
! !

!ASTTranslatorForValue methodsFor: 'visitor-double dispatching' stamp: 'ajh 7/8/2004 10:45'!
acceptVariableNode: aVariableNode

	aVariableNode binding emitValue: methodBuilder from: aVariableNode owningScope.
! !


!ASTTranslatorForValue methodsFor: 'emit messages' stamp: 'ajh 2/25/2003 16:50'!
emitToDoNode: aMessageNode

	super emitToDoNode: aMessageNode.
	methodBuilder pushLiteral: nil.
! !

!ASTTranslatorForValue methodsFor: 'emit messages' stamp: 'ajh 2/25/2003 16:49'!
emitWhileNode: aMessageNode

	super emitWhileNode: aMessageNode.
	methodBuilder pushLiteral: nil.
! !


Object subclass: #LexicalScope
	instanceVariableNames: 'outerScope'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Semantics'!
!LexicalScope commentStamp: 'ajh 3/24/2003 21:53' prior: 0!
I am a symbol table where variable names are associated with SemVars.  Each context (method/closure) get a fresh scope that inherits from its outer scope.!


!LexicalScope methodsFor: 'levels' stamp: 'ajh 7/8/2004 20:02'!
hasOuter: scope

	outerScope ifNil: [^ false].
	^ outerScope = scope or: [outerScope hasOuter: scope]! !

!LexicalScope methodsFor: 'levels' stamp: 'ajh 6/25/2004 17:27'!
newBlockScope

	^ self newFunctionScope! !

!LexicalScope methodsFor: 'levels' stamp: 'ajh 6/25/2004 17:26'!
newFunctionScope

	^ FunctionScope new outerScope: self! !

!LexicalScope methodsFor: 'levels' stamp: 'ajh 6/29/2004 16:37'!
newMethodScope

	^ self newFunctionScope isHome: true! !

!LexicalScope methodsFor: 'levels' stamp: 'ajh 2/26/2003 13:44'!
outerScope

	^ outerScope! !

!LexicalScope methodsFor: 'levels' stamp: 'ajh 3/20/2003 11:27'!
scopeLevel

	outerScope ifNil: [^ 0].
	^ outerScope scopeLevel + 1! !


!LexicalScope methodsFor: 'adding' stamp: 'ajh 2/25/2003 22:44'!
addSelector: string

	^ string asSymbol! !


!LexicalScope methodsFor: 'lookup' stamp: 'ajh 6/25/2004 22:00'!
captureVar: name

	^ self lookupVar: name! !

!LexicalScope methodsFor: 'lookup' stamp: 'ajh 2/27/2003 00:40'!
classEncoding

	^ self outerScope classEncoding! !

!LexicalScope methodsFor: 'lookup' stamp: 'ajh 2/25/2003 22:41'!
lookupSelector: name

	Symbol hasInterned: name ifTrue: [ :sym | ^ sym].
	^ nil! !

!LexicalScope methodsFor: 'lookup' stamp: 'ajh 3/11/2003 20:19'!
lookupVar: name
	"subclass responsibility"

	^ self outerScope lookupVar: name! !

!LexicalScope methodsFor: 'lookup' stamp: 'ajh 2/25/2003 22:40'!
possibleSelectorsFor: string

	^ Symbol possibleSelectorsFor: string! !

!LexicalScope methodsFor: 'lookup' stamp: 'ajh 2/25/2003 22:39'!
possibleVarsFor: string

	| list |
	list _ self possibleVarsFor: string continued: nil.
	^ string correctAgainst: nil continuedFrom: list
! !

!LexicalScope methodsFor: 'lookup' stamp: 'ajh 3/11/2003 20:17'!
possibleVarsFor: name continued: listOrNil
	"subclass responsibility"

	^ self outerScope possibleVarsFor: name continued: listOrNil! !

!LexicalScope methodsFor: 'lookup' stamp: 'ajh 3/25/2003 23:38'!
rawVar: name
	"subclass responsibility"

	^ self outerScope rawVar: name! !


!LexicalScope methodsFor: 'initializing' stamp: 'ajh 2/25/2003 21:27'!
initialize
! !

!LexicalScope methodsFor: 'initializing' stamp: 'ajh 2/26/2003 13:43'!
outerScope: aSemScope

	outerScope _ aSemScope! !

!LexicalScope methodsFor: 'initializing' stamp: 'ajh 2/26/2003 20:01'!
parseScope

	^ self! !


!LexicalScope methodsFor: 'old protocol' stamp: 'ajh 2/25/2003 20:09'!
requestor: editor
	"Old protocol, do nothing"
! !


!LexicalScope methodsFor: 'scope' stamp: 'ajh 7/8/2004 18:39'!
isDoItScope

	^ false! !

!LexicalScope methodsFor: 'scope' stamp: 'ajh 6/25/2004 15:20'!
isFunctionScope

	^ false! !

!LexicalScope methodsFor: 'scope' stamp: 'ajh 7/8/2004 20:26'!
isInstanceScope

	^ false! !


LexicalScope subclass: #ClassScope
	instanceVariableNames: 'class'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Semantics'!
!ClassScope commentStamp: 'ajh 3/24/2003 21:44' prior: 0!
I include all variable enries for class, pool and global vars.!


!ClassScope methodsFor: 'initializing' stamp: 'ajh 2/25/2003 20:10'!
class: aBehavior

	class _ aBehavior! !


!ClassScope methodsFor: 'levels' stamp: 'ajh 6/24/2004 04:03'!
instanceScope

	^ InstanceScope new
		vars: class allInstVarNames;
		outerScope: self! !

!ClassScope methodsFor: 'levels' stamp: 'ajh 6/26/2004 17:56'!
newFunctionScope

	^ self instanceScope newFunctionScope! !


!ClassScope methodsFor: 'lookup' stamp: 'ajh 2/25/2003 20:10'!
classEncoding

	^ class! !

!ClassScope methodsFor: 'lookup' stamp: 'ajh 6/30/2004 14:18'!
lookupVar: name
	"Return a SemVar for my pool var with this name.  Return nil if none found"

	class scopeHas: name asSymbol ifTrue:
		[:assoc | ^ GlobalVar new assoc: assoc; scope: self].
	^ nil
! !

!ClassScope methodsFor: 'lookup' stamp: 'ajh 3/18/2003 15:42'!
possibleVarsFor: name continued: listOrNil
	"Return my pool var names that are close to name"

	name first isUppercase ifFalse: [^ listOrNil].
	^ class possibleVariablesFor: name continuedFrom: listOrNil! !

!ClassScope methodsFor: 'lookup' stamp: 'ajh 6/30/2004 14:18'!
rawVar: name
	"Return a SemVar for my pool var with this name.  Return nil if none found"

	class scopeHas: name asSymbol ifTrue:
		[:assoc | ^ GlobalVar new assoc: assoc; scope: self].
	^ nil
! !


!ClassScope methodsFor: 'printing' stamp: 'ajh 2/25/2003 20:10'!
printOn: stream

	class printOn: stream.
	stream nextPutAll: ' parseScope'.
! !


LexicalScope subclass: #FunctionScope
	instanceVariableNames: 'thisContextVar thisEnvVar tempVars capturedVars isHome hasInnerFreeVars hideTemps'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Semantics'!
!FunctionScope commentStamp: 'ajh 7/8/2004 18:33' prior: 0!
I am the symbol table (for declared temp vars) for methods and blocks (non-inlined closures). Scopes nest mirroring the block nesting (skipping inlined blocks like ifTrue: blocks). Variable references are looked up in the closest scope first then if not found searches the outer scope and so on. Temp variables found in an outer scope become captured in that outer scope. The emitPrologue: checks if any temp were captured and if so generates code that creates a ClosureEnvironment for those captured vars and copies any captured rargs (receiver and arguments) into it. This way MethodContexts are LIFO. Even a non-local (remote) return does not hold onto its home context, but rather holds onto its home context's environment. Upon return the caller stack is searched for the context that points to that home environment and returns to it. This allows stacks to be copied and blocks to still return to the correct context within the current thread.
!


!FunctionScope methodsFor: 'scope' stamp: 'ajh 6/26/2004 13:57'!
isBlockScope

	^ self isMethodScope not! !

!FunctionScope methodsFor: 'scope' stamp: 'ajh 6/25/2004 15:18'!
isFunctionScope

	^ true! !

!FunctionScope methodsFor: 'scope' stamp: 'ajh 6/29/2004 16:38'!
isHome
	"the home scope is the one that returns (^) return from"

	^ isHome = true! !

!FunctionScope methodsFor: 'scope' stamp: 'ajh 6/29/2004 16:38'!
isHome: bool
	"the home scope is the one that returns (^) return from"

	isHome _ bool! !

!FunctionScope methodsFor: 'scope' stamp: 'ajh 6/29/2004 16:38'!
isMethodScope

	^ self isHome! !

!FunctionScope methodsFor: 'scope' stamp: 'ajh 7/8/2004 14:19'!
popScope
	"Propogate free var usages to their outer vars, then return outer scope"

	^ self outerScope! !


!FunctionScope methodsFor: 'lookup' stamp: 'ajh 7/8/2004 19:52'!
captureVar: name
	"Return the ScopeVar with this name. If a temp move to captured."

	| var |
	capturedVars at: name ifPresent: [:v | ^ v].
	self hideTemps ifFalse: [
		tempVars at: name ifPresent: [:v | ^ self moveToCaptured: v]].
	(name = 'top env' and: [self isHome]) ifTrue: [^ self forceThisEnv].
	var _ self outerScope captureVar: name.
	(var notNil and: [var isGlobal not]) ifTrue: [self hasInnerFreeVars: true].
	^ var! !

!FunctionScope methodsFor: 'lookup' stamp: 'ajh 7/8/2004 14:16'!
lookupVar: name
	"Return the ScopeVar with this name. If found in outer scope then move it to captured."

	tempVars at: name ifPresent: [:v | ^ v].
	capturedVars at: name ifPresent: [:v | ^ v].
	name = 'thisEnv' ifTrue: [^ self forceThisEnv].
	name = 'thisContext' ifTrue: [^ thisContextVar].
	(name = 'top env' and: [self isHome]) ifTrue: [^ nil]. "'top env' used by return node"
	^ self outerScope captureVar: name! !

!FunctionScope methodsFor: 'lookup' stamp: 'ajh 7/8/2004 16:10'!
possibleVarsFor: name continued: listOrNil
	"Return my var names that are close in spelling to name"

	| list |
	list _ listOrNil.
	name first isLowercase ifTrue: [
		{tempVars. capturedVars} do: [:vars |
			list _ name correctAgainstDictionary: vars continuedFrom: list]
	].
	^ self outerScope possibleVarsFor: name continued: list! !

!FunctionScope methodsFor: 'lookup' stamp: 'ajh 7/8/2004 14:23'!
rawVar: name
	"Return the var with this name without capturing temps"

	tempVars at: name ifPresent: [:v | ^ v].
	capturedVars at: name ifPresent: [:v | ^ v].
	name = 'thisEnv' ifTrue: [^ self forceThisEnv].
	name = 'thisContext' ifTrue: [^ thisContextVar].
	^ self outerScope captureVar: name! !


!FunctionScope methodsFor: 'temp vars' stamp: 'ajh 6/26/2004 17:46'!
addTemp: name

	^ tempVars add: (TempVar new
			name: name;
			index: tempVars size; "zero-based; receiver is at 0"
			scope: self)! !

!FunctionScope methodsFor: 'temp vars' stamp: 'ajh 7/8/2004 19:50'!
hideTemps
	"If true do not allow inner scopes to capture my temps. Used by do-it scopes, which are fetching from an existing context env which no longer has its temps"

	^ hideTemps = true! !

!FunctionScope methodsFor: 'temp vars' stamp: 'ajh 6/25/2004 17:12'!
receiverVar

	^ self tempVars first! !

!FunctionScope methodsFor: 'temp vars' stamp: 'ajh 6/25/2004 22:05'!
removeTemp: tempVar

	tempVars remove: tempVar.
	tempVars do: [:var |
		var index > tempVar index ifTrue: [
			var index: var index - 1]].
! !

!FunctionScope methodsFor: 'temp vars' stamp: 'ajh 6/25/2004 22:23'!
tempVars

	^ tempVars asSortedCollection: [:x :y | x index <= y index]! !

!FunctionScope methodsFor: 'temp vars' stamp: 'ajh 7/8/2004 19:52'!
tempsAreNoLongerAvailable
	"Hide my temps and outer temps so they don't get captured (becuase they weren't captured and their values are not in the do-it env chain)"

	hideTemps _ true.
	self outerScope tempsAreNoLongerAvailable.
! !


!FunctionScope methodsFor: 'initializing' stamp: 'ajh 7/8/2004 20:23'!
asDoItScope

	^ (self primitiveChangeClassTo: ContextDoItScope basicNew) hideOuterTemps! !

!FunctionScope methodsFor: 'initializing' stamp: 'ajh 7/8/2004 14:24'!
initialize

	tempVars _ KeyedSet keyBlock: [:var | var name].
	capturedVars _ KeyedSet keyBlock: [:var | var name].
	thisContextVar _ ThisContextVar new
		name: 'thisContext';
		index: -2;
		scope: self.
! !


!FunctionScope methodsFor: 'decompiling' stamp: 'ajh 7/3/2004 18:08'!
receiverVarAt: instVarIndex

	^ self vars detect: [:var | var isFieldVar and: [var index = instVarIndex]] ifNone: [
		| name |
		name _ 'f', instVarIndex asString.
		self vars at: name put: 
			("ScopeFieldVar" self halt new
				name: name;
				receiverVar: self receiverVar;
				index: instVarIndex)
	  ]! !

!FunctionScope methodsFor: 'decompiling' stamp: 'ajh 6/27/2004 15:33'!
tempVarAt: tempIndex

	tempIndex = -2 ifTrue: [^ self thisContextVar].
	tempIndex = -1 ifTrue: [^ self thisEnvVar].
	^ tempVars detect: [:var | var index = tempIndex] ifNone: [
		| name |
		name _ String streamContents: [:str |
			self scopeLevel timesRepeat: [str nextPut: $t].
			tempIndex printOn: str].
		tempVars add: 
			(TempVar new
				name: name;
				index: tempIndex;
				scope: self)
	  ]! !


!FunctionScope methodsFor: 'printing' stamp: 'ajh 3/20/2003 11:29'!
printOn: stream

	super printOn: stream.
	stream space.
	self scopeLevel printOn: stream.
! !


!FunctionScope methodsFor: 'env' stamp: 'ajh 7/8/2004 14:37'!
forceThisEnv
	"create local env (if not already)"

	thisEnvVar ifNotNil: [^ thisEnvVar].
	^ thisEnvVar _ ThisEnvVar new
			name: 'thisEnv';
			index: -1;
			scope: self! !

!FunctionScope methodsFor: 'env' stamp: 'ajh 7/8/2004 14:30'!
hasEscapingEnv
	"has own env?"

	^ thisEnvVar notNil! !

!FunctionScope methodsFor: 'env' stamp: 'ajh 6/25/2004 17:11'!
thisContextVar

	^ thisContextVar! !

!FunctionScope methodsFor: 'env' stamp: 'ajh 6/25/2004 17:56'!
thisEnvVar

	^ thisEnvVar! !


!FunctionScope methodsFor: 'captured vars' stamp: 'ajh 7/8/2004 14:36'!
addCaptured: name

	self forceThisEnv.
	^ capturedVars add: (CapturedVar new
			name: name;
			index: capturedVars size + 1;
			scope: self)! !

!FunctionScope methodsFor: 'captured vars' stamp: 'ajh 6/25/2004 22:24'!
capturedVars

	^ capturedVars asSortedCollection: [:x :y | x index <= y index]! !

!FunctionScope methodsFor: 'captured vars' stamp: 'ajh 7/8/2004 14:38'!
moveToCaptured: tempVar
	"tempVar has been captured by an inner scope, move it from temps to captured. If tempVar is an arg, just copy it"

	| var |
	self forceThisEnv.
	tempVar isArg ifTrue: [
		var _ self addCaptured: tempVar name.
		var privUsage: tempVar privUsage. 
		var sourceTemp: tempVar.
	] ifFalse: [
		self removeTemp: tempVar.
		"tempVar becomes a captured var so parse tree var nodes will keep pointer to its var"
		var _ tempVar becomeCaptured.
		capturedVars add: var.
		var index: capturedVars size.
	].
	^ var! !


!FunctionScope methodsFor: 'parent env' stamp: 'ajh 7/8/2004 16:01'!
addSlotForCapturedParentEnv

	| receiver |
	self hasInnerFreeVars ifFalse: [^ false].
	receiver _ self receiverVar.
	capturedVars do: [:var | var sourceTemp = receiver ifTrue: [^ false]].
	^ true! !

!FunctionScope methodsFor: 'parent env' stamp: 'ajh 7/8/2004 16:00'!
capturedParentEnv

	| receiver |
	self hasInnerFreeVars ifFalse: [^ nil].
	receiver _ self receiverVar.
	capturedVars do: [:var | var sourceTemp = receiver ifTrue: [^ var]].
	"not added to captured list, but slot is created for it in emitCreateEnv.
	 It's not added since it was not captured directly by a variable reference in the code (only indirectly)"
	^ CapturedVar new
		name: 'env parent';
		index: capturedVars size + 1;
		scope: self! !

!FunctionScope methodsFor: 'parent env' stamp: 'ajh 7/8/2004 14:26'!
closestEnvScope
	"If self has its own env return self, otherwise return first outer scope with its own env"

	self hasEscapingEnv ifTrue: [^ self].
	^ self outerScope closestEnvScope! !

!FunctionScope methodsFor: 'parent env' stamp: 'ajh 7/8/2004 15:38'!
hasInnerFreeVars
	"True if has inner scope that references to outer scope (non-global) variables. Indicates whether or not to add parent ref to local env"

	^ hasInnerFreeVars = true! !

!FunctionScope methodsFor: 'parent env' stamp: 'ajh 7/8/2004 15:37'!
hasInnerFreeVars: bool
	"True if has inner scope that references to outer scope (non-global) variables. Indicates whether or not to add parent ref to local env"

	hasInnerFreeVars _ bool! !

!FunctionScope methodsFor: 'parent env' stamp: 'ajh 7/8/2004 14:26'!
outerEnvScope

	^ outerScope closestEnvScope! !


!FunctionScope methodsFor: 'emitting' stamp: 'ajh 7/8/2004 16:02'!
emitCreateEnv: methodBuilder
	"Create a heap environment for my captured vars"

	methodBuilder
		pushLiteral: ClosureEnvironment;
		pushLiteral: capturedVars size + self addSlotForCapturedParentEnv asBit;
		send: #new:.
	self thisEnvVar emitLocalStore: methodBuilder.
	methodBuilder popTop.
! !

!FunctionScope methodsFor: 'emitting' stamp: 'ajh 7/8/2004 17:59'!
emitEnvParentEnv: methodBuilder

	methodBuilder getField: self capturedParentEnv index.

	"return which scope my parent env points to"
	^ self outerEnvScope! !

!FunctionScope methodsFor: 'emitting' stamp: 'ajh 7/8/2004 15:40'!
emitLocalParentEnv: methodBuilder
	"local parent is held in receiver"

	self receiverVar emitLocalValue: methodBuilder.

	"return which scope my receiver points to"
	^ self outerEnvScope! !

!FunctionScope methodsFor: 'emitting' stamp: 'ajh 7/8/2004 18:38'!
emitMoveCapturedRargs: methodBuilder

	self capturedVars do: [:var |
		| sourceTemp |
		sourceTemp _ var sourceTemp.
		sourceTemp ifNotNil: [
			sourceTemp emitLocalValue: methodBuilder.
			self thisEnvVar emitLocalValue: methodBuilder.
			methodBuilder setField: var index; popTop]].

	self addSlotForCapturedParentEnv ifTrue: [
		self receiverVar emitLocalValue: methodBuilder.
		self thisEnvVar emitLocalValue: methodBuilder.
		methodBuilder setField: capturedVars size + 1; popTop].
! !

!FunctionScope methodsFor: 'emitting' stamp: 'ajh 7/8/2004 17:12'!
emitMyEnv: methodBuilder

	thisEnvVar emitLocalValue: methodBuilder! !

!FunctionScope methodsFor: 'emitting' stamp: 'ajh 6/26/2004 17:31'!
emitPrologue: methodBuilder

	self hasEscapingEnv ifFalse: [^ self].
	self emitCreateEnv: methodBuilder.
	self emitMoveCapturedRargs: methodBuilder.
! !


FunctionScope subclass: #ContextDoItScope
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Semantics'!

!ContextDoItScope methodsFor: 'initializing' stamp: 'ajh 7/8/2004 19:53'!
hideOuterTemps
	"Hide outer temps so they don't get captured (becuase they weren't captured and their values are not in the do-it env chain)"

	self outerScope tempsAreNoLongerAvailable! !


!ContextDoItScope methodsFor: 'lookup' stamp: 'ajh 7/8/2004 17:06'!
captureVar: name
	"Don't convert temps to captured"

	^ self lookupVar: name! !


!ContextDoItScope methodsFor: 'parent env' stamp: 'ajh 7/8/2004 17:56'!
closestEnvScope
	"my temps are like inst vars, hence my own environment"

	^ self! !


!ContextDoItScope methodsFor: 'emitting' stamp: 'ajh 7/8/2004 17:36'!
emitEnvParentEnv: methodBuilder

	methodBuilder getField: (MethodContext allInstVarNames indexOf: #receiver).

	"return which scope my receiver points to"
	^ self outerEnvScope! !

!ContextDoItScope methodsFor: 'emitting' stamp: 'ajh 7/8/2004 19:30'!
emitMyEnv: methodBuilder

	self flag: #myEnv.
	methodBuilder getField: MethodContext myEnvFieldIndex.! !


!ContextDoItScope methodsFor: 'scope' stamp: 'ajh 7/8/2004 17:15'!
isDoItScope

	^ true! !


LexicalScope subclass: #GlobalScope
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Semantics'!
!GlobalScope commentStamp: 'ajh 3/24/2003 21:52' prior: 0!
When the class is not known, use me to at least find global vars.!


!GlobalScope methodsFor: 'lookup' stamp: 'ajh 2/26/2003 13:42'!
classEncoding
	"subclass responsibility, default is nil"

	^ nil! !

!GlobalScope methodsFor: 'lookup' stamp: 'ajh 6/30/2004 14:18'!
lookupVar: name
	"subclass responsibility, default is to look up global vars"

	ProtoObject scopeHas: name asSymbol ifTrue:
		[:assoc | ^ GlobalVar new assoc: assoc; scope: self].
	^ nil
! !

!GlobalScope methodsFor: 'lookup' stamp: 'ajh 3/12/2003 12:10'!
possibleVarsFor: name continued: listOrNil
	"subclass responsibility, default is to look up global vars"

	name first isLowercase ifTrue: [^ listOrNil].
	^ ProtoObject possibleVariablesFor: name continuedFrom: listOrNil! !

!GlobalScope methodsFor: 'lookup' stamp: 'ajh 6/30/2004 14:18'!
rawVar: name
	"subclass responsibility, default is to look up global vars"

	ProtoObject scopeHas: name asSymbol ifTrue:
		[:assoc | ^ GlobalVar new assoc: assoc; scope: self].
	^ nil
! !


LexicalScope subclass: #InstanceScope
	instanceVariableNames: 'vars'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Semantics'!

!InstanceScope methodsFor: 'initializing' stamp: 'ajh 7/8/2004 20:26'!
isInstanceScope

	^ true! !

!InstanceScope methodsFor: 'initializing' stamp: 'ajh 7/8/2004 18:16'!
tempsAreNoLongerAvailable
	"done"
! !

!InstanceScope methodsFor: 'initializing' stamp: 'ajh 6/24/2004 04:10'!
vars: names

	self vars: names offset: 0! !

!InstanceScope methodsFor: 'initializing' stamp: 'ajh 7/8/2004 16:22'!
vars: names offset: k

	vars _ Dictionary new.
	names withIndexDo: [:name :index |
		vars at: name put:
			(CapturedVar new
				name: name;
				index: k + index;
				scope: self)
				markGiven].
! !


!InstanceScope methodsFor: 'lookup' stamp: 'ajh 7/8/2004 14:55'!
closestEnvScope
	"If self has its own env return self, otherwise return first outer scope with its own env"

	^ self! !

!InstanceScope methodsFor: 'lookup' stamp: 'ajh 6/28/2004 16:36'!
hasEscapingEnv

	^ true! !

!InstanceScope methodsFor: 'lookup' stamp: 'ajh 6/24/2004 03:02'!
lookupVar: name
	"Return a SemVar for my pool var with this name.  Return nil if none found"

	^ vars at: name ifAbsent: [self outerScope lookupVar: name]! !

!InstanceScope methodsFor: 'lookup' stamp: 'ajh 6/24/2004 03:03'!
possibleVarsFor: name continued: listOrNil
	"Return my var names that are close in spelling to name"

	| list |
	list _ name first isLowercase
		ifTrue: [name correctAgainstDictionary: vars continuedFrom: listOrNil]
		ifFalse: [listOrNil].
	^ self outerScope possibleVarsFor: name continued: list! !

!InstanceScope methodsFor: 'lookup' stamp: 'ajh 6/24/2004 03:04'!
rawVar: name
	"Return a ScopeVar for my inst var with this name.  Return nil if none found"

	^ vars at: name ifAbsent: [self outerScope rawVar: name]! !


Object subclass: #ScopeVar
	instanceVariableNames: 'scope usage'
	classVariableNames: 'UsageAutomaton'
	poolDictionaries: ''
	category: 'Compiler-Semantics'!
!ScopeVar commentStamp: 'ajh 3/24/2003 22:38' prior: 0!
I am an entry in a SemScope that gets associated with variable nodes of the same name.  There a three different subclasses of vars: temp vars, field vars, and pool/global vars.!


!ScopeVar methodsFor: 'initializing' stamp: 'ajh 7/8/2004 16:17'!
scope: aLexicalScope

	scope _ aLexicalScope! !


!ScopeVar methodsFor: 'accessing' stamp: 'ajh 3/16/2003 20:08'!
asString

	^ self name! !

!ScopeVar methodsFor: 'accessing' stamp: 'ajh 6/23/2004 22:44'!
isCaptured

	^ false! !

!ScopeVar methodsFor: 'accessing' stamp: 'ajh 7/8/2004 16:25'!
isGlobal

	^ false! !

!ScopeVar methodsFor: 'accessing' stamp: 'ajh 3/18/2003 11:13'!
isTemp

	^ false! !

!ScopeVar methodsFor: 'accessing' stamp: 'ajh 6/23/2004 22:47'!
name

	^ self subclassResponsibility! !

!ScopeVar methodsFor: 'accessing' stamp: 'ajh 2/27/2003 00:38'!
printOn: stream

	stream nextPutAll: self name! !

!ScopeVar methodsFor: 'accessing' stamp: 'ajh 6/23/2004 22:52'!
scope

	^ scope! !


!ScopeVar methodsFor: 'emitting' stamp: 'ajh 7/8/2004 10:47'!
emitStore: methodBuilder from: refScope

	self subclassResponsibility! !

!ScopeVar methodsFor: 'emitting' stamp: 'ajh 7/8/2004 10:47'!
emitValue: methodBuilder from: refScope

	self subclassResponsibility! !


!ScopeVar methodsFor: 'read/write usage' stamp: 'ajh 6/25/2004 11:58'!
isUndefined

	^ false! !

!ScopeVar methodsFor: 'read/write usage' stamp: 'ajh 7/8/2004 16:18'!
isUnused

	^ usage isNil! !

!ScopeVar methodsFor: 'read/write usage' stamp: 'ajh 7/8/2004 16:18'!
markGiven

	usage ifNil: [usage _ #arg].! !

!ScopeVar methodsFor: 'read/write usage' stamp: 'ajh 7/8/2004 16:19'!
markRead

	(usage isNil or: [usage = #arg]) ifTrue: [usage _ #read]! !

!ScopeVar methodsFor: 'read/write usage' stamp: 'ajh 7/8/2004 16:19'!
markWrite

	usage _ #write! !

!ScopeVar methodsFor: 'read/write usage' stamp: 'ajh 6/28/2004 10:13'!
privUsage

	^ usage! !

!ScopeVar methodsFor: 'read/write usage' stamp: 'ajh 6/28/2004 10:13'!
privUsage: usageSymbol

	usage _ usageSymbol! !


ScopeVar subclass: #GlobalVar
	instanceVariableNames: 'assoc'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Semantics'!
!GlobalVar commentStamp: 'ajh 3/24/2003 21:53' prior: 0!
I am a global, pool, or class variable.!


!GlobalVar methodsFor: 'initializing' stamp: 'ajh 2/26/2003 17:31'!
assoc: anAssociation

	assoc _ anAssociation! !


!GlobalVar methodsFor: 'accessing' stamp: 'ajh 7/8/2004 16:25'!
isGlobal

	^ true! !

!GlobalVar methodsFor: 'accessing' stamp: 'ajh 7/2/2004 14:15'!
name

	^ assoc name! !


!GlobalVar methodsFor: 'emitting' stamp: 'ajh 7/8/2004 10:47'!
emitStore: methodBuilder from: refScope

	methodBuilder pushLiteral: assoc.
	methodBuilder send: #privSetInHolder:.
! !

!GlobalVar methodsFor: 'emitting' stamp: 'ajh 7/8/2004 10:48'!
emitValue: methodBuilder from: refScope

	methodBuilder pushLiteral: assoc.
	methodBuilder send: #value.
! !


ScopeVar subclass: #LocalVar
	instanceVariableNames: 'name index'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Semantics'!
!LocalVar commentStamp: 'ajh 3/24/2003 21:54' prior: 0!
I am an argument or temporary variable of a method or block.!


!LocalVar methodsFor: 'initializing' stamp: 'ajh 6/23/2004 22:43'!
index: n

	index _ n! !

!LocalVar methodsFor: 'initializing' stamp: 'ajh 6/23/2004 22:47'!
name: string

	name _ string! !


!LocalVar methodsFor: 'accessing' stamp: 'ajh 3/18/2003 11:30'!
index

	^ index! !

!LocalVar methodsFor: 'accessing' stamp: 'ajh 6/23/2004 22:51'!
name

	^ name! !


!LocalVar methodsFor: 'read/write usage' stamp: 'ajh 7/8/2004 16:21'!
isUndefined

	^ usage isNil! !

!LocalVar methodsFor: 'read/write usage' stamp: 'ajh 6/28/2004 10:10'!
markArg
	"mark given"

	usage _ #arg! !


!LocalVar methodsFor: 'emitting' stamp: 'ajh 7/8/2004 15:07'!
emitEnvStore: methodBuilder from: refScope

	| nextParent |
	refScope = self scope ifTrue: [^ self emitEnvStore: methodBuilder].

	"self is being referenced from an inner scope (refScope), emit its parent env and recurse"
	nextParent _ refScope emitEnvParentEnv: methodBuilder.
	self emitEnvStore: methodBuilder from: nextParent.
! !

!LocalVar methodsFor: 'emitting' stamp: 'ajh 7/8/2004 15:07'!
emitEnvValue: methodBuilder from: refScope

	| nextParent |
	refScope = self scope ifTrue: [^ self emitEnvValue: methodBuilder].

	"self is being referenced from an inner scope (refScope), emit its parent env and recurse"
	nextParent _ refScope emitEnvParentEnv: methodBuilder.
	self emitEnvValue: methodBuilder from: nextParent.
! !

!LocalVar methodsFor: 'emitting' stamp: 'ajh 7/8/2004 15:08'!
emitStore: methodBuilder from: refScope

	| nextParent |
	refScope = self scope ifTrue: [^ self emitLocalStore: methodBuilder].

	"self is being referenced from an inner scope (refScope), emit its parent env and recurse"
	nextParent _ refScope emitLocalParentEnv: methodBuilder.
	self emitEnvStore: methodBuilder from: nextParent.
! !

!LocalVar methodsFor: 'emitting' stamp: 'ajh 7/8/2004 15:08'!
emitValue: methodBuilder from: refScope

	| nextParent |
	refScope = self scope ifTrue: [^ self emitLocalValue: methodBuilder].

	"self is being referenced from an inner scope (refScope), emit its parent env and recurse"
	nextParent _ refScope emitLocalParentEnv: methodBuilder.
	self emitEnvValue: methodBuilder from: nextParent.
! !


LocalVar subclass: #CapturedVar
	instanceVariableNames: 'sourceTemp'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Semantics'!
!CapturedVar commentStamp: 'ajh 3/24/2003 21:52' prior: 0!
I'm a field var of some receiverVar.  if receiverVar is 'self' than I am an instance variable.  If receiverVar is a closure than I am a free variable (outer temp reference) and my outerVar is the outer reference var.!


!CapturedVar methodsFor: 'initializing' stamp: 'ajh 6/25/2004 22:18'!
sourceTemp: tempVarOrNil

	sourceTemp _ tempVarOrNil! !


!CapturedVar methodsFor: 'accessing' stamp: 'ajh 6/24/2004 15:15'!
isCaptured

	^ true! !

!CapturedVar methodsFor: 'accessing' stamp: 'ajh 6/25/2004 22:18'!
sourceTemp

	^ sourceTemp! !


!CapturedVar methodsFor: 'emitting' stamp: 'ajh 7/8/2004 17:16'!
emitEnvStore: methodBuilder

	self scope isDoItScope ifTrue: [self scope emitMyEnv: methodBuilder].
	methodBuilder setField: self index! !

!CapturedVar methodsFor: 'emitting' stamp: 'ajh 7/8/2004 17:16'!
emitEnvValue: methodBuilder

	self scope isDoItScope ifTrue: [self scope emitMyEnv: methodBuilder].
	methodBuilder getField: self index.
! !

!CapturedVar methodsFor: 'emitting' stamp: 'ajh 7/8/2004 14:08'!
emitLocalStore: methodBuilder

	self scope emitMyEnv: methodBuilder.
	methodBuilder setField: self index.
! !

!CapturedVar methodsFor: 'emitting' stamp: 'ajh 7/8/2004 14:09'!
emitLocalValue: methodBuilder

	self scope emitMyEnv: methodBuilder.
	methodBuilder getField: self index.
! !


Notification subclass: #SemanticWarning
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Semantics'!
!SemanticWarning commentStamp: 'ajh 7/7/2004 17:36' prior: 0!
If a variable is not found or some other problem occurs during checking (ASTChecker) than a particular subclass instance of me is signal, usually causing a notification to the user. If not handled, the default handling is done, and compiling continues.!


!SemanticWarning methodsFor: 'as yet unclassified' stamp: 'ajh 3/12/2003 14:28'!
correctIn: compiler

	self subclassResponsibility! !

!SemanticWarning methodsFor: 'as yet unclassified' stamp: 'ajh 3/11/2003 14:54'!
isResumable

	^ true! !


SemanticWarning subclass: #ShadowVariableWarning
	instanceVariableNames: 'variableNode shadowedVar'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Semantics'!
!ShadowVariableWarning commentStamp: 'ajh 3/24/2003 22:21' prior: 0!
I get signalled when a variable in a block or method scope shadows a variable of the same name in an outer scope.  The default action is to allow it.!


!ShadowVariableWarning methodsFor: 'as yet unclassified' stamp: 'ajh 3/19/2003 16:42'!
correctIn: compiler

	self resume: (compiler variable: self variableNode shadows: self shadowedVar)! !

!ShadowVariableWarning methodsFor: 'as yet unclassified' stamp: 'ajh 3/19/2003 13:34'!
defaultAction
	"allow variable to shadow previous/inst var of the same name"

	^ nil! !

!ShadowVariableWarning methodsFor: 'as yet unclassified' stamp: 'ajh 3/19/2003 13:32'!
shadowedVar

	^ shadowedVar! !

!ShadowVariableWarning methodsFor: 'as yet unclassified' stamp: 'ajh 3/19/2003 13:32'!
shadowedVar: semVar

	shadowedVar _ semVar! !

!ShadowVariableWarning methodsFor: 'as yet unclassified' stamp: 'ajh 3/19/2003 13:19'!
variableNode

	^ variableNode! !

!ShadowVariableWarning methodsFor: 'as yet unclassified' stamp: 'ajh 3/19/2003 13:24'!
variableNode: aVariableNode

	variableNode _ aVariableNode.
	messageText _ 'Temp shadows: ', aVariableNode name.
! !


LocalVar subclass: #TempVar
	instanceVariableNames: 'isArg'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Semantics'!

!TempVar methodsFor: 'initializing' stamp: 'ajh 6/29/2004 17:05'!
becomeCaptured
	"TempVar has to become a captured var so parse tree var nodes will point to the captured var. Use change-class instead of become for speed. In order to do this both TempVar and CapturedVar has to have the same number of inst vars"

	isArg "sourceTemp" _ nil.
	self primitiveChangeClassTo: CapturedVar basicNew.
! !


!TempVar methodsFor: 'accessing' stamp: 'ajh 6/23/2004 23:18'!
isTemp

	^ true! !


!TempVar methodsFor: 'emitting' stamp: 'ajh 7/8/2004 17:40'!
emitEnvStore: methodBuilder

	self scope isDoItScope ifFalse: [self error: 'should have been captured'].
	methodBuilder setField: (MethodContext instSize + self index).
! !

!TempVar methodsFor: 'emitting' stamp: 'ajh 7/8/2004 17:40'!
emitEnvValue: methodBuilder

	self scope isDoItScope ifFalse: [self error: 'should have been captured'].
	methodBuilder getField: (MethodContext instSize + self index).
! !

!TempVar methodsFor: 'emitting' stamp: 'ajh 7/8/2004 14:10'!
emitLocalStore: methodBuilder

	methodBuilder storeTemp: self.
! !

!TempVar methodsFor: 'emitting' stamp: 'ajh 7/8/2004 14:10'!
emitLocalValue: methodBuilder

	methodBuilder pushTemp: self.
! !


!TempVar methodsFor: 'read/write usage' stamp: 'ajh 6/25/2004 22:11'!
isArg

	^ isArg = true! !

!TempVar methodsFor: 'read/write usage' stamp: 'ajh 6/25/2004 22:11'!
markArg

	isArg _ true.
	usage _ #arg! !


LocalVar subclass: #ThisContextVar
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Semantics'!

!ThisContextVar methodsFor: 'accessing' stamp: 'ajh 6/25/2004 16:52'!
isUndefined

	^ false! !


!ThisContextVar methodsFor: 'emitting' stamp: 'ajh 7/8/2004 14:12'!
emitLocalValue: methodBuilder

	methodBuilder pushThisContext.
! !


LocalVar subclass: #ThisEnvVar
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Semantics'!

!ThisEnvVar methodsFor: 'accessing' stamp: 'ajh 6/25/2004 17:54'!
isUndefined

	^ false! !


!ThisEnvVar methodsFor: 'emitting' stamp: 'ajh 7/8/2004 14:13'!
emitEnvValue: methodBuilder

	"do nothing, env is on top"
! !

!ThisEnvVar methodsFor: 'emitting' stamp: 'ajh 7/8/2004 14:14'!
emitLocalStore: methodBuilder

	methodBuilder storeThisEnv.
! !

!ThisEnvVar methodsFor: 'emitting' stamp: 'ajh 7/8/2004 14:14'!
emitLocalValue: methodBuilder

	methodBuilder pushThisEnv.
! !


SemanticWarning subclass: #UndeclaredSelectorWarning
	instanceVariableNames: 'messageNode'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Semantics'!
!UndeclaredSelectorWarning commentStamp: 'ajh 3/24/2003 22:23' prior: 0!
I get signalled when no selector is defined for a message send, indicating a possible mispelling.  My default action is to create a new selector.!


!UndeclaredSelectorWarning methodsFor: 'as yet unclassified' stamp: 'ajh 3/12/2003 14:28'!
correctIn: compiler

	self resume: (compiler correctSelector: self messageNode)! !

!UndeclaredSelectorWarning methodsFor: 'as yet unclassified' stamp: 'ajh 3/24/2003 22:30'!
defaultAction
	"create new selector"

	^ messageNode selectorString asSymbol! !

!UndeclaredSelectorWarning methodsFor: 'as yet unclassified' stamp: 'ajh 3/11/2003 22:43'!
messageNode

	^ messageNode! !

!UndeclaredSelectorWarning methodsFor: 'as yet unclassified' stamp: 'ajh 3/11/2003 23:37'!
messageNode: aMessageNode

	messageNode _ aMessageNode.
	messageText _ 'Unknown selector: ', aMessageNode selectorString.
! !


SemanticWarning subclass: #UndeclaredVariableWarning
	instanceVariableNames: 'variableNode'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Semantics'!
!UndeclaredVariableWarning commentStamp: 'ajh 3/24/2003 22:33' prior: 0!
I get signalled when an unknown variable it references.  My default action is to raise an warning.!


!UndeclaredVariableWarning methodsFor: 'as yet unclassified' stamp: 'ajh 3/12/2003 14:28'!
correctIn: compiler

	self resume: (compiler correctVariable: self variableNode)! !

!UndeclaredVariableWarning methodsFor: 'as yet unclassified' stamp: 'ajh 3/11/2003 12:05'!
variableNode

	^ variableNode! !

!UndeclaredVariableWarning methodsFor: 'as yet unclassified' stamp: 'ajh 3/11/2003 23:37'!
variableNode: aVariableNode

	variableNode _ aVariableNode.
	messageText _ 'Undeclared variable: ', aVariableNode name.
! !


SemanticWarning subclass: #UninitializedVariableWarning
	instanceVariableNames: 'variableNode'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Semantics'!
!UninitializedVariableWarning commentStamp: 'ajh 3/24/2003 22:34' prior: 0!
I get signalled when a temporary variable is used before it is assigned to.  My default action is to allow it since all temps get initialized to nil.!


!UninitializedVariableWarning methodsFor: 'as yet unclassified' stamp: 'ajh 3/12/2003 14:28'!
correctIn: compiler

	self resume: (compiler queryUninitializedTemp: self variableNode)! !

!UninitializedVariableWarning methodsFor: 'as yet unclassified' stamp: 'ajh 3/12/2003 13:37'!
defaultAction
	"unitialized temp will be initialized to nil, so ignore this warning if not handled"

	^ nil! !

!UninitializedVariableWarning methodsFor: 'as yet unclassified' stamp: 'ajh 3/12/2003 13:25'!
variableNode

	^ variableNode! !

!UninitializedVariableWarning methodsFor: 'as yet unclassified' stamp: 'ajh 3/12/2003 14:08'!
variableNode: aVariableNode

	variableNode _ aVariableNode.
	messageText _ 'Uninitialized temp: ', aVariableNode name.
! !


SemanticWarning subclass: #UnusedVariableWarning
	instanceVariableNames: 'variableNode'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Semantics'!
!UnusedVariableWarning commentStamp: 'ajh 3/24/2003 22:35' prior: 0!
I get signalled when a temporary variable is declared but never used.  My default action is to allow it since it is benign.!


!UnusedVariableWarning methodsFor: 'as yet unclassified' stamp: 'ajh 3/12/2003 14:28'!
correctIn: compiler

	self resume: (compiler removeUnusedTemp: self variableNode)! !

!UnusedVariableWarning methodsFor: 'as yet unclassified' stamp: 'ajh 3/12/2003 14:07'!
defaultAction
	"unused temp will be ignored, so ignore this warning if not handled"

	^ nil! !

!UnusedVariableWarning methodsFor: 'as yet unclassified' stamp: 'ajh 3/12/2003 14:06'!
variableNode

	^ variableNode! !

!UnusedVariableWarning methodsFor: 'as yet unclassified' stamp: 'ajh 3/12/2003 14:07'!
variableNode: aVariableNode

	variableNode _ aVariableNode.
	messageText _ 'Unused temp: ', aVariableNode name.
! !
