Object subclass: #Parser2
	instanceVariableNames: 'source requestor scope doitFlag failBlock endOfMethodPattern'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Syntax'!
!Parser2 commentStamp: 'ajh 3/24/2003 21:36' prior: 0!
This mirrors the old Parser protocol but delegates to the real parser which is SqueakParser.!


!Parser2 methodsFor: 'public access' stamp: 'ajh 3/11/2003 14:35'!
parse: sourceStreamOrString class: behavior

	^ self parse: sourceStreamOrString readStream class: behavior
		noPattern: false notifying: nil ifFail: [nil]! !

!Parser2 methodsFor: 'public access' stamp: 'ajh 9/14/2002 09:28'!
parse: sourceStream class: behavior noPattern: noPattern context: context notifying: req ifFail: aBlock
	"Backwards compatibilty"

	context ifNotNil: [self error: 'Use Compiler-evaluate... instead'].
	^ self parse: sourceStream class: behavior noPattern: noPattern notifying: req ifFail: aBlock! !

!Parser2 methodsFor: 'public access' stamp: 'ajh 7/7/2004 17:37'!
parse: sourceStream class: parseScope noPattern: doitBool notifying: req ifFail: aBlock
	"Parse sourceStream into a embedded BlockNode if doitFlag is true (no method header) or a MethodNode if doitFlag is false.  Parsing is done with respect to parseScope to find non-local variables.  Errors in parsing are reported to req if not nil followed by executing the fail block."

	| parser |
	source _ sourceStream.
	requestor _ req.
	doitFlag _ doitBool.
	scope _ parseScope parseScope.
	failBlock _ [^ aBlock value].

	parser _ (req notNil and: [RequestAlternateSyntaxSetting signal and: [(sourceStream isKindOf: FileStream) not]])
		ifTrue: ["Squeak00Parser" self toDo]
		ifFalse: [SqueakParser].

	^ [	| tree |
		tree _ doitFlag 
			ifTrue: [parser parseDoIt: source]
			ifFalse: [parser parseMethod: source].
		[tree verifyIn: scope] on: SemanticWarning do: [:ex | ex correctIn: self]
	 ] on: UnhandledError do: [:uEx | | ex |
		ex _ uEx exception.
		((SmaCCParserError handles: ex) or: [PrimitiveParseError handles: ex])
			ifTrue: [self notify: ex description at: ex tag position]
			ifFalse: [uEx pass]
	 ]! !

!Parser2 methodsFor: 'public access' stamp: 'ajh 3/6/2003 00:42'!
parseArgsAndTemps: aString notifying: req 
	"No initialization required.  Parse the argument, aString, notifying req if an error occurs. Otherwise, answer a two-element Array containing Arrays of strings (the argument names and temporary variable names)."

	aString == nil ifTrue: [^ #()].

	^ (self parse: aString readStream class: nil noPattern: false notifying: req ifFail: [^ nil]) tempNames! !

!Parser2 methodsFor: 'public access' stamp: 'ajh 3/11/2003 19:59'!
parseMethodComment: aString setPattern: aBlock
	"Answer the method comment for the argument, aString. Evaluate aBlock with the methodNode containing the selector and arguments"

	| methodNode comments |
	methodNode _ SqueakParser parseMethodPattern: aString.
	endOfMethodPattern _ methodNode methodPatternStop.
	aBlock value: methodNode.
	comments _ methodNode comments.
	comments isEmpty ifTrue: [^ nil].
	^ comments first! !

!Parser2 methodsFor: 'public access' stamp: 'ajh 3/11/2003 20:00'!
parseSelector: aString 
	"Answer the message selector for the argument, aString, which should parse successfully up to the temporary declaration or the end of the method header."

	| methodNode |
	methodNode _ SqueakParser parseMethodPattern: aString.
	endOfMethodPattern _ methodNode methodPatternStop.
	^ methodNode selector! !


!Parser2 methodsFor: 'error handling' stamp: 'ajh 3/6/2003 01:03'!
fail

	^ failBlock value! !

!Parser2 methodsFor: 'error handling' stamp: 'ajh 5/19/2004 17:31'!
interactive

	requestor == nil ifTrue: [^ false].
	(requestor isKindOf: SyntaxError) ifTrue: [^ false].
	(requestor respondsTo: #interactive) ifTrue: [^ requestor interactive].
	^ true! !

!Parser2 methodsFor: 'error handling' stamp: 'ajh 9/12/2002 01:10'!
notify: string at: location
	requestor isNil
		ifTrue: [SyntaxError 
					errorInClass: (self classEncoding ifNil: [FakeClassPool])
					withCode: 
						(source contents
							copyReplaceFrom: location
							to: location - 1
							with: string , ' ->')
					doitFlag: doitFlag]
		ifFalse: [requestor
					notify: string , ' ->'
					at: location
					in: source].
	^self fail! !


!Parser2 methodsFor: 'accessing' stamp: 'ajh 3/11/2003 12:16'!
classEncoding

	^ scope classEncoding! !

!Parser2 methodsFor: 'accessing' stamp: 'ajh 3/25/2003 23:45'!
encoder

	^ self! !

!Parser2 methodsFor: 'accessing' stamp: 'ajh 3/10/2003 22:17'!
requestor
	"Return the source code editor"

	^ requestor! !

!Parser2 methodsFor: 'accessing' stamp: 'ajh 6/9/2002 19:47'!
requestor: editor
	"set the source code editor"

	requestor _ editor! !


!Parser2 methodsFor: 'error correction' stamp: 'ajh 3/13/2003 04:22'!
correctSelector: messageNode
	"Correct the proposed selector in messageNode, correcting the original text if such action is indicated.  Fail if the proposedKeyword couldn't be converted into a valid selector.  Spots is an ordered collection of intervals within the test stream of the for each of the keyword parts."

	| alternatives aStream choice userSelection lines firstLine sel spots newSel |

	sel _ messageNode selectorString.
	"If we can't ask the user, assume that the keyword will be defined later"
	self interactive ifFalse: [^ sel asSymbol].

	spots _ messageNode selectorParts collect: [:token | token sourceInterval].
	userSelection _ requestor selectionInterval.
	requestor selectFrom: spots first first to: spots last last.
	requestor select.
	alternatives _ messageNode owningScope possibleSelectorsFor: sel.

	aStream _ WriteStream on: (String new: 200).
	aStream nextPutAll: (sel contractTo: 35); cr.
	firstLine _ 1.
 	alternatives do: [:s | aStream nextPutAll: s; cr].
	aStream nextPutAll: 'cancel'.
	lines _ Array with: firstLine with: (alternatives size + firstLine).
	
	choice _ (PopUpMenu labels: aStream contents lines: lines)
		startUpWithCaption: 
'Unknown selector, please 
confirm, correct, or cancel'.

	(choice = 0) | (choice > (lines at: 2)) ifTrue: [^ self fail].

	requestor deselect.
	requestor selectInvisiblyFrom: userSelection first to: userSelection last.

	choice = 1 ifTrue: [^ sel asSymbol].
	newSel _ alternatives at: choice - 1.
	self substituteSelector: newSel in: messageNode.
	((sel last ~~ $:) and: [newSel last == $:]) ifTrue: [^ self fail].
	^ (messageNode owningScope lookupSelector: newSel)
		ifNil: [self halt: 'chosen selector expected to exist']
! !

!Parser2 methodsFor: 'error correction' stamp: 'ajh 6/30/2004 14:20'!
correctVariable: variableNode
	"Correct the variableNode to a known variable, or declare it as a new
	variable if such action is requested.  We support declaring lowercase
	variables as temps, and uppercase variables as Globals or ClassVars,
	depending on whether the context is nil (class=UndefinedObject)."

	| varName spot alternatives aStream choice userSelection temp binding declareSize |

	varName _ variableNode name.
	"If we can't ask the user for correction, make it undeclared"
	self interactive ifFalse: [
		Transcript show: ' (' , varName , ' is Undeclared) '.
		varName _ varName asSymbol.
		Undeclared at: varName put: nil.
		^ GlobalVar new assoc: (Undeclared associationAt: varName); scope: scope
	].

	spot _ variableNode sourceInterval.
	temp _ varName first isLowercase.
	"First check to see if the requestor knows anything about the variable"
	(temp and: [(binding _ requestor bindingOf: varName) notNil])
		ifTrue: [^ GlobalVar new assoc: binding; scope: scope].
	userSelection _ requestor selectionInterval.
	requestor selectFrom: spot first to: spot last.
	requestor select.

	alternatives _ variableNode owningScope possibleVarsFor: varName.

	aStream _ WriteStream on: (String new: 200).
	temp ifTrue: [
		declareSize _ 1.
		aStream nextPutAll: 'declare temp'; cr.
	] ifFalse: [
		declareSize _ 1.
		self classEncoding ifNotNil: [
			declareSize _ declareSize + 1.  
			aStream nextPutAll: 'declare class variable'; cr].
		aStream nextPutAll: 'declare global'; cr.
	].
	alternatives do: [:name | aStream nextPutAll: name; cr].
	aStream nextPutAll: 'cancel'.

	choice _ (PopUpMenu labels: aStream contents
				lines: {declareSize. declareSize + alternatives size})
		startUpWithCaption:
(('Unknown variable: ', varName, '
please correct, or cancel:') asText makeBoldFrom: 19 to: 19 + varName size).

	(choice = 0 or: [choice > (declareSize + alternatives size)]) ifTrue: [^ self fail].

	requestor deselect.
	requestor selectInvisiblyFrom: userSelection first to: userSelection last.
	temp ifTrue: [
		choice = 1 ifTrue: [^ self declareTempAndPaste: variableNode]
	] ifFalse: [
		declareSize = 1 ifTrue: [
			choice = 1 ifTrue: [^ self declareGlobal: variableNode].
		] ifFalse: [ "declareSize = 2"
			choice = 1 ifTrue: [^ self declareClassVar: variableNode].
			choice = 2 ifTrue: [^ self declareGlobal: variableNode].
		].
	].

	"Spelling correction"
	varName _ alternatives at: choice - declareSize.
	variableNode changeToken: (SmaCCToken value: varName start: spot first).
	self substituteWord: varName wordInterval: spot offset: 0.
	^ (variableNode owningScope lookupVar: varName)
		ifNil: [self halt: 'var should have been found']! !

!Parser2 methodsFor: 'error correction' stamp: 'ajh 3/20/2003 01:30'!
declareClassVar: variableNode

	self classEncoding theNonMetaClass
		addClassVarName: variableNode name asSymbol.
	^ (variableNode owningScope lookupVar: variableNode name)
		ifNil: [self halt: 'should be found']! !

!Parser2 methodsFor: 'error correction' stamp: 'ajh 3/20/2003 01:30'!
declareGlobal: variableNode

	Smalltalk at: variableNode name asSymbol put: nil.
	^ (variableNode owningScope lookupVar: variableNode name)
		ifNil: [self halt: 'should be found']! !

!Parser2 methodsFor: 'error correction' stamp: 'ajh 3/20/2003 01:31'!
declareTempAndPaste: variableNode

	| name insertion theTextString methodBody pos start c var tempNode prev |

	name _ variableNode name.
	theTextString _ requestor text string.
	methodBody _ variableNode root body.
	methodBody temporaries notEmpty ifTrue: [
  		"Paste it after last temp"
		prev _ pos _ methodBody temporaries last stop + 1.
		(theTextString at: pos) isSeparator ifTrue: [
			insertion _ name, ' '.
			pos _ pos + 1.  start _ pos.
		] ifFalse: [
			insertion _ ' ', name.
			start _ pos + 1.
		].
		self substituteWord: insertion
			wordInterval: (pos to: pos - 1) offset: 0.
	] ifFalse: [
		"No bars - insert some with CR & tab"
		insertion _ '| ', name, ' |', String cr.
		prev _ pos _ methodBody statements first start - 1.
		start _ 0.
		[pos = 0 or: [(c _ theTextString at: pos) = Character cr]] whileFalse: [
			c isSeparator ifTrue: [
				insertion _ c asString, insertion.
				start _ start + 1.
			].
			pos _ pos - 1.
		].
		pos _ pos + 1.
		self substituteWord: insertion
			wordInterval: (pos to: pos - 1) offset: 0.
		start _ pos + 2 + start.
	].
	var _ methodBody owningScope addTemp: name.
	tempNode _ RBVariableNode new
		identifierToken: (SmaCCToken value: name start: start);
		binding: var.
	methodBody root adjustPositionsAfter: prev by: insertion size.
	methodBody temporaries: (methodBody temporaries copyWith: tempNode).
	^ (variableNode owningScope lookupVar: name)
		ifNil: [self halt: 'should have been found']! !

!Parser2 methodsFor: 'error correction' stamp: 'ajh 3/12/2003 14:34'!
queryUninitializedTemp: variableNode

	| varStart varName |
	self interactive ifFalse: [^ self].
	varName _ variableNode name.
	varStart _ variableNode start.
	requestor selectFrom: varStart to: varStart + varName size - 1; select.
	((PopUpMenu labels:
'yes
no') startUpWithCaption:
		((varName , ' appears to be
undefined at this point.
Proceed anyway?') asText makeBoldFrom: 1 to: varName size))
		= 1 ifFalse: [^ self fail]! !

!Parser2 methodsFor: 'error correction' stamp: 'ajh 3/15/2003 13:27'!
removeUnusedTemp: variableNode
	"Removing unused temp, variableNode, if verified by the user"

	| varName interval |
	self interactive ifFalse: [^ self].  "don't remove and continue"
	varName _ variableNode name.
	((PopUpMenu labels: 'yes\no' withCRs) startUpWithCaption:
		((varName , ' appears to be
unused in this method.
OK to remove it?') asText makeBoldFrom: 1 to: varName size))
			= 1 ifFalse: [^ self].  "don't remove and continue"

	"Remove temp then continue"
	interval _ variableNode sourceInterval.
	(requestor text at: interval last + 1) isSeparator ifTrue: [
		interval _ interval first to: interval last + 1].
	self substituteWord: '' wordInterval: interval offset: 0.
	variableNode root adjustPositionsAfter: interval first by: 0 - interval size.
	variableNode parent removeTemporaryNamed: varName.
	^ self! !

!Parser2 methodsFor: 'error correction' stamp: 'ajh 3/17/2003 13:07'!
substituteSelector: newSel in: messageNode
	"Substitute the correctSelector into the (presuamed interactive) receiver."

	| offset newParts |
	newParts _ OrderedCollection new.
	offset _ 0.
	newSel keywords
	  with: (messageNode selectorParts collect: [:tok | tok sourceInterval])
	  do: [:word :interval |
		newParts add: (SmaCCToken value: word start: interval first + offset).
		offset _ self substituteWord: word wordInterval: interval offset: offset.
	  ].
	messageNode changeSelectorParts: newParts asArray.
! !

!Parser2 methodsFor: 'error correction' stamp: 'ajh 3/11/2003 23:22'!
substituteWord: correctWord wordInterval: spot offset: o
	"Substitute the correctSelector into the (presuamed interactive) receiver."

	requestor
		correctFrom: (spot first + o)
		to: (spot last + o)
		with: correctWord.
	^ o + correctWord size - spot size! !

!Parser2 methodsFor: 'error correction' stamp: 'ajh 3/19/2003 16:43'!
variable: varNode shadows: semVar

	self interactive ifFalse: [
		^ Transcript show: '(', varNode name, ' is shadowed)'].
	self notify: 'Name already defined' at: varNode start.
! !


!Parser2 methodsFor: 'old protocol' stamp: 'ajh 3/11/2003 20:01'!
endOfLastToken

	^ endOfMethodPattern! !

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

Parser2 class
	instanceVariableNames: ''!

!Parser2 class methodsFor: 'as yet unclassified' stamp: 'ajh 7/7/2004 13:49'!
initialize

	Smalltalk at: #ParseError put: SmaCCParserError.
! !


Object subclass: #PrimitiveNode
	instanceVariableNames: 'primitiveNum spec'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Syntax'!
!PrimitiveNode commentStamp: 'ajh 3/24/2003 21:35' prior: 0!
I represent a primitive.  I am more than just a number if I am a named primitive.

Structure:

 num	<Integer>	Primitive number.
 spec	<Object>		Stored in first literal when num is 117 or 120.
!


!PrimitiveNode methodsFor: 'as yet unclassified' stamp: 'ajh 7/14/2001 12:37'!
num

	^ primitiveNum! !

!PrimitiveNode methodsFor: 'as yet unclassified' stamp: 'ajh 7/14/2001 12:23'!
num: n

	primitiveNum _ n! !

!PrimitiveNode methodsFor: 'as yet unclassified' stamp: 'ajh 3/12/2003 12:26'!
printOn: aStream

	aStream nextPutAll: 'primitive '; print: primitiveNum! !

!PrimitiveNode methodsFor: 'as yet unclassified' stamp: 'ajh 3/19/2003 22:06'!
printPrimitiveOn: aStream 
	"Print the primitive on aStream"

	| primIndex primDecl |
	primIndex _ primitiveNum.
	primIndex = 0 ifTrue: [^ self].
	primIndex = 120 ifTrue: [
		"External call spec"
		^ aStream print: spec].
	aStream nextPutAll: '<primitive: '.
	primIndex = 117 ifTrue: [
		primDecl _ spec.
		aStream nextPut: $';
			nextPutAll: (primDecl at: 2);
			nextPut: $'.
		(primDecl at: 1) ifNotNil: [
			aStream nextPutAll: ' module: ';
				nextPut: $';
				nextPutAll: (primDecl at: 1);
				nextPut: $'].
	] ifFalse: [aStream print: primIndex].
	aStream nextPut: $>.
	(primIndex ~= 117 and: [primIndex ~= 120]) ifTrue: [
		Smalltalk at: #Interpreter ifPresent: [:cls |
			aStream nextPutAll: ' "', 
				((cls classPool at: #PrimitiveTable) at: primIndex + 1) , '" '
		].
	].
! !

!PrimitiveNode methodsFor: 'as yet unclassified' stamp: 'ajh 3/19/2003 22:02'!
sourceText

	^ String streamContents: [:stream |
		self printPrimitiveOn: stream]! !

!PrimitiveNode methodsFor: 'as yet unclassified' stamp: 'ajh 7/14/2001 12:37'!
spec

	^ spec! !

!PrimitiveNode methodsFor: 'as yet unclassified' stamp: 'ajh 7/14/2001 12:30'!
spec: literal

	spec _ literal! !

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

PrimitiveNode class
	instanceVariableNames: ''!

!PrimitiveNode class methodsFor: 'as yet unclassified' stamp: 'ajh 7/14/2001 12:47'!
null

	^ self new num: 0! !


Error subclass: #PrimitiveParseError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Syntax'!
!PrimitiveParseError commentStamp: 'ajh 3/24/2003 18:21' prior: 0!
I get signalled if there is a parsing error while parsing the primitive text between angle brackets <...>!


!PrimitiveParseError methodsFor: 'as yet unclassified' stamp: 'ajh 3/11/2003 15:26'!
isResumable

	^ true! !


Scanner subclass: #PrimitiveParser
	instanceVariableNames: 'here hereType hereMark hereEnd prevMark prevEnd startPosition spec'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Syntax'!
!PrimitiveParser commentStamp: 'ajh 3/24/2003 18:23' prior: 0!
I parse the the primitive text between angle brackets.  I am basically a copy of the original Parser but with only the primitive parse methods retained.!


!PrimitiveParser methodsFor: 'public access' stamp: 'ajh 3/11/2003 11:32'!
parse: primitiveString

	| prim |
	super scan: primitiveString readStream.
	prevMark _ hereMark _ mark.
	self advance.

	prim _ self primitive.
	^ PrimitiveNode new
		num: prim;
		spec: spec! !

!PrimitiveParser methodsFor: 'public access' stamp: 'ajh 3/11/2003 18:34'!
startPosition: n
	"Needed only for error reporting"

	startPosition _ n! !


!PrimitiveParser methodsFor: 'scanning' stamp: 'ajh 3/11/2003 11:10'!
advance
	| this |
	prevMark _ hereMark.
	prevEnd _ hereEnd.
	this _ here.
	here _ token.
	hereType _ tokenType.
	hereMark _ mark.
	hereEnd _ source position - (source atEnd ifTrue: [hereChar == 30 asCharacter ifTrue: [0] ifFalse: [1]] ifFalse: [2]).
	self scanToken.
	"Transcript show: 'here: ', here printString, ' mark: ', hereMark printString, ' end: ', hereEnd printString; cr."
	^this! !

!PrimitiveParser methodsFor: 'scanning' stamp: 'ajh 3/11/2003 11:10'!
endOfLastToken

	^ prevEnd ifNil: [mark]! !

!PrimitiveParser methodsFor: 'scanning' stamp: 'ajh 3/11/2003 11:10'!
match: type 
	"Answer with true if next tokens type matches."

	hereType == type
		ifTrue: 
			[self advance.
			^true].
	^false! !

!PrimitiveParser methodsFor: 'scanning' stamp: 'ajh 3/11/2003 11:10'!
matchReturn

	^ self match: #upArrow! !

!PrimitiveParser methodsFor: 'scanning' stamp: 'ajh 3/11/2003 11:10'!
matchToken: thing 
	"Matches the token, not its type."

	here = thing ifTrue: [self advance. ^true].
	^false! !

!PrimitiveParser methodsFor: 'scanning' stamp: 'ajh 3/11/2003 11:10'!
startOfNextToken
	"Return starting position in source of next token."

	hereType == #doIt ifTrue: [^source position + 1].
	^hereMark! !


!PrimitiveParser methodsFor: 'primitives' stamp: 'ajh 3/11/2003 11:32'!
allocateLiteral: lit

	spec _ lit! !

!PrimitiveParser methodsFor: 'primitives' stamp: 'ajh 3/15/2003 16:28'!
externalFunctionDeclaration
	"Parse the function declaration for a call to an external library."
	| descriptorClass callType retType externalName args argType module fn |
	descriptorClass _ Smalltalk at: #ExternalFunction ifAbsent:[nil].
	descriptorClass == nil ifTrue:[^0].
	callType _ descriptorClass callingConventionFor: here.
	callType == nil ifTrue:[^0].
	"Parse return type"
	self advance.
	retType _ self externalType: descriptorClass.
	retType == nil ifTrue:[^self expected:'return type'].
	"Parse function name or index"
	externalName _ here.
	(self match: #string) 
		ifTrue:[externalName _ externalName asSymbol]
		ifFalse:[(self match:#number) ifFalse:[^self expected:'function name or index']].
	(self matchToken:'(' asSymbol) ifFalse:[^self expected:'argument list'].
	args _ WriteStream on: Array new.
	[here == #')'] whileFalse:[
		argType _ self externalType: descriptorClass.
		argType == nil ifTrue:[^self expected:'argument'].
		argType isVoid & argType isPointerType not ifFalse:[args nextPut: argType].
	].
	(self matchToken:')' asSymbol) ifFalse:[^self expected:')'].
	(self matchToken: 'module:') ifTrue:[
		module _ here.
		(self match: #string) ifFalse:[^self expected: 'String'].
		module _ module asSymbol].
	Smalltalk at: #ExternalLibraryFunction ifPresent:[:xfn|
		fn _ xfn name: externalName 
				module: module 
				callType: callType
				returnType: retType
				argumentTypes: args contents.
		self allocateLiteral: fn.
	].
	^120! !

!PrimitiveParser methodsFor: 'primitives' stamp: 'ajh 3/11/2003 11:10'!
externalType: descriptorClass
	"Parse an return an external type"
	| xType |
	xType _ descriptorClass atomicTypeNamed: here.
	xType == nil ifTrue:["Look up from class scope"
		Symbol hasInterned: here ifTrue:[:sym|
			xType _ descriptorClass structTypeNamed: sym]].
	xType == nil ifTrue:[
		"Raise an error if user is there"
		self interactive ifTrue:[^nil].
		"otherwise go over it silently"
		xType _ descriptorClass forceTypeNamed: here].
	self advance.
	(self matchToken:#*)
		ifTrue:[^xType asPointerType]
		ifFalse:[^xType]! !

!PrimitiveParser methodsFor: 'primitives' stamp: 'ajh 3/11/2003 11:10'!
primitive
	| n |
	(self matchToken: #<) ifFalse: [^ 0].
	n _ self primitiveDeclarations.
	(self matchToken: #>) ifFalse: [^ self expected: '>'].
	^ n! !

!PrimitiveParser methodsFor: 'primitives' stamp: 'ajh 3/11/2003 11:10'!
primitiveDeclarations
	| prim module |
	(self matchToken: 'primitive:') ifFalse:[^self externalFunctionDeclaration].
	prim _ here.
	(self match: #number) ifTrue:[^prim].	"Indexed primitives"
	(self match: #string) ifFalse:[^self expected:'Integer or String'].
	(self matchToken: 'module:') ifTrue:[
		module _ here.
		(self match: #string) ifFalse:[^self expected: 'String'].
		module _ module asSymbol].
	(self allocateLiteral: (Array with: module with: prim asSymbol with: 0 with: 0)).
	^117! !


!PrimitiveParser methodsFor: 'error handling' stamp: 'ajh 3/11/2003 15:04'!
expected: aString 
	"Notify a problem at token 'here'."

	tokenType == #doIt ifTrue: [hereMark _ hereMark + 1].
	hereType == #doIt ifTrue: [hereMark _ hereMark + 1].
	^ self notify: aString , ' expected'! !

!PrimitiveParser methodsFor: 'error handling' stamp: 'ajh 3/11/2003 18:36'!
localPosition

	^ hereMark! !

!PrimitiveParser methodsFor: 'error handling' stamp: 'ajh 3/11/2003 15:17'!
notify: aString
	"Notify problem at token at 'self position'."

	^ PrimitiveParseError new
		tag: self;
		signal: aString! !

!PrimitiveParser methodsFor: 'error handling' stamp: 'ajh 3/11/2003 18:37'!
position

	^ self startPosition + self localPosition - 1! !

!PrimitiveParser methodsFor: 'error handling' stamp: 'ajh 3/11/2003 18:34'!
startPosition

	^ startPosition! !


SmaCCParser subclass: #SqueakParser
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Syntax'!
!SqueakParser commentStamp: 'ajh 3/24/2003 18:24' prior: 0!
I parse Smalltalk text into a Refactory abstract syntax tree, whose root is RBMethodNode or RBDoItNode.  Methods under 'generated-*' categories were automatically generated using SmaCC.!


!SqueakParser methodsFor: 'generated-tables' stamp: 'ajh 7/19/2004 11:54'!
reduceTable
	^#(
#(33 0 #reduceFor:)
#(33 2 #reduceFor:)
#(33 2 #reduceFor:)
#(33 2 #reduceFor:)
#(33 2 #reduceFor:)
#(33 2 #reduceFor:)
#(33 2 #reduceFor:)
#(33 2 #reduceFor:)
#(33 2 #reduceFor:)
#(33 2 #reduceFor:)
#(33 2 #reduceFor:)
#(34 1 #array)
#(34 3 #second:)
#(35 1 #firstIn:)
#(35 3 #add3:)
#(36 0 #reduceActionForOptionalXXXperiodX1:)
#(36 1 #reduceActionForOptionalXXXperiodX2:)
#(37 1 #first:)
#(37 1 #first:)
#(37 1 #first:)
#(38 2 #secondIn:)
#(38 3 #add3:)
#(39 5 #blockWithArgs:)
#(39 3 #blockNoArgs:)
#(39 3 #blockArgs:)
#(39 7 #blockWithTemps:)
#(40 3 #assignment:)
#(41 1 #first:)
#(41 3 #cascade:)
#(42 3 #second:)
#(42 1 #first:)
#(42 1 #first:)
#(42 1 #first:)
#(42 1 #first:)
#(43 3 #array:)
#(44 1 #litTrue:)
#(44 1 #litFalse:)
#(44 1 #litNil:)
#(44 1 #litNumber:)
#(44 1 #litNumber:)
#(44 1 #litChar:)
#(44 1 #litString:)
#(44 2 #litStringSymbol:)
#(44 2 #litSymbol:)
#(44 2 #litSymbol:)
#(44 2 #litSymbol:)
#(44 2 #litSymbol:)
#(44 2 #litSymbol:)
#(44 4 #litArray:)
#(44 4 #litArray:)
#(45 3 #primitiveMessage:)
#(46 1 #first:)
#(46 1 #first:)
#(46 1 #first:)
#(47 2 #messagePart:)
#(47 3 #addMessagePart:)
#(48 2 #messageSend:)
#(48 2 #messageSend:)
#(48 2 #messageSend:)
#(49 2 #messageSend:)
#(49 2 #messageSend:)
#(49 2 #messageSend:)
#(50 2 #messageSend:)
#(50 2 #messageSend:)
#(51 1 #unaryMessage:)
#(52 2 #messagePart:)
#(53 2 #messagePart:)
#(53 3 #addMessagePart:)
#(54 1 #first:)
#(54 1 #first:)
#(54 1 #first:)
#(55 1 #first:)
#(55 1 #first:)
#(56 1 #variable:)
#(57 1 #unaryMessage:)
#(57 2 #messagePart:)
#(57 1 #first:)
#(58 1 #sequence:)
#(58 2 #sequenceWithTemps:)
#(59 1 #reduceFor:)
#(60 0 #byteStream)
#(60 2 #byteStreamPut:)
#(61 0 #stream)
#(61 2 #streamPut:)
#(62 1 #value:)
#(62 1 #valueSymbol:)
#(62 1 #valueSymbol:)
#(62 1 #valueSymbol:)
#(62 1 #valueSymbol:)
#(62 1 #valueSymbol:)
#(62 3 #contents2:)
#(62 3 #contents2:)
#(62 1 #valueSymbol:)
#(63 0 #array)
#(63 2 #first:)
#(63 5 #returnAdd:)
#(63 3 #return:)
#(66 1 #first:)
#(66 1 #first:)
#(66 1 #first:)
#(67 2 #method:)
#(67 3 #methodPrim:)
#(67 4 #methodTempsPrim:)
#(68 0 #array)
#(68 2 #add:)
)! !

!SqueakParser methodsFor: 'generated-tables' stamp: 'ajh 7/19/2004 11:54'!
transitionTable
	^#(
#(3 17 21 21 22 25 24 29 47 33 57 37 67)
#(3 41 1 45 3 49 4 53 8 57 9 61 10 65 11 69 12 73 13 77 15 81 18 85 19 89 20 93 21 97 29 101 34 105 35 109 37 113 39 117 40 121 41 125 42 129 43 133 44 137 46 141 48 145 49 149 50 153 56 157 58 161 63 378 64)
#(3 17 21 21 22 25 24 29 47 165 57)
#(2 302 1 3 4 6 8 9 10 11 12 13 15 18 19 20 21 29 64)
#(3 93 21 169 56)
#(3 93 21 173 56)
#(3 310 1 310 3 310 4 310 6 310 8 310 9 310 10 310 11 310 12 310 13 310 15 310 18 310 19 310 20 310 21 177 22 310 29 310 64)
#(3 41 1 45 3 49 4 181 6 53 8 57 9 61 10 65 11 69 12 73 13 77 15 81 18 85 19 89 20 93 21 97 29 185 34 105 35 109 37 113 39 117 40 121 41 125 42 129 43 133 44 189 45 137 46 141 48 145 49 149 50 153 56 193 58 161 63 378 64)
#(2 0 64)
#(3 41 1 378 2 45 3 49 4 197 5 53 8 57 9 61 10 65 11 69 12 73 13 77 15 81 18 85 19 89 20 93 21 97 29 101 34 105 35 109 37 201 38 113 39 117 40 121 41 125 42 129 43 133 44 137 46 141 48 145 49 149 50 153 56 205 58 161 63)
#(3 209 1 213 13 217 20 221 21 225 22 229 23 233 24 237 26)
#(3 41 1 45 3 61 10 65 11 69 12 73 13 77 15 81 18 85 19 89 20 93 21 97 29 241 37 113 39 117 40 121 41 125 42 129 43 133 44 137 46 141 48 145 49 149 50 153 56)
#(2 50 1 2 3 4 6 10 11 12 13 15 18 19 20 21 29 64)
#(3 418 9 418 21 245 68)
#(2 150 1 2 3 7 10 11 12 13 14 15 16 17 18 19 20 21 22 23 24 26 29 30 32 64)
#(2 154 1 2 3 7 10 11 12 13 14 15 16 17 18 19 20 21 22 23 24 26 29 30 32 64)
#(2 146 1 2 3 7 10 11 12 13 14 15 16 17 18 19 20 21 22 23 24 26 29 30 32 64)
#(3 41 1 45 3 61 10 65 11 69 12 73 13 77 15 81 18 85 19 89 20 93 21 97 29 249 37 113 39 117 40 121 41 125 42 129 43 133 44 137 46 141 48 145 49 149 50 153 56)
#(3 41 1 45 3 49 4 61 10 65 11 69 12 73 13 378 14 77 15 81 18 85 19 89 20 93 21 97 29 105 35 109 37 113 39 117 40 121 41 125 42 129 43 133 44 137 46 141 48 145 49 149 50 153 56 253 63)
#(2 158 1 2 3 7 10 11 12 13 14 15 16 17 18 19 20 21 22 23 24 26 29 30 32 64)
#(2 162 1 2 3 7 10 11 12 13 14 15 16 17 18 19 20 21 22 23 24 26 29 30 32 64)
#(2 170 1 2 3 7 10 11 12 13 14 15 16 17 18 19 20 21 22 23 24 26 29 30 32 64)
#(2 298 1 2 3 4 5 6 8 9 10 11 12 13 14 15 16 17 18 19 20 21 22 24 25 29 30 64)
#(2 166 1 2 3 7 10 11 12 13 14 15 16 17 18 19 20 21 22 23 24 26 29 30 32 64)
#(3 41 1 378 2 45 3 49 4 61 10 65 11 69 12 73 13 77 15 81 18 85 19 89 20 93 21 97 29 105 35 109 37 113 39 117 40 121 41 125 42 129 43 133 44 137 46 141 48 145 49 149 50 153 56 257 63 378 64)
#(3 66 2 66 14 261 30 265 36 66 64)
#(2 58 2 14 30 64)
#(2 130 2 14 16 17 21 22 24 30 64)
#(2 74 2 14 16 30 64)
#(3 78 2 78 14 78 16 269 17 78 30 78 64)
#(3 82 2 82 14 82 16 273 21 277 22 281 24 82 30 285 51 289 52 293 53 82 64)
#(2 126 2 14 16 17 21 22 24 30 64)
#(2 134 2 14 16 17 21 22 24 30 64)
#(2 114 2 14 16 17 30 64)
#(2 210 2 14 16 17 30 64)
#(3 214 2 214 14 214 16 214 17 277 22 281 24 214 30 297 52 301 53 214 64)
#(3 218 2 218 14 218 16 218 17 273 21 277 22 281 24 218 30 305 51 309 52 313 53 218 64)
#(3 138 2 138 14 138 16 138 21 138 22 138 24 317 25 138 30 138 64)
#(2 0 64)
#(2 314 2 64)
#(2 0 64)
#(2 222 1 3 4 6 8 9 10 11 12 13 15 18 19 20 21 22 29 64)
#(2 306 1 3 4 6 8 9 10 11 12 13 15 18 19 20 21 29 64)
#(3 93 21 321 56)
#(3 6 3 6 7 6 10 6 11 6 12 6 13 6 14 6 15 6 16 6 17 6 18 6 19 6 20 6 21 6 22 6 24 6 29 6 30 325 33)
#(3 41 1 45 3 49 4 181 6 61 10 65 11 69 12 73 13 77 15 81 18 85 19 89 20 93 21 97 29 105 35 109 37 113 39 117 40 121 41 125 42 129 43 133 44 329 45 137 46 141 48 145 49 149 50 153 56 257 63 378 64)
#(3 41 1 45 3 49 4 53 8 57 9 61 10 65 11 69 12 73 13 77 15 81 18 85 19 89 20 93 21 97 29 101 34 105 35 109 37 113 39 117 40 121 41 125 42 129 43 133 44 137 46 141 48 145 49 149 50 153 56 333 58 161 63 378 64)
#(2 406 64)
#(3 93 21 337 56)
#(3 341 2 345 5 349 8 353 9)
#(2 357 2)
#(3 326 2 326 18 361 60)
#(3 334 1 334 3 334 10 334 11 334 12 334 13 334 16 334 18 334 19 334 20 334 21 334 22 334 23 334 24 334 26 334 29 334 32 365 61)
#(2 174 1 2 3 7 10 11 12 13 14 15 16 17 18 19 20 21 22 23 24 26 29 30 32 64)
#(2 178 1 2 3 7 10 11 12 13 14 15 16 17 18 19 20 21 22 23 24 26 29 30 32 64)
#(2 186 1 2 3 7 10 11 12 13 14 15 16 17 18 19 20 21 22 23 24 26 29 30 32 64)
#(2 190 1 2 3 7 10 11 12 13 14 15 16 17 18 19 20 21 22 23 24 26 29 30 32 64)
#(2 182 1 2 3 7 10 11 12 13 14 15 16 17 18 19 20 21 22 23 24 26 29 30 32 64)
#(2 194 1 2 3 7 10 11 12 13 14 15 16 17 18 19 20 21 22 23 24 26 29 30 32 64)
#(3 66 2 66 14 369 30 373 36 66 64)
#(3 377 9 93 21 381 56)
#(2 385 16)
#(2 389 14)
#(2 318 2 64)
#(3 41 1 70 2 45 3 393 4 61 10 65 11 69 12 73 13 70 14 77 15 81 18 85 19 89 20 93 21 97 29 397 37 113 39 117 40 121 41 125 42 129 43 133 44 137 46 141 48 145 49 149 50 153 56 70 64)
#(2 382 2 14 64)
#(3 273 21 277 22 281 24 401 51 405 52 409 53 413 66)
#(2 262 2 14 16 17 21 22 24 30 64)
#(3 41 1 45 3 61 10 65 11 69 12 73 13 77 15 81 18 85 19 89 20 93 21 97 29 113 39 417 42 129 43 133 44 421 49 425 50 429 54 433 56)
#(3 41 1 45 3 61 10 65 11 69 12 73 13 77 15 81 18 85 19 89 20 93 21 97 29 113 39 437 42 129 43 133 44 441 50 445 55 433 56)
#(2 258 2 14 16 17 21 22 24 30 64)
#(2 250 2 14 16 17 22 24 30 64)
#(3 238 2 238 14 238 16 238 17 449 22 238 30 238 64)
#(2 242 2 14 16 17 22 24 30 64)
#(3 230 2 230 14 230 16 230 17 449 22 230 30 230 64)
#(2 254 2 14 16 17 21 22 24 30 64)
#(2 246 2 14 16 17 22 24 30 64)
#(3 234 2 234 14 234 16 234 17 449 22 234 30 234 64)
#(3 41 1 45 3 61 10 65 11 69 12 73 13 77 15 81 18 85 19 89 20 93 21 97 29 453 37 113 39 117 40 121 41 125 42 129 43 133 44 137 46 141 48 145 49 149 50 153 56)
#(2 226 1 3 4 6 8 9 10 11 12 13 15 18 19 20 21 22 29 64)
#(3 45 3 457 7 61 10 65 11 69 12 461 13 465 14 469 15 473 16 477 17 81 18 85 19 89 20 481 21 485 22 489 24 97 29 493 30 497 44)
#(3 41 1 45 3 49 4 61 10 65 11 69 12 73 13 77 15 81 18 85 19 89 20 93 21 97 29 105 35 109 37 113 39 117 40 121 41 125 42 129 43 133 44 137 46 141 48 145 49 149 50 153 56 501 63 378 64)
#(2 410 64)
#(2 86 2 5 8 9)
#(2 102 2 14 16 17 21 22 24 30 64)
#(3 93 21 505 56)
#(3 418 9 418 21 509 68)
#(3 41 1 378 2 45 3 49 4 53 8 57 9 61 10 65 11 69 12 73 13 77 15 81 18 85 19 89 20 93 21 97 29 101 34 105 35 109 37 113 39 117 40 121 41 125 42 129 43 133 44 137 46 141 48 145 49 149 50 153 56 513 58 161 63)
#(2 98 2 14 16 17 21 22 24 30 64)
#(3 517 2 521 18)
#(3 525 1 45 3 61 10 65 11 69 12 529 13 533 16 81 18 85 19 89 20 537 21 541 22 545 23 549 24 553 26 97 29 557 32 561 44 565 62)
#(2 70 2 14 64)
#(2 390 2 14 64)
#(2 54 1 2 3 4 6 10 11 12 13 15 18 19 20 21 29 64)
#(2 422 9 21)
#(2 122 2 14 16 17 21 22 24 30 64)
#(2 142 2 14 16 17 21 22 24 30 64)
#(3 41 1 45 3 61 10 65 11 69 12 73 13 77 15 81 18 85 19 89 20 93 21 97 29 569 37 113 39 117 40 121 41 125 42 129 43 133 44 137 46 141 48 145 49 149 50 153 56)
#(2 62 2 14 30 64)
#(2 394 2 14 16 17 30 64)
#(2 398 2 14 16 17 30 64)
#(3 402 2 402 14 402 16 402 17 449 22 402 30 402 64)
#(2 118 2 14 16 17 30 64)
#(3 286 2 286 14 286 16 286 17 273 21 286 22 281 24 286 30 285 51 289 52 286 64)
#(3 278 2 278 14 278 16 278 17 278 22 281 24 278 30 297 52 278 64)
#(3 282 2 282 14 282 16 282 17 273 21 282 22 281 24 282 30 305 51 309 52 282 64)
#(2 270 2 14 16 17 22 30 64)
#(2 138 2 14 16 17 21 22 24 30 64)
#(3 294 2 294 14 294 16 294 17 273 21 294 22 294 24 294 30 285 51 294 64)
#(3 290 2 290 14 290 16 290 17 273 21 290 22 290 24 290 30 305 51 290 64)
#(2 266 2 14 16 17 22 24 30 64)
#(3 41 1 45 3 61 10 65 11 69 12 73 13 77 15 81 18 85 19 89 20 93 21 97 29 113 39 417 42 129 43 133 44 421 49 425 50 573 54 433 56)
#(2 110 2 14 16 30 64)
#(2 206 1 3 4 8 9 10 11 12 13 15 18 19 20 21 29 64)
#(2 18 3 7 10 11 12 13 14 15 16 17 18 19 20 21 22 24 29 30)
#(2 30 3 7 10 11 12 13 14 15 16 17 18 19 20 21 22 24 29 30)
#(2 26 3 7 10 11 12 13 14 15 16 17 18 19 20 21 22 24 29 30)
#(2 22 3 7 10 11 12 13 14 15 16 17 18 19 20 21 22 24 29 30)
#(2 34 3 7 10 11 12 13 14 15 16 17 18 19 20 21 22 24 29 30)
#(2 42 3 7 10 11 12 13 14 15 16 17 18 19 20 21 22 24 29 30)
#(2 10 3 7 10 11 12 13 14 15 16 17 18 19 20 21 22 24 29 30)
#(2 46 3 7 10 11 12 13 14 15 16 17 18 19 20 21 22 24 29 30)
#(2 38 3 7 10 11 12 13 14 15 16 17 18 19 20 21 22 24 29 30)
#(2 14 3 7 10 11 12 13 14 15 16 17 18 19 20 21 22 24 29 30)
#(2 414 64)
#(2 90 2 5 8 9)
#(3 577 9 93 21 381 56)
#(2 581 2)
#(2 198 1 2 3 7 10 11 12 13 14 15 16 17 18 19 20 21 22 23 24 26 29 30 32 64)
#(2 330 2 18)
#(3 326 2 326 18 585 60)
#(3 334 1 334 3 334 10 334 11 334 12 334 13 334 16 334 18 334 19 334 20 334 21 334 22 334 23 334 24 334 26 334 29 334 32 589 61)
#(2 202 1 2 3 7 10 11 12 13 14 15 16 17 18 19 20 21 22 23 24 26 29 30 32 64)
#(2 346 1 3 10 11 12 13 16 18 19 20 21 22 23 24 26 29 32)
#(2 354 1 3 10 11 12 13 16 18 19 20 21 22 23 24 26 29 32)
#(2 358 1 3 10 11 12 13 16 18 19 20 21 22 23 24 26 29 32)
#(2 350 1 3 10 11 12 13 16 18 19 20 21 22 23 24 26 29 32)
#(2 362 1 3 10 11 12 13 16 18 19 20 21 22 23 24 26 29 32)
#(2 374 1 3 10 11 12 13 16 18 19 20 21 22 23 24 26 29 32)
#(2 342 1 3 10 11 12 13 16 18 19 20 21 22 23 24 26 29 32)
#(2 338 1 3 10 11 12 13 16 18 19 20 21 22 23 24 26 29 32)
#(3 66 2 66 14 369 30 593 36 66 64)
#(2 274 2 14 16 17 22 30 64)
#(3 41 1 378 2 45 3 49 4 61 10 65 11 69 12 73 13 77 15 81 18 85 19 89 20 93 21 97 29 105 35 109 37 113 39 117 40 121 41 125 42 129 43 133 44 137 46 141 48 145 49 149 50 153 56 597 63)
#(2 94 2 14 16 17 21 22 24 30 64)
#(3 601 2 521 18)
#(3 525 1 45 3 61 10 65 11 69 12 529 13 605 16 81 18 85 19 89 20 537 21 541 22 545 23 549 24 553 26 97 29 557 32 561 44 565 62)
#(2 386 2 14 64)
#(2 609 2)
#(2 370 1 3 10 11 12 13 16 18 19 20 21 22 23 24 26 29 32)
#(2 366 1 3 10 11 12 13 16 18 19 20 21 22 23 24 26 29 32)
#(2 106 2 14 16 17 21 22 24 30 64)
)! !


!SqueakParser methodsFor: 'generated-reduction actions' stamp: 'ajh 7/19/2004 11:54'!
reduceActionForOptionalXXXperiodX1: nodes 
	^ nil! !

!SqueakParser methodsFor: 'generated-reduction actions' stamp: 'ajh 7/19/2004 11:54'!
reduceActionForOptionalXXXperiodX2: nodes 
	^ nodes at: 1! !


!SqueakParser methodsFor: 'private' stamp: 'ajh 7/19/2004 12:05'!
actionForCurrentToken

	| ids action |
	ids := currentToken id.
	1 to: ids size do: [:i | 
		action := self actionFor: (ids at: i).
		(action bitAnd: self actionMask) = self errorAction ifFalse: [^ action].

		"Convert negative number to binarySymbol"
		(ids at: i) = scanner negativeNumberId ifTrue: [
			currentToken := SmaCCToken 
				value: '-'
				start: currentToken startPosition
				id: (Array with: scanner binarySymbolId).
			scanner position: currentToken stopPosition.
			^ self actionForCurrentToken].

		"Ignore repeating periods"
		((ids at: i) = scanner periodId and: [(nodeStack last isKindOf: SmaCCToken) and: [nodeStack last id first = scanner periodId]]) ifTrue: [
			currentToken := nil.
			self getNextToken.
			^ self actionForCurrentToken].

	].
	^ self errorAction! !

!SqueakParser methodsFor: 'private' stamp: 'ajh 3/6/2003 00:02'!
nodeStack

	^ nodeStack! !

!SqueakParser methodsFor: 'private' stamp: 'ajh 3/6/2003 00:19'!
scanner

	^ scanner! !


!SqueakParser methodsFor: 'reduction actions' stamp: 'ajh 3/15/2003 00:01'!
add3: nodes

	^ nodes first copyWith: nodes third! !

!SqueakParser methodsFor: 'reduction actions' stamp: 'ajh 3/14/2003 23:59'!
add: nodes

	^ nodes first copyWith: nodes second! !

!SqueakParser methodsFor: 'reduction actions' stamp: 'ajh 3/14/2003 23:54'!
addMessagePart: nodes

	^ {nodes first first copyWith: nodes second.
	   nodes first second copyWith: nodes third}! !

!SqueakParser methodsFor: 'reduction actions' stamp: 'ajh 3/14/2003 23:58'!
array

	^ #()! !

!SqueakParser methodsFor: 'reduction actions' stamp: 'ajh 3/15/2003 00:30'!
array: nodes

	^ RBArrayNode
		leftBrace: nodes first start
		rightBrace: nodes third stop
		statements: nodes second! !

!SqueakParser methodsFor: 'reduction actions' stamp: 'ajh 3/15/2003 00:07'!
assignment: nodes

	^ RBAssignmentNode
		variable: nodes first
		value: nodes third! !

!SqueakParser methodsFor: 'reduction actions' stamp: 'ajh 3/15/2003 00:04'!
blockArgs: nodes

	^ (RBBlockNode
		arguments: nodes second
		body: (RBSequenceNode statements: #()))
		left: nodes first start;
		right: nodes third stop! !

!SqueakParser methodsFor: 'reduction actions' stamp: 'ajh 3/15/2003 00:03'!
blockNoArgs: nodes

	^ (RBBlockNode
		body: nodes second)
		left: nodes first start;
		right: nodes third stop! !

!SqueakParser methodsFor: 'reduction actions' stamp: 'ajh 3/15/2003 00:02'!
blockWithArgs: nodes

	^ (RBBlockNode
		arguments: nodes second
		body: nodes fourth)
		left: nodes first start;
		right: nodes fifth stop! !

!SqueakParser methodsFor: 'reduction actions' stamp: 'ajh 3/15/2003 00:06'!
blockWithTemps: nodes

	^ (RBBlockNode
		arguments: nodes second
		body: (RBSequenceNode
			temporaries: nodes fourth
			statements: nodes sixth))
		left: nodes first start;
		right: nodes seventh stop! !

!SqueakParser methodsFor: 'reduction actions' stamp: 'ajh 3/15/2003 00:23'!
byteStream

	^ ByteArray new writeStream! !

!SqueakParser methodsFor: 'reduction actions' stamp: 'ajh 3/15/2003 00:25'!
byteStreamPut: nodes

	^ nodes first
		nextPut: nodes second value asNumber;
		yourself! !

!SqueakParser methodsFor: 'reduction actions' stamp: 'ajh 3/15/2003 00:12'!
cascade: nodes

	^ (nodes first isKindOf: RBMessageNode)
		ifTrue: [RBCascadeNode messages:
			{nodes first.
			 RBMessageNode new
				receiver: nodes first receiver
				selectorParts: nodes third first
				arguments: nodes third last}]
		ifFalse: [RBCascadeNode messages:
			(nodes first messages copyWith:
				(RBMessageNode new
					receiver: nodes first messages first receiver
					selectorParts: nodes third first
					arguments: nodes third last))]! !

!SqueakParser methodsFor: 'reduction actions' stamp: 'ajh 3/15/2003 00:27'!
contents2: nodes

	^ nodes second contents! !

!SqueakParser methodsFor: 'reduction actions' stamp: 'ajh 3/14/2003 23:53'!
first: nodes

	^ nodes first! !

!SqueakParser methodsFor: 'reduction actions' stamp: 'ajh 3/15/2003 00:01'!
firstIn: nodes

	^ {nodes first}! !

!SqueakParser methodsFor: 'reduction actions' stamp: 'ajh 3/17/2003 12:10'!
litArray: nodes

	^ RBLiteralNode
		literalToken: (SmaCCToken
			value: (String new: (nodes fourth stop - nodes second start + 1))
			start: nodes second start)
		value: nodes third contents! !

!SqueakParser methodsFor: 'reduction actions' stamp: 'ajh 3/15/2003 00:16'!
litChar: nodes

	^ RBLiteralNode
		literalToken: nodes first
		value: (nodes first value at: 2)! !

!SqueakParser methodsFor: 'reduction actions' stamp: 'ajh 3/15/2003 00:14'!
litFalse: nodes

	^ RBLiteralNode
		literalToken: nodes first
		value: false! !

!SqueakParser methodsFor: 'reduction actions' stamp: 'ajh 3/15/2003 00:14'!
litNil: nodes

	^ RBLiteralNode
		literalToken: nodes first
		value: nil! !

!SqueakParser methodsFor: 'reduction actions' stamp: 'ajh 3/15/2003 13:42'!
litNumber: nodes

	| str num |
	str _ nodes first value readStream.
	num _ Number readFrom: str.
	str atEnd ifFalse: [
		currentToken _ nil.
		scanner position: scanner position - (str originalContents size - str position) + 1.
		self reportErrorMessage: 'Digit out of range'].
	^ RBLiteralNode
		literalToken: nodes first
		value: num! !

!SqueakParser methodsFor: 'reduction actions' stamp: 'ajh 3/15/2003 00:18'!
litString: nodes

	| s |
	s _ nodes first value.
	^ RBLiteralNode
		literalToken: nodes first
		value: ((s copyFrom: 2 to: s size - 1)
				copyReplaceAll: '''''' with: '''')! !

!SqueakParser methodsFor: 'reduction actions' stamp: 'ajh 3/15/2003 00:19'!
litStringSymbol: nodes

	| s |
	s _ nodes second value.
	^ RBLiteralNode
		literalToken: nodes second
		value: ((s copyFrom: 2 to: s size - 1)
				copyReplaceAll: '''''' with: '''') asSymbol! !

!SqueakParser methodsFor: 'reduction actions' stamp: 'ajh 3/15/2003 00:20'!
litSymbol: nodes

	^ RBLiteralNode
		literalToken: nodes second
		value: nodes second value asSymbol! !

!SqueakParser methodsFor: 'reduction actions' stamp: 'ajh 3/15/2003 00:14'!
litTrue: nodes

	^ RBLiteralNode
		literalToken: nodes first
		value: true! !

!SqueakParser methodsFor: 'reduction actions' stamp: 'ajh 3/14/2003 23:52'!
messagePart: nodes

	^ {{nodes first}. {nodes second}}! !

!SqueakParser methodsFor: 'reduction actions' stamp: 'ajh 3/15/2003 00:13'!
messageSend: nodes

	^ RBMessageNode new
		receiver: nodes first
		selectorParts: nodes second first
		arguments: nodes second last! !

!SqueakParser methodsFor: 'reduction actions' stamp: 'ajh 3/14/2003 23:48'!
method: nodes

	^ RBMethodNode new
		selectorParts: nodes first first
		arguments: nodes first last;
		body: nodes second! !

!SqueakParser methodsFor: 'reduction actions' stamp: 'ajh 3/14/2003 23:49'!
methodPrim: nodes

	^ RBMethodNode new
		selectorParts: nodes first first
		arguments: nodes first last;
		body: nodes third;
		tags: nodes second! !

!SqueakParser methodsFor: 'reduction actions' stamp: 'ajh 3/14/2003 23:50'!
methodTempsPrim: nodes

	^ RBMethodNode new
		selectorParts: nodes first first
		arguments: nodes first last;
		body: (RBSequenceNode
			temporaries: nodes second
			statements: nodes fourth);
		tags: nodes third! !

!SqueakParser methodsFor: 'reduction actions' stamp: 'ajh 3/14/2003 23:55'!
primitiveMessage: nodes

	^ {nodes first start to: nodes third stop}! !

!SqueakParser methodsFor: 'reduction actions' stamp: 'ajh 3/19/2003 13:49'!
return: nodes

	^ {RBReturnNode
		return: nodes first start
		value: nodes second}! !

!SqueakParser methodsFor: 'reduction actions' stamp: 'ajh 3/19/2003 13:49'!
returnAdd: nodes

	^ nodes first copyWith: (RBReturnNode
		return: nodes third start
		value: nodes fourth)! !

!SqueakParser methodsFor: 'reduction actions' stamp: 'ajh 3/14/2003 23:58'!
second: nodes

	^ nodes second! !

!SqueakParser methodsFor: 'reduction actions' stamp: 'ajh 3/15/2003 00:06'!
secondIn: nodes

	^ {nodes second}! !

!SqueakParser methodsFor: 'reduction actions' stamp: 'ajh 3/14/2003 23:57'!
sequence: nodes

	^ RBSequenceNode
		statements: nodes first! !

!SqueakParser methodsFor: 'reduction actions' stamp: 'ajh 3/14/2003 23:58'!
sequenceWithTemps: nodes

	^ RBSequenceNode
		temporaries: nodes first
		statements: nodes second! !

!SqueakParser methodsFor: 'reduction actions' stamp: 'ajh 3/15/2003 00:23'!
stream

	^ Array new writeStream! !

!SqueakParser methodsFor: 'reduction actions' stamp: 'ajh 3/15/2003 00:26'!
streamPut: nodes

	^ nodes first
		nextPut: nodes second;
		yourself! !

!SqueakParser methodsFor: 'reduction actions' stamp: 'ajh 3/14/2003 23:52'!
unaryMessage: nodes

	^ {{nodes first}. #()}! !

!SqueakParser methodsFor: 'reduction actions' stamp: 'ajh 3/15/2003 00:27'!
value: nodes

	^ nodes first value! !

!SqueakParser methodsFor: 'reduction actions' stamp: 'ajh 3/15/2003 00:27'!
valueSymbol: nodes

	^ nodes first value asSymbol! !

!SqueakParser methodsFor: 'reduction actions' stamp: 'ajh 3/15/2003 00:13'!
variable: nodes

	^ RBVariableNode new
		identifierToken: nodes first! !

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

SqueakParser class
	instanceVariableNames: ''!

!SqueakParser class methodsFor: 'generated-starting states' stamp: 'ajh 7/19/2004 11:54'!
startingStateForMethod
	^1! !

!SqueakParser class methodsFor: 'generated-starting states' stamp: 'ajh 7/19/2004 11:54'!
startingStateForMethodPattern
	^3! !

!SqueakParser class methodsFor: 'generated-starting states' stamp: 'ajh 7/19/2004 11:54'!
startingStateForSequence
	^2! !


!SqueakParser class methodsFor: 'generated-accessing' stamp: 'ajh 7/19/2004 11:54'!
scannerClass
	^SqueakScanner! !


!SqueakParser class methodsFor: 'generated-comments' stamp: 'ajh 7/19/2004 11:54'!
parserDefinitionComment

	"%id <name> <number> <negativeNumber> <binarySymbol> <period>;
%start Sequence MethodPattern;

Method:
	MethodPattern Sequence						{#method:}
|	MethodPattern Primitive Sequence			{#methodPrim:}
|	MethodPattern Temporaries Primitive Statements {#methodTempsPrim:};

MethodPattern:
	<name>										{#unaryMessage:}
|	<binarySymbol> Variable						{#messagePart:}
|	KeywordMethodPattern						{#first:};

KeywordMethodPattern:
	<keyword> Variable							{#messagePart:}
|	KeywordMethodPattern <keyword> Variable	{#addMessagePart:};

Primitive:
	""<"" PrimitiveMessage "">""						{#primitiveMessage:};

Sequence:
	Statements									{#sequence:}
|	Temporaries Statements						{#sequenceWithTemps:};

Temporaries:
	""||""											{#array}
|	""|"" TemporaryVariables ""|""					{#second:};

TemporaryVariables:
												{#array}
|	TemporaryVariables Variable				{#add:};

Statements:
												{#array}
|	StatementList <period>?						{#first:}
|	StatementList <period> ""^"" Expression <period>?	{#returnAdd:}
|	""^"" Expression <period>?						{#return:};

StatementList:
	Expression									{#firstIn:}
|	StatementList <period> Expression				{#add3:};

Block:
	""["" BlockArgs ""|"" Sequence ""]""					{#blockWithArgs:}
|	""["" Sequence ""]""								{#blockNoArgs:}
|	""["" BlockArgs ""]""								{#blockArgs:}
|	""["" BlockArgs ""||"" TemporaryVariables ""|"" Statements ""]"" {#blockWithTemps:};

BlockArgs:
	"":"" Variable									{#secondIn:}
|	BlockArgs "":"" Variable						{#add3:};

Expression:
	Assignment									{#first:}
|	Cascade										{#first:}
|	Primary									{#first:};

Primary:
	""("" Expression "")""								{#second:}
|	Array										{#first:}
|	Block										{#first:}
|	Literal										{#first:}
|	Variable									{#first:};

Assignment:
	Variable <assignment> Expression				{#assignment:};

Cascade:
	MessageSend								{#first:}
|	Cascade "";"" Message							{#cascade:};

MessageSend:
	KeywordMessageSend						{#first:}
|	BinaryMessageSend							{#first:}
|	UnaryMessageSend							{#first:};

Message:
	UnaryMessage								{#first:}
|	BinaryMessage								{#first:}
|	KeywordMessage							{#first:};

KeywordMessageSend:
	BinaryMessageSend KeywordMessage			{#messageSend:}
|	UnaryMessageSend KeywordMessage			{#messageSend:}
|	Primary KeywordMessage					{#messageSend:};

KeywordMessage:
	<keyword> KeywordArgument				{#messagePart:}
|	KeywordMessage <keyword> KeywordArgument {#addMessagePart:};

KeywordArgument:
	BinaryMessageSend							{#first:}
|	UnaryMessageSend							{#first:}
|	Primary									{#first:};

BinaryMessageSend:
	BinaryMessageSend BinaryMessage			{#messageSend:}
|	UnaryMessageSend BinaryMessage			{#messageSend:}
|	Primary BinaryMessage						{#messageSend:};

BinaryMessage :
	<binarySymbol> BinaryArgument				{#messagePart:};

BinaryArgument:
	UnaryMessageSend							{#first:}
|	Primary									{#first:};

UnaryMessageSend :
	UnaryMessageSend UnaryMessage			{#messageSend:}
|	Primary UnaryMessage						{#messageSend:};

UnaryMessage :
	<name>										{#unaryMessage:};

Array:
	""{"" Statements ""}""								{#array:};

Variable: 
	<name>										{#variable:};

Literal:
	""true""										{#litTrue:}
|	""false""										{#litFalse:}
|	""nil""										{#litNil:}
|	<number>									{#litNumber:}
|	<negativeNumber>							{#litNumber:} 
|	<character>									{#litChar:}
|	<string>										{#litString:}
|	""#"" <string>									{#litStringSymbol:}
|	""#"" <name>									{#litSymbol:}
|	""#"" <binarySymbol>							{#litSymbol:}
|	""#"" <keyword>								{#litSymbol:}
|	""#"" <multikeyword>							{#litSymbol:}
|	""#"" <alternateKeyword>						{#litSymbol:}
|	""#"" ""["" ByteArray ""]""							{#litArray:}
|	""#"" ""("" LiteralArray "")""						{#litArray:};

ByteArray:
												{#byteStream}
|	ByteArray <number>							{#byteStreamPut:};

LiteralArray:
												{#stream}
|	LiteralArray ArrayLiteral					{#streamPut:};

ArrayLiteral:
	Literal										{#value:}
|	<name>										{#valueSymbol:}
|	<binarySymbol>								{#valueSymbol:}
|	<keyword>									{#valueSymbol:}
|	<multikeyword>								{#valueSymbol:}
|	<alternateKeyword>							{#valueSymbol:}
|	""("" LiteralArray "")""							{#contents2:}
|	""["" ByteArray ""]""							{#contents2:}
|	<anyChar>									{#valueSymbol:};

PrimitiveMessage:
	| PrimitiveMessage <keyword>
	| PrimitiveMessage Literal
	| PrimitiveMessage ""(""
	| PrimitiveMessage "")""
	| PrimitiveMessage ""{""
	| PrimitiveMessage ""}""
	| PrimitiveMessage "";""
	| PrimitiveMessage <period>
	| PrimitiveMessage <name>
	| PrimitiveMessage <binarySymbol>;
"! !


!SqueakParser class methodsFor: 'parsing' stamp: 'ajh 2/27/2003 16:40'!
parseDoIt: stringOrStream

	| sequence |
	sequence _ self
		parseStream: stringOrStream readStream
		startingAt: self startingStateForSequence.
	^ (RBDoItNode body: sequence)
		source: stringOrStream contents! !

!SqueakParser class methodsFor: 'parsing' stamp: 'ajh 2/27/2003 16:40'!
parseMethod: stringOrStream

	^ (self
		parseStream: stringOrStream readStream
		startingAt: self startingStateForMethod)
		source: stringOrStream contents! !

!SqueakParser class methodsFor: 'parsing' stamp: 'ajh 6/22/2003 21:06'!
parseMethodPattern: stringOrStream

	| array comments |
	array _
		[self
			parseStream: stringOrStream readStream
			startingAt: self startingStateForMethodPattern
		] on: SmaCCParserError do: [:ex | | parser |
			parser _ ex tag.
			comments _ parser scanner comments.
			parser nodeStack size = 1 ifFalse: [ex pass].
			parser nodeStack last
		].
	^ RBMethodNode new
		selectorParts: array first arguments: array last;
		comments: comments! !


SmaCCScanner subclass: #SqueakScanner
	instanceVariableNames: 'comments'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Syntax'!
!SqueakScanner commentStamp: 'ajh 3/24/2003 21:31' prior: 0!
I parse Smalltalk text into tokens which are used by the SqueakParser.  Methods under 'generated-*' categories were automatically generated using SmaCC.!


!SqueakScanner methodsFor: 'generated-tokens' stamp: 'ajh 7/19/2004 11:54'!
binarySymbolId
	^24! !

!SqueakScanner methodsFor: 'generated-tokens' stamp: 'ajh 7/19/2004 11:54'!
emptySymbolTokenId
	^64! !

!SqueakScanner methodsFor: 'generated-tokens' stamp: 'ajh 7/19/2004 11:54'!
errorTokenId
	^65! !

!SqueakScanner methodsFor: 'generated-tokens' stamp: 'ajh 7/19/2004 11:54'!
nameId
	^21! !

!SqueakScanner methodsFor: 'generated-tokens' stamp: 'ajh 7/19/2004 11:54'!
negativeNumberId
	^19! !

!SqueakScanner methodsFor: 'generated-tokens' stamp: 'ajh 7/19/2004 11:54'!
numberId
	^18! !

!SqueakScanner methodsFor: 'generated-tokens' stamp: 'ajh 7/19/2004 11:54'!
periodId
	^30! !


!SqueakScanner methodsFor: 'generated-scanner' stamp: 'ajh 7/19/2004 11:54'!
scan1
	self step.
	currentCharacter isDigit 
		ifTrue: 
			[
			[self recordMatch: #(18).
			self step.
			currentCharacter isDigit] 
					whileTrue.
			^ self reportLastMatch].
	currentCharacter == $- 
		ifTrue: 
			[self step.
			currentCharacter isDigit 
				ifTrue: 
					[
					[self recordMatch: #(18).
					self step.
					currentCharacter isDigit] 
							whileTrue.
					^ self reportLastMatch].
			^ self reportLastMatch].
	^ self reportLastMatch! !

!SqueakScanner methodsFor: 'generated-scanner' stamp: 'ajh 7/19/2004 11:54'!
scan2
	self step.
	currentCharacter isDigit 
		ifTrue: 
			[
			[self recordMatch: #(19).
			self step.
			currentCharacter isDigit] 
					whileTrue.
			^ self reportLastMatch].
	currentCharacter == $- 
		ifTrue: 
			[self step.
			currentCharacter isDigit 
				ifTrue: 
					[
					[self recordMatch: #(19).
					self step.
					currentCharacter isDigit] 
							whileTrue.
					^ self reportLastMatch].
			^ self reportLastMatch].
	^ self reportLastMatch! !

!SqueakScanner methodsFor: 'generated-scanner' stamp: 'ajh 7/19/2004 11:54'!
scan3
	
	[self step.
	currentCharacter ~~ $'] whileTrue.
	^ self scan4! !

!SqueakScanner methodsFor: 'generated-scanner' stamp: 'ajh 7/19/2004 11:54'!
scan4
	self recordMatch: #(20).
	self step.
	currentCharacter == $' ifTrue: [^ self scan3].
	^ self reportLastMatch! !

!SqueakScanner methodsFor: 'generated-scanner' stamp: 'ajh 7/19/2004 11:54'!
scan5
	
	[self step.
	currentCharacter isDigit or: 
			[(currentCharacter between: $A and: $Z) 
				or: [currentCharacter between: $a and: $z]]] 
			whileTrue.
	currentCharacter == $: 
		ifTrue: 
			[self recordMatch: #(23).
			self step.
			((currentCharacter between: $A and: $Z) 
				or: [currentCharacter between: $a and: $z]) ifTrue: [^ self scan5].
			^ self reportLastMatch].
	^ self reportLastMatch! !

!SqueakScanner methodsFor: 'generated-scanner' stamp: 'ajh 7/19/2004 11:54'!
scan6
	
	[self step.
	currentCharacter isDigit or: 
			[(currentCharacter between: $A and: $Z) 
				or: [currentCharacter between: $a and: $z]]] 
			whileTrue.
	currentCharacter == $: 
		ifTrue: 
			[self recordMatch: #(26).
			self step.
			((currentCharacter between: $A and: $Z) 
				or: [currentCharacter between: $a and: $z]) ifTrue: [^ self scan6].
			^ self reportLastMatch].
	^ self reportLastMatch! !

!SqueakScanner methodsFor: 'generated-scanner' stamp: 'ajh 7/19/2004 11:54'!
scan7
	self step.
	currentCharacter isDigit 
		ifTrue: 
			[
			[self recordMatch: #(18).
			self step.
			currentCharacter isDigit] 
					whileTrue.
			currentCharacter == $e ifTrue: [^ self scan1].
			^ self reportLastMatch].
	^ self reportLastMatch! !

!SqueakScanner methodsFor: 'generated-scanner' stamp: 'ajh 7/19/2004 11:54'!
scan8
	self step.
	(currentCharacter isDigit or: [currentCharacter between: $A and: $Z]) 
		ifTrue: 
			[
			[self recordMatch: #(18).
			self step.
			currentCharacter isDigit or: [currentCharacter between: $A and: $Z]] 
					whileTrue.
			currentCharacter == $. 
				ifTrue: 
					[self step.
					(currentCharacter isDigit or: [currentCharacter between: $A and: $Z]) 
						ifTrue: 
							[
							[self recordMatch: #(18).
							self step.
							currentCharacter isDigit or: [currentCharacter between: $A and: $Z]] 
									whileTrue.
							currentCharacter == $e ifTrue: [^ self scan1].
							^ self reportLastMatch].
					^ self reportLastMatch].
			currentCharacter == $e ifTrue: [^ self scan1].
			^ self reportLastMatch].
	^ self reportLastMatch! !

!SqueakScanner methodsFor: 'generated-scanner' stamp: 'ajh 7/19/2004 11:54'!
scan9
	self recordMatch: #(22).
	self step.
	((currentCharacter between: $A and: $Z) 
		or: [currentCharacter between: $a and: $z]) ifTrue: [^ self scan5].
	currentCharacter == $= 
		ifTrue: [^ self recordAndReportMatch: #variableAssignment].
	^ self reportLastMatch! !

!SqueakScanner methodsFor: 'generated-scanner' stamp: 'ajh 7/19/2004 11:54'!
scanForToken
	self step.
	(currentCharacter <= $ or: 
			[(currentCharacter between: $ and: $) 
				or: [currentCharacter == $` or: [currentCharacter >= $]]]) 
		ifTrue: [^ self recordAndReportMatch: #(32)].
	((currentCharacter between: $A and: $Z) 
		or: [currentCharacter between: $a and: $z]) 
			ifTrue: 
				[self recordMatch: #(21 32).
				self step.
				(currentCharacter isDigit or: 
						[(currentCharacter between: $A and: $Z) 
							or: [currentCharacter between: $a and: $z]]) 
					ifTrue: 
						[
						[self recordMatch: #(21).
						self step.
						currentCharacter isDigit or: 
								[(currentCharacter between: $A and: $Z) 
									or: [currentCharacter between: $a and: $z]]] 
								whileTrue.
						currentCharacter == $: ifTrue: [^ self scan9].
						^ self reportLastMatch].
				currentCharacter == $: ifTrue: [^ self scan9].
				^ self reportLastMatch].
	(currentCharacter == $!! or: 
			[(currentCharacter between: $% and: $&) or: 
					[(currentCharacter between: $* and: $,) or: 
							[currentCharacter == $/ or: 
									[currentCharacter == $= or: 
											[(currentCharacter between: $? and: $@) 
												or: [currentCharacter == $\ or: [currentCharacter == $~]]]]]]]) 
		ifTrue: 
			[self recordMatch: #binarySymbol.
			self step.
			currentCharacter isSpecial 
				ifTrue: 
					[
					[self recordMatch: #binarySymbol.
					self step.
					currentCharacter isSpecial] 
							whileTrue.
					^ self reportLastMatch].
			^ self reportLastMatch].
	currentCharacter isDigit 
		ifTrue: 
			[self recordMatch: #(18 32).
			self step.
			currentCharacter isDigit 
				ifTrue: 
					[
					[self recordMatch: #(18).
					self step.
					currentCharacter isDigit] 
							whileTrue.
					currentCharacter == $. ifTrue: [^ self scan7].
					currentCharacter == $e ifTrue: [^ self scan1].
					currentCharacter == $r ifTrue: [^ self scan8].
					^ self reportLastMatch].
			currentCharacter == $. ifTrue: [^ self scan7].
			currentCharacter == $e ifTrue: [^ self scan1].
			currentCharacter == $r ifTrue: [^ self scan8].
			^ self reportLastMatch].
	(currentCharacter isSeparator or: [currentCharacter == $]) 
		ifTrue: 
			[self recordMatch: #whitespace.
			self step.
			(currentCharacter isSeparator or: [currentCharacter == $]) 
				ifTrue: 
					[
					[self recordMatch: #whitespace.
					self step.
					currentCharacter isSeparator or: [currentCharacter == $]] 
							whileTrue.
					^ self reportLastMatch].
			^ self reportLastMatch].
	currentCharacter == $" 
		ifTrue: 
			[self recordMatch: #(32).
			self step.
			currentCharacter ~~ $" 
				ifTrue: 
					[
					[self step.
					currentCharacter ~~ $"] whileTrue.
					^ self recordAndReportMatch: #comment].
			currentCharacter == $" ifTrue: [^ self recordAndReportMatch: #comment].
			^ self reportLastMatch].
	currentCharacter == $# ifTrue: [^ self recordAndReportMatch: #(3 32)].
	currentCharacter == $$ 
		ifTrue: 
			[self recordMatch: #(32).
			self step.
			^ self recordAndReportMatch: #(29)].
	currentCharacter == $' 
		ifTrue: 
			[self recordMatch: #(32).
			self step.
			currentCharacter ~~ $' ifTrue: [^ self scan3].
			currentCharacter == $' ifTrue: [^ self scan4].
			^ self reportLastMatch].
	currentCharacter == $( ifTrue: [^ self recordAndReportMatch: #(13 32)].
	currentCharacter == $) ifTrue: [^ self recordAndReportMatch: #(16 32)].
	currentCharacter == $- 
		ifTrue: 
			[self recordMatch: #binarySymbol.
			self step.
			currentCharacter isSpecial 
				ifTrue: 
					[
					[self recordMatch: #binarySymbol.
					self step.
					currentCharacter isSpecial] 
							whileTrue.
					^ self reportLastMatch].
			currentCharacter isDigit 
				ifTrue: 
					[
					[self recordMatch: #(19).
					self step.
					currentCharacter isDigit] 
							whileTrue.
					currentCharacter == $. 
						ifTrue: 
							[self step.
							currentCharacter isDigit 
								ifTrue: 
									[
									[self recordMatch: #(19).
									self step.
									currentCharacter isDigit] 
											whileTrue.
									currentCharacter == $e ifTrue: [^ self scan2].
									^ self reportLastMatch].
							^ self reportLastMatch].
					currentCharacter == $e ifTrue: [^ self scan2].
					currentCharacter == $r 
						ifTrue: 
							[self step.
							(currentCharacter isDigit or: [currentCharacter between: $A and: $Z]) 
								ifTrue: 
									[
									[self recordMatch: #(19).
									self step.
									currentCharacter isDigit or: [currentCharacter between: $A and: $Z]] 
											whileTrue.
									currentCharacter == $. 
										ifTrue: 
											[self step.
											(currentCharacter isDigit or: [currentCharacter between: $A and: $Z]) 
												ifTrue: 
													[
													[self recordMatch: #(19).
													self step.
													currentCharacter isDigit or: [currentCharacter between: $A and: $Z]] 
															whileTrue.
													currentCharacter == $e ifTrue: [^ self scan2].
													^ self reportLastMatch].
											^ self reportLastMatch].
									currentCharacter == $e ifTrue: [^ self scan2].
									^ self reportLastMatch].
							^ self reportLastMatch].
					^ self reportLastMatch].
			^ self reportLastMatch].
	currentCharacter == $. ifTrue: [^ self recordAndReportMatch: #(30 32)].
	currentCharacter == $: 
		ifTrue: 
			[self recordMatch: #(5 32).
			self step.
			((currentCharacter between: $A and: $Z) 
				or: [currentCharacter between: $a and: $z]) ifTrue: [^ self scan6].
			currentCharacter == $= ifTrue: [^ self recordAndReportMatch: #(25)].
			^ self reportLastMatch].
	currentCharacter == $; ifTrue: [^ self recordAndReportMatch: #(17 32)].
	currentCharacter == $< 
		ifTrue: 
			[self recordMatch: #(6 24 32).
			self step.
			currentCharacter isSpecial 
				ifTrue: 
					[
					[self recordMatch: #binarySymbol.
					self step.
					currentCharacter isSpecial] 
							whileTrue.
					^ self reportLastMatch].
			^ self reportLastMatch].
	currentCharacter == $> 
		ifTrue: 
			[self recordMatch: #(7 24 32).
			self step.
			currentCharacter isSpecial 
				ifTrue: 
					[
					[self recordMatch: #binarySymbol.
					self step.
					currentCharacter isSpecial] 
							whileTrue.
					^ self reportLastMatch].
			^ self reportLastMatch].
	currentCharacter == $[ ifTrue: [^ self recordAndReportMatch: #(1 32)].
	currentCharacter == $] ifTrue: [^ self recordAndReportMatch: #(2 32)].
	currentCharacter == $^ ifTrue: [^ self recordAndReportMatch: #(4 32)].
	currentCharacter == $_ ifTrue: [^ self recordAndReportMatch: #(25 32)].
	currentCharacter == ${ ifTrue: [^ self recordAndReportMatch: #(15 32)].
	currentCharacter == $| 
		ifTrue: 
			[self recordMatch: #(9 24 32).
			self step.
			(currentCharacter == $!! or: 
					[(currentCharacter between: $% and: $&) or: 
							[(currentCharacter between: $* and: $-) or: 
									[currentCharacter == $/ or: 
											[(currentCharacter between: $< and: $@) 
												or: [currentCharacter == $\ or: [currentCharacter == $~]]]]]]) 
				ifTrue: 
					[
					[self recordMatch: #binarySymbol.
					self step.
					currentCharacter isSpecial] 
							whileTrue.
					^ self reportLastMatch].
			currentCharacter == $| 
				ifTrue: 
					[self recordMatch: #(8 24).
					self step.
					currentCharacter isSpecial 
						ifTrue: 
							[
							[self recordMatch: #binarySymbol.
							self step.
							currentCharacter isSpecial] 
									whileTrue.
							^ self reportLastMatch].
					^ self reportLastMatch].
			^ self reportLastMatch].
	currentCharacter == $} ifTrue: [^ self recordAndReportMatch: #(14 32)].
	^ self reportLastMatch! !


!SqueakScanner methodsFor: 'token handling' stamp: 'ajh 7/19/2004 11:08'!
binarySymbol
	"Negative number takes precedence over binary symbol, so for example, '0@-1' means '0 @ -1', rather than '0 @- 1', which is how the scanner reads it. If the symbol is just $- and the next char is a digit, then the negative number token takes precedence and SqueakParser>>actionForCurrentToken converts it to a binary symbol. If there is space between $- and the digit, then it reaches here and we convert it to a negativeNumber token (and let SqueakParser>>actionForCurrentToken deal with it)."

	| string negPos prevReturnMatchBlock |
	string _ outputStream contents.
	string last = $- ifFalse: [
		matchActions _ {self binarySymbolId}.
		^ self createTokenFor: outputStream contents].

	stream peek isDigit ifTrue: [
		"Back up one and let scanner find negativeNumber token"
		outputStream skip: -1.
		stream skip: -1.
		matchActions _ {self binarySymbolId}.
		^ self createTokenFor: outputStream contents].

	(stream peek = $  and: [string size = 1]) ifFalse: [
		matchActions _ {self binarySymbolId}.
		^ self createTokenFor: outputStream contents].

	"Allow space between negative sign and number (wierd, but allowed in Smalltalk ANSI standard)"
	negPos _ stream position.
	[stream peek = $ ] whileTrue: [stream next].

	stream peek isDigit ifFalse: [
		stream position: negPos.
		matchActions _ {self binarySymbolId}.
		^ self createTokenFor: outputStream contents].

	"scan number but catch return and convert it to a negative number"
	prevReturnMatchBlock _ returnMatchBlock.
	returnMatchBlock _ [:token | 
		returnMatchBlock _ prevReturnMatchBlock.
		token id first = self numberId ifFalse: [self error: 'unexpected token'].
		token value: '-', token value
			start: token start - 1
			id: {self negativeNumberId}.
		returnMatchBlock value: token].
	self whitespace.  "eats whitespace and scans next token"! !

!SqueakScanner methodsFor: 'token handling' stamp: 'ajh 3/6/2003 00:15'!
comment

	comments add: outputStream contents.
	super comment.
! !

!SqueakScanner methodsFor: 'token handling' stamp: 'ajh 7/16/2004 10:57'!
variableAssignment

	outputStream skip: -2.
	stream skip: -2.
	matchActions _ {self nameId}.
	self createTokenFor: outputStream contents.
! !


!SqueakScanner methodsFor: 'private' stamp: 'ajh 3/6/2003 00:16'!
comments

	^ comments! !

!SqueakScanner methodsFor: 'private' stamp: 'ajh 3/6/2003 00:13'!
initialize

	super initialize.
	comments _ OrderedCollection new.
! !

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

SqueakScanner class
	instanceVariableNames: ''!

!SqueakScanner class methodsFor: 'generated-initialization' stamp: 'ajh 7/19/2004 11:54'!
initializeKeywordMap
	keywordMap _ Dictionary new.
	#(#(21 'false' 10) #(21 'nil' 11) #(21 'true' 12)) do: 
			[:each | 
			(keywordMap at: each first ifAbsentPut: [Dictionary new]) at: (each at: 2)
				put: each last].
	^ keywordMap! !


!SqueakScanner class methodsFor: 'generated-comments' stamp: 'ajh 7/19/2004 11:54'!
scannerDefinitionComment

	"<decimalNumber>:	[0-9]+ (\. [0-9]+)? ;
<radixNumber>:		[0-9]+ r [0-9A-Z]+ (\. [0-9A-Z]+)? ;
<exponentNumber>:	(<decimalNumber> | <radixNumber>) e \-? [0-9]+ ;
<number>:			<decimalNumber> | <radixNumber> | <exponentNumber> ;
<negativeNumber>:	\- <number> ;
<string>:				\'  [^\']* \' (\' [^\']* \')* ;
<name>:				[a-zA-Z] [a-zA-Z0-9]* ;
<keyword>:			<name> \: ;
<multikeyword>:		<name> \: (<name> \: )+ ;
<binarySymbol>:		[\~\!!\@\%\&\*\-\+\=\\\|\?\/\>\<\,] [\~\!!\@\%\&\*\-\+\=\\\|\?\/\>\<\,]* ;
<assignment>:		\: \= | \_ ;
<alternateKeyword>:	\: <name> \: (<name> \:)* ;
<whitespace>:		\s+ ;
<comment>:			\"" [^\""]* \"" ;
<character>:			\$ . ;
<period>:				\. ;
<variableAssignment>:	<name> \: \= ;

<anyChar>:			. ; # For VW literal arrays that handle #(;) -> #(#';')
"! !

Parser2 initialize!